//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file PorousMediaGPU.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================

#include "blockforest/Initialization.h"
#include "blockforest/SetupBlock.h"
#include "blockforest/SetupBlockForest.h"
#include "blockforest/loadbalancing/StaticCurve.h"
#include "blockforest/StructuredBlockForest.h"
#include "blockforest/communication/NonUniformBufferedScheme.h"

#include "core/Abort.h"
#include "core/DataTypes.h"
#include "core/SharedFunctor.h"
#include "core/debug/CheckFunctions.h"
#include "core/logging/Logging.h"
#include "core/math/Vector3.h"
#include "core/math/Constants.h"
#include "core/Environment.h"
#include "core/mpi/MPIManager.h"
#include "core/MemoryUsage.h"
#include "core/mpi/Reduce.h"
#include "core/timing/RemainingTimeLogger.h"
#include "core/timing/TimingPool.h"

#include "field/AddToStorage.h"
#include "field/CellCounter.h"
#include "field/FlagField.h"
#include "field/StabilityChecker.h"
#include "field/adaptors/AdaptorCreators.h"
#include "field/iterators/FieldIterator.h"
#include "field/vtk/VTKWriter.h"
#include "field/vtk/FlagFieldCellFilter.h"

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
#   include "gpu/AddGPUFieldToStorage.h"
#   include "gpu/DeviceSelectMPI.h"
#   include "gpu/ErrorChecking.h"
#   include "gpu/FieldCopy.h"
#   include "gpu/HostFieldAllocator.h"
#   include "gpu/ParallelStreams.h"
#   include "gpu/communication/NonUniformGPUScheme.h"
#   include "gpu/communication/UniformGPUScheme.h"
#endif

#include "geometry/InitBoundaryHandling.h"
#include "geometry/mesh/TriangleMeshIO.h"

#include "lbm_generated/communication/NonuniformGeneratedPdfPackInfo.h"
#include "lbm_generated/evaluation/PerformanceEvaluation.h"
#include "lbm_generated/field/AddToStorage.h"
#include "lbm_generated/field/PdfField.h"
#include "lbm_generated/refinement/BasicRecursiveTimeStep.h"

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
#   include "lbm_generated/gpu/AddToStorage.h"
#   include "lbm_generated/gpu/BasicRecursiveTimeStepGPU.h"
#   include "lbm_generated/gpu/GPUPdfField.h"
#   include "lbm_generated/gpu/NonuniformGeneratedGPUPdfPackInfo.h"
#   include "lbm_generated/gpu/UniformGeneratedGPUPdfPackInfo.h"
#endif

#include "timeloop/SweepTimeloop.h"
#include "vtk/VTKOutput.h"

#include <cstdlib>
#include <iostream>
#include <memory>

#include "ChannelStaticDefines.h"
#include "ChannelInfoHeader.h"

using namespace walberla;
using FlagField_T          = FlagField< uint8_t >;

using StorageSpecification_T = lbm::ChannelStorageSpecification;
using Stencil_T              = StorageSpecification_T::Stencil;
using CommunicationStencil_T = StorageSpecification_T::CommunicationStencil;
using BoundaryCollection_T = lbm::ChannelBoundaryCollection< FlagField_T >;
using SweepCollection_T = lbm::ChannelSweepCollection;

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
using PdfField_T = lbm_generated::PdfField< StorageSpecification_T >;
using GPUPdfField_T = lbm_generated::GPUPdfField< StorageSpecification_T >;
using gpu::communication::NonUniformGPUScheme;
using gpu::communication::UniformGPUScheme;

using lbm_generated::NonuniformGeneratedGPUPdfPackInfo;
using lbm_generated::UniformGeneratedGPUPdfPackInfo;
using RefinementTimeStep = lbm_generated::BasicRecursiveTimeStepGPU< GPUPdfField_T, SweepCollection_T, BoundaryCollection_T >;
#else
using PdfField_T = lbm_generated::PdfField< StorageSpecification_T >;
using blockforest::communication::NonUniformBufferedScheme;

using lbm_generated::NonuniformGeneratedPdfPackInfo;
using RefinementTimeStep = lbm_generated::BasicRecursiveTimeStep< PdfField_T, SweepCollection_T, BoundaryCollection_T >;
#endif

using macroFieldType = VelocityField_T::value_type;
using RefinementSelectionFunction = std::function<void (SetupBlockForest &)>;

struct IDs
{
   BlockDataID pdfField;
   BlockDataID velocityField;
   BlockDataID densityField;
   BlockDataID flagField;

   BlockDataID pdfFieldGPU;
   BlockDataID velocityFieldGPU;
   BlockDataID densityFieldGPU;
};

void createBlockForest(shared_ptr< BlockForest >& bfs, std::shared_ptr<Config>& config, const RefinementSelectionFunction& refinementSelection)
{
   auto domainParameters  = config->getOneBlock("DomainSetup");

   WALBERLA_LOG_INFO_ON_ROOT("Generating SetupBlockForest...")

   const uint_t numProcesses = uint_c(mpi::MPIManager::instance()->numProcesses());
   const std::string blockForestFilestem = "Channel";

   SetupBlockForest setupBfs;
   setupBfs.addRefinementSelectionFunction(refinementSelection);

   Vector3< uint_t > cellsPerBlock = domainParameters.getParameter< Vector3< uint_t > >("cellsPerBlock");
   Vector3< real_t > domainSize = domainParameters.getParameter< Vector3< real_t > >("domainSize");
   const real_t coarseMeshSize = domainParameters.getParameter< real_t >("resolution");
   Vector3< bool >  periodic    = domainParameters.getParameter< Vector3< bool > >("periodic");
   uint_t refinementLevels      = domainParameters.getParameter< uint_t >("refinementLevels");

   Vector3<uint_t> rootBlocks;
   rootBlocks[0] = uint_c(std::ceil( (domainSize[0] / coarseMeshSize) / real_c(cellsPerBlock[0])));
   rootBlocks[1] = uint_c(std::ceil( (domainSize[1] / coarseMeshSize) / real_c(cellsPerBlock[1])));
   rootBlocks[2] = uint_c(std::ceil( (domainSize[2] / coarseMeshSize) / real_c(cellsPerBlock[2])));

   const AABB domain(real_c(0.0), real_c(0.0), real_c(0.0), domainSize[0], domainSize[1], domainSize[2]);
   setupBfs.addWorkloadMemorySUIDAssignmentFunction(blockforest::uniformWorkloadAndMemoryAssignment);
   setupBfs.init(domain, rootBlocks[0], rootBlocks[1], rootBlocks[2],
                         periodic[0], periodic[1], periodic[2]);
   setupBfs.balanceLoad(blockforest::StaticLevelwiseCurveBalanceWeighted(), numProcesses);
   bfs = std::make_shared< BlockForest >(uint_c(MPIManager::instance()->worldRank()), setupBfs);

   WALBERLA_LOG_INFO_ON_ROOT("===========================  BLOCK FOREST STATISTICS ============================");
   WALBERLA_LOG_INFO_ON_ROOT("Blocks created: " << setupBfs.getNumberOfBlocks())
   for (uint_t level = 0; level <= refinementLevels; level++){
      const uint_t numberOfBlocks = setupBfs.getNumberOfBlocks(level);
      WALBERLA_LOG_INFO_ON_ROOT("Level " << level << " Blocks: " << numberOfBlocks)
   }

   const real_t avgBlocksPerProc = real_c(setupBfs.getNumberOfBlocks()) / real_c(setupBfs.getNumberOfProcesses());
   WALBERLA_LOG_INFO_ON_ROOT("Average blocks per process: " << avgBlocksPerProc);

   const uint_t blockCells        = cellsPerBlock[0] * cellsPerBlock[1] * cellsPerBlock[2];;
   const uint_t totalNumberCells     = setupBfs.getNumberOfBlocks() * blockCells;
   const real_t averageCellsPerGPU   = avgBlocksPerProc * real_c(blockCells);

   const uint_t PDFsPerCell            = StorageSpecification_T::inplace ? Stencil_T::Q : 2 * Stencil_T::Q;
   const uint_t valuesPerCell          = (PDFsPerCell + VelocityField_T::F_SIZE + ScalarField_T::F_SIZE);
   const uint_t sizePerValue           = sizeof(StorageSpecification_T::value_type);
   const double expectedMemory         = double_c(totalNumberCells * valuesPerCell * sizePerValue) * 1e-9;

   WALBERLA_LOG_INFO_ON_ROOT( "Total number of cells will be " << totalNumberCells << " fluid cells (in total on all levels)")
   WALBERLA_LOG_INFO_ON_ROOT( "Expected total memory demand will be " << expectedMemory << " GB")

   WALBERLA_LOG_INFO_ON_ROOT("=================================================================================")

   if(mpi::MPIManager::instance()->numProcesses() > 1)
      return;

   auto logging           = config->getOneBlock("Logging");
   bool writeSetupForestAndReturn    = logging.getParameter< bool >("writeSetupForestAndReturn", false);

   if(writeSetupForestAndReturn){
      std::ostringstream oss;
      oss << blockForestFilestem << ".bfs";
      setupBfs.saveToFile(oss.str().c_str());

      setupBfs.writeVTKOutput(blockForestFilestem);
   }
}

void refineWalls(SetupBlockForest& forest, const uint_t level)
{
   auto domain = forest.getDomain();
   const AABB minAABB(domain.xMin(), domain.yMin(), domain.zMin(), domain.xMax(), domain.yMin() + 1.0, domain.zMax());

   for (auto block = forest.begin(); block != forest.end(); ++block){
      const AABB& aabb = block->getAABB();

      if (block->getLevel() < level && minAABB.intersects(aabb) ){
         block->setMarker(true);
      }
   }
}

void defineObstacles(const shared_ptr< StructuredBlockStorage >& blocks, const BlockDataID flagFieldID, const FlagUID obstacleUID )
{
   auto domain = blocks->getDomain();
   const real_t xMin = domain.xMin();
   const real_t xMax = domain.xMax();

   for (auto bIt = blocks->begin(); bIt != blocks->end(); ++bIt){
      Block& b             = dynamic_cast< Block& >(*bIt);
      const auto flagField    = b.getData< FlagField_T >(flagFieldID);
      const auto boundaryFlag = flagField->getOrRegisterFlag(obstacleUID);

      for( auto it = flagField->beginWithGhostLayer(1); it != flagField->end(); ++it ){
         Vector3< real_t > cellCenter = blocks->getBlockLocalCellCenter( b, it.cell() );
         blocks->mapToPeriodicDomain(cellCenter);

         if(cellCenter[0] > 1.0 && cellCenter[0] < 2.0 && cellCenter[1] > 1.2 && cellCenter[1] < 2.0 && cellCenter[2] > 1.2 && cellCenter[2] < 2.0 ){
            addFlag(flagField->get(it.x(), it.y(), it.z()), boundaryFlag);

         }
      }
   }
}

int main(int argc, char** argv){
   Environment env( argc, argv );
   mpi::MPIManager::instance()->useWorldComm();
   #if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
      gpu::selectDeviceBasedOnMpiRank();
   #endif

   auto config = env.config();

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                        SETUP AND CONFIGURATION                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto parameters        = config->getOneBlock("Parameters");
   auto loggingParameters = config->getOneBlock("Logging");
   auto boundariesConfig  = config->getBlock("Boundaries");
   auto domainParameters  = config->getOneBlock("DomainSetup");


   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                      Parameters and initialisation                                         ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   Vector3< real_t > domainSize = domainParameters.getParameter< Vector3< real_t > >("domainSize");
   const real_t coarseMeshSize = domainParameters.getParameter< real_t >("resolution");

   const real_t reynoldsNumber = parameters.getParameter< real_t >("reynoldsNumber");
   const real_t referenceVelocity = parameters.getParameter< real_t >("referenceVelocity");
   const real_t drivingForce = parameters.getParameter< real_t >("drivingForce");
   
   const real_t referenceVelocity_L = parameters.getParameter< real_t >("latticeVelocity");
   const real_t referenceLength_L = (domainSize[0] / coarseMeshSize) / real_c(2.0);
   const real_t dt = referenceVelocity_L / referenceVelocity * coarseMeshSize;
   const real_t viscosity_L = real_c((referenceVelocity_L * referenceLength_L) / reynoldsNumber);
   const real_t omega = real_c(real_c(1.0) / (real_c(3.0) * viscosity_L + real_c(0.5)));
   
   const bool initialiseWithVelocity = parameters.getParameter< bool >("initialiseWithVelocity");
   const uint_t timesteps = parameters.getParameter< uint_t >("timesteps");

   bool writeSetupForestAndReturn = loggingParameters.getParameter< bool >("writeSetupForestAndReturn", false);
   if (uint_c(MPIManager::instance()->numProcesses()) > 1) writeSetupForestAndReturn = false;

   const uint_t refinementLevels = domainParameters.getParameter< uint_t >("refinementLevels");
   Vector3< uint_t > cellsPerBlock = domainParameters.getParameter< Vector3< uint_t > >("cellsPerBlock");
   
   auto refinementSelection = std::bind(refineWalls, std::placeholders::_1, refinementLevels);

   shared_ptr< BlockForest > bfs;
   createBlockForest(bfs, config, refinementSelection);

   WALBERLA_LOG_INFO_ON_ROOT( "Benchmark run data:"
                           "\n- simulation parameters:"   << std::setprecision(16) <<
                           "\n   + collision model:     " << infoMap["collisionOperator"] <<
                           "\n   + stencil:             " << infoMap["stencil"] <<
                           "\n   + streaming:           " << infoMap["streamingPattern"] <<
                           "\n   + compressible:        " << ( StorageSpecification_T::compressible ? "yes" : "no" ) <<
                           "\n   + mesh levels:         " << refinementLevels + uint_c(1) <<
                           "\n   + resolution:          " << coarseMeshSize << " - on the coarsest grid" <<
                           "\n- simulation properties:  "
                           "\n   + viscosity LB:        " << viscosity_L  << " [dx*dx/dt] (on the coarsest grid)" <<
                           "\n   + omega:               " << omega << " (on the coarsest grid)" <<
                           "\n   + inflow velocity:     " << referenceVelocity << " [m/s]" <<
                           "\n   + lattice velocity:    " << referenceVelocity_L << " [dx/dt]" <<
                           "\n   + Reynolds number:     " << reynoldsNumber <<
                           "\n   + dt (coarsest grid):  " << dt << " [s]"
                           "\n   + #time steps:         " << timesteps << " (on the coarsest grid, " << ( real_c(1.0) / dt ) << " for 1s of real time)" <<
                           "\n   + simulation time:     " << ( real_c( timesteps ) * dt ) << " [s]" )

   if (writeSetupForestAndReturn && mpi::MPIManager::instance()->numProcesses() == 1){
      WALBERLA_LOG_INFO_ON_ROOT("BlockForest has been created and writen to file. Returning program")
      logging::Logging::printFooterOnStream();
      return EXIT_SUCCESS;
   }

   auto blocks = std::make_shared< StructuredBlockForest >(bfs, cellsPerBlock[0], cellsPerBlock[1], cellsPerBlock[2]);
   blocks->createCellBoundingBoxes();

   auto finalDomain = blocks->getDomain();

   const StorageSpecification_T StorageSpec = StorageSpecification_T();
   // Creating fields
   uint_t step = 0;
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Allocating data")
   IDs ids;

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   auto PDFAllocator = make_shared< gpu::HostFieldAllocator< PdfField_T::value_type > >();
   auto allocator = make_shared< gpu::HostFieldAllocator< VelocityField_T::value_type > >();
   ids.pdfField = lbm_generated::addPdfFieldToStorage(blocks, "pdfs", StorageSpec, uint_c(2), field::fzyx, PDFAllocator);
   ids.velocityField = field::addToStorage< VelocityField_T >(blocks, "velocity", real_c(0.0), field::fzyx, uint_c(2), allocator);
   ids.densityField = field::addToStorage< ScalarField_T >(blocks, "density", real_c(1.0), field::fzyx, uint_c(2), allocator);
   ids.flagField = field::addFlagFieldToStorage< FlagField_T >(blocks, "Boundary Flag Field", uint_c(3));

   ids.pdfFieldGPU = lbm_generated::addGPUPdfFieldToStorage< PdfField_T >(blocks, ids.pdfField, StorageSpec, "pdfs on GPU", true);
   ids.velocityFieldGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.velocityField, "velocity on GPU", true);
   ids.densityFieldGPU = gpu::addGPUFieldToStorage< ScalarField_T >(blocks, ids.densityField, "density on GPU", true);

   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#else
   ids.pdfField = lbm_generated::addPdfFieldToStorage(blocks, "pdfs", StorageSpec, uint_c(2), field::fzyx);
   ids.velocityField = field::addToStorage< VelocityField_T >(blocks, "vel", VelocityField_T::value_type(0.0), field::fzyx, uint_c(2));
   ids.densityField = field::addToStorage< ScalarField_T >(blocks, "density", ScalarField_T::value_type(1.0), field::fzyx, uint_c(2));
   ids.flagField = field::addFlagFieldToStorage< FlagField_T >(blocks, "Boundary Flag Field", uint_c(3));
#endif


#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   SweepCollection_T sweepCollection(blocks, ids.pdfFieldGPU, ids.densityFieldGPU, ids.velocityFieldGPU, drivingForce, omega);
#else
   SweepCollection_T sweepCollection(blocks, ids.pdfField, ids.densityField, ids.velocityField, drivingForce, omega);
#endif

   if(initialiseWithVelocity){
      for (auto& block : *blocks){
         auto velocity = block.getData< VelocityField_T >(ids.velocityField);
         WALBERLA_FOR_ALL_CELLS_INCLUDING_GHOST_LAYER_XYZ(velocity,
            velocity->get(x, y, z, 0) = referenceVelocity_L;
         )
      }
   }

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   gpu::fieldCpy< gpu::GPUField< VelocityField_T::value_type >, VelocityField_T >(blocks, ids.velocityFieldGPU, ids.velocityField);
#endif

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Initialising data")
   for (auto& block : *blocks){
      sweepCollection.initialise(&block, cell_idx_c(1));
   }

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                      LB SWEEPS AND BOUNDARY HANDLING                                       ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto fluidUID    = FlagUID("Fluid");
   auto obstacleUID = FlagUID("NoSlip");

   // Boundaries
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setting boundary conditions")

   geometry::initBoundaryHandling< FlagField_T >(*blocks, ids.flagField, boundariesConfig);
   defineObstacles(blocks, ids.flagField, obstacleUID);
   geometry::setNonBoundaryCellsToDomain< FlagField_T >(*blocks, ids.flagField, fluidUID, 2);


#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfFieldGPU, fluidUID, referenceVelocity_L, ids.pdfField);
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#else
   BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfField, fluidUID, referenceVelocity_L);
#endif


   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                           COMMUNICATION SCHEME                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_LOG_INFO_ON_ROOT("Setting up communication...")
   const bool gpuEnabledMPI       = parameters.getParameter< bool >("gpuEnabledMPI", false);

   auto nonUniformCommunication = std::make_shared< NonUniformGPUScheme< CommunicationStencil_T > >(blocks, gpuEnabledMPI);
   auto nonUniformPackInfo      = lbm_generated::setupNonuniformGPUPdfCommunication< GPUPdfField_T >(blocks, ids.pdfFieldGPU);
   nonUniformCommunication->addPackInfo(nonUniformPackInfo);
#else
   WALBERLA_LOG_INFO_ON_ROOT("Setting up communication...")
   auto nonUniformCommunication = std::make_shared< NonUniformBufferedScheme< CommunicationStencil_T > >(blocks);
   auto nonUniformPackInfo      = lbm_generated::setupNonuniformPdfCommunication< PdfField_T >(blocks, ids.pdfField);
   nonUniformCommunication->addPackInfo(nonUniformPackInfo);

#endif

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                          TIME STEP DEFINITIONS                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   SweepTimeloop timeloop(blocks->getBlockStorage(), timesteps);

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   auto LBMRefinement = std::make_shared<RefinementTimeStep>(blocks, ids.pdfFieldGPU, sweepCollection, boundaryCollection, nonUniformCommunication, nonUniformPackInfo);
   LBMRefinement->addRefinementToTimeLoop(timeloop, uint_c(0));
#else
   auto LBMRefinement = std::make_shared<RefinementTimeStep>(blocks, ids.pdfField, sweepCollection, boundaryCollection, nonUniformCommunication, nonUniformPackInfo);
   LBMRefinement->addRefinementToTimeLoop(timeloop);
#endif

   auto dxF = blocks->dz(refinementLevels);
   const AABB sliceXY(finalDomain.xMin(), finalDomain.yMin(), finalDomain.center()[2] - dxF,
                      finalDomain.xMax(), finalDomain.yMax(), finalDomain.center()[2] + dxF);


   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                             VTK Output                                                     ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   // VTtimeloop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");K
   auto VTKWriter = config->getOneBlock("VTKWriter");
   const uint_t vtkWriteFrequency       = VTKWriter.getParameter< uint_t >("vtkWriteFrequency", 0);
   const real_t sliceThickness          = VTKWriter.getParameter< real_t >("sliceThickness", real_c(1.0));

   if (vtkWriteFrequency > 0){
      auto vtkOutput =
         vtk::createVTKOutput_BlockData(*blocks, "vtk", vtkWriteFrequency, 0, false, "vtk_out",
                                        "simulation_step", false, true, true, false, 0, false, false);

      vtkOutput->addBeforeFunction([&]() {
         for (auto& block : *blocks){
            sweepCollection.calculateMacroscopicParameters(&block);
         }
      });


      field::FlagFieldCellFilter<FlagField_T> fluidFilter( ids.flagField );
      fluidFilter.addFlag(obstacleUID );
      vtkOutput->addCellExclusionFilter(fluidFilter);


      if (VTKWriter.getParameter< bool >("writeOnlySlice", true)){
         vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXY));
      }

      if (VTKWriter.getParameter< bool >("velocity")){
         if (VTKWriter.getParameter< bool >("writeVelocityAsMagnitude")){
            auto velWriter = make_shared< field::VTKMagnitudeWriter< VelocityField_T, float32 > >(ids.velocityField, "velocityMagnitude");
            vtkOutput->addCellDataWriter(velWriter);
         }
         else{
            auto velWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(ids.velocityField, "velocity");
            vtkOutput->addCellDataWriter(velWriter);
         }
      }
      if (VTKWriter.getParameter< bool >("density")){
         auto densityWriter = make_shared< field::VTKWriter< ScalarField_T, float32 > >(ids.densityField, "density");
         vtkOutput->addCellDataWriter(densityWriter);
      }
      if (VTKWriter.getParameter< bool >("flag")){
         auto flagWriter = make_shared< field::VTKWriter< FlagField_T > >(ids.flagField, "flag");
         vtkOutput->addCellDataWriter(flagWriter);
      }
      timeloop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");
   }

   // log remaining time
   timeloop.addFuncAfterTimeStep(
      timing::RemainingTimeLogger(timeloop.getNrOfTimeSteps(), 20),
      "remaining time logger");

   auto CheckerParameters      = config->getOneBlock("StabilityChecker");
   const uint_t checkFrequency = CheckerParameters.getParameter< uint_t >("checkFrequency", uint_t(0));
   if (checkFrequency > 0){
      auto checkFunction = [](VelocityField_T::value_type value) {  return value < math::abs(VelocityField_T::value_type(10)); };
      timeloop.addFuncAfterTimeStep(
         makeSharedFunctor(field::makeStabilityChecker< VelocityField_T, FlagField_T >(
            config, blocks, ids.velocityField, ids.flagField, fluidUID, checkFunction)),
         "Stability check");
   }

   WALBERLA_MPI_BARRIER()

   const lbm_generated::PerformanceEvaluation< FlagField_T > performance(blocks, ids.flagField, fluidUID);
   field::CellCounter< FlagField_T > fluidCells(blocks, ids.flagField, fluidUID);
   fluidCells();

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Starting Simulation with " << timesteps << " timesteps")
   WcTimingPool timeloopTiming;
   WcTimer simTimer;

   WALBERLA_MPI_BARRIER()

   simTimer.start();
   timeloop.run(timeloopTiming);
   WALBERLA_MPI_BARRIER()
   simTimer.end();
   WALBERLA_LOG_INFO_ON_ROOT("Simulation finished")
   double time = double_c(simTimer.max());
   WALBERLA_MPI_SECTION() { walberla::mpi::reduceInplace(time, walberla::mpi::MAX); }
   performance.logResultOnRoot(timesteps, time);

   const auto reducedTimeloopTiming = timeloopTiming.getReduced();
   WALBERLA_LOG_RESULT_ON_ROOT("Time loop timing:\n" << *reducedTimeloopTiming)

   printResidentMemoryStatistics();
   return EXIT_SUCCESS;
}