//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file PorousMediaGPU.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================

#include "blockforest/Initialization.h"
#include "blockforest/StructuredBlockForest.h"
#include "blockforest/communication/UniformBufferedScheme.h"

#include "core/Abort.h"
#include "core/DataTypes.h"
#include "core/SharedFunctor.h"
#include "core/debug/CheckFunctions.h"
#include "core/logging/Logging.h"
#include "core/math/Vector3.h"
#include "core/math/Constants.h"
#include "core/Environment.h"
#include "core/mpi/MPIManager.h"
#include "core/MemoryUsage.h"
#include "core/mpi/Reduce.h"
#include "core/timing/RemainingTimeLogger.h"
#include "core/timing/TimingPool.h"

#include "field/AddToStorage.h"
#include "field/CellCounter.h"
#include "field/FlagField.h"
#include "field/StabilityChecker.h"
#include "field/adaptors/AdaptorCreators.h"
#include "field/iterators/FieldIterator.h"
#include "field/vtk/VTKWriter.h"
#include "field/vtk/FlagFieldCellFilter.h"

#include "geometry/InitBoundaryHandling.h"
#include "geometry/mesh/TriangleMeshIO.h"

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
#   include "gpu/AddGPUFieldToStorage.h"
#   include "gpu/DeviceSelectMPI.h"
#   include "gpu/ErrorChecking.h"
#   include "gpu/FieldCopy.h"
#   include "gpu/HostFieldAllocator.h"
#   include "gpu/ParallelStreams.h"
#   include "gpu/communication/NonUniformGPUScheme.h"
#endif

#include "lbm_generated/communication/UniformGeneratedPdfPackInfo.h"
#include "lbm_generated/evaluation/PerformanceEvaluation.h"
#include "lbm_generated/field/AddToStorage.h"
#include "lbm_generated/field/PdfField.h"

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
#   include "lbm_generated/gpu/AddToStorage.h"
#   include "lbm_generated/gpu/GPUPdfField.h"
#   include "lbm_generated/gpu/NonuniformGeneratedGPUPdfPackInfo.h"
#   include "lbm_generated/gpu/BasicRecursiveTimeStepGPU.h"
#endif

#include "postprocessing/FieldToSurfaceMesh.h"

#include "timeloop/SweepTimeloop.h"
#include "vtk/VTKOutput.h"

#include <cstdlib>
#include <iostream>
#include <memory>

#include "DragEvaluation.h"
#include "Inflow.h"
#include "InitializerFunctions.h"
#include "GridGeneration.h"
#include "Setup.h"
#include "Probe.h"
#include "TPMS.h"
#include "Types.h"
#include "util.h"

#include "PorousMediaGPUStaticDefines.h"

using namespace walberla;
using macroFieldType = VelocityField_T::value_type;

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
using GPUPdfField_T = lbm_generated::GPUPdfField< StorageSpecification_T >;
using gpu::communication::NonUniformGPUScheme;
using lbm_generated::NonuniformGeneratedGPUPdfPackInfo;
#else
using PdfField_T = lbm_generated::PdfField< StorageSpecification_T >;
using blockforest::communication::UniformBufferedScheme;

using lbm_generated::UniformGeneratedPdfPackInfo;
#endif

inline Vector3< uint_t > inflowSliceSize( const shared_ptr< StructuredBlockStorage > & blocks, IBlock * const block ){
   return Vector3<uint_t>( uint_c(1), blocks->getNumberOfYCells( *block ), blocks->getNumberOfZCells( *block ) );
}

class HemisphereWallDistance
{
 public:
   HemisphereWallDistance(const real_t radius, const real_t hemispheresDistance) : radius_(radius), hemispheresDistance_(hemispheresDistance){} 

   real_t operator()(const Cell& fluidCell, const Cell& boundaryCell,
                     const shared_ptr< StructuredBlockForest >& SbF, IBlock& block){

      auto domain = SbF->getDomain();
      const real_t yMin = domain.yMin();
      const real_t yMax = domain.yMax();

      Vector3< real_t > boundary = SbF->getBlockLocalCellCenter( block, boundaryCell );
      Vector3< real_t > fluid    = SbF->getBlockLocalCellCenter( block, fluidCell );

      auto center = getCenterOfHemisphere(boundary, radius_, hemispheresDistance_, yMin, yMax);

      hemispheresContains(boundary, center, radius_);

      WALBERLA_CHECK(!hemispheresContains(fluid, center, radius_), "fluid cell is in contained in Hemisphere (" << fluid[0] << ", " << fluid[1] << ", " << fluid[2] << "). The block local cell is " << fluidCell)
      WALBERLA_CHECK(hemispheresContains(boundary, center, radius_), "boundary cell is not in contained in Hemisphere (" << boundary[0] << ", " << boundary[1] << ", " << boundary[2] << "). The block local cell is " << boundaryCell)

      return delta(fluid, boundary, center);
   }

   real_t delta(const Vector3< real_t >& fluid, const Vector3< real_t >& boundary, const Vector3< real_t >& center) const
   {
      // http://devmag.org.za/2009/04/17/basic-collision-detection-in-2d-part-2/
      const Vector3< real_t > f = fluid - center;
      const Vector3< real_t > d = (boundary - center) - f;

      const real_t dDotd = d[0] * d[0] + d[1] * d[1] + d[2] * d[2];
      const real_t fDotf = f[0] * f[0] + f[1] * f[1] + f[2] * f[2];
      const real_t dDotf = d[0] * f[0] + d[1] * f[1] + d[2] * f[2];

      const real_t b = real_c(2.0) * dDotf;
      const real_t c = fDotf - (radius_ * radius_);

      const real_t bb4ac = b * b - (real_c(4.0) * dDotd * c);
      WALBERLA_CHECK_GREATER_EQUAL(bb4ac, real_c(0.0))

      const real_t sqrtbb4ac = std::sqrt(bb4ac);
      const real_t alpha = std::min((-b + sqrtbb4ac) / (real_c(2.0) * dDotd), (-b - sqrtbb4ac) / (real_c(2.0) * dDotd));

      WALBERLA_CHECK_GREATER_EQUAL(alpha, real_c(0.0))
      WALBERLA_CHECK_LESS_EQUAL(alpha, real_c(1.0))

      return alpha;
   }


private:
   const real_t radius_;
   const real_t hemispheresDistance_;
};

real_t defaultWallDistance(const Cell& /*fluidCell*/, const Cell& /*boundaryCell*/, const shared_ptr< StructuredBlockForest >& /*SbF*/, IBlock& /*block*/){
   return real_c(0.5);
}

void refineWalls(SetupBlockForest& forest, const real_t hemispheresRadius, const real_t dx, const uint_t level)
{
   auto domain = forest.getDomain();
   const real_t yMin = domain.yMin();
   const real_t yMax = domain.yMax();

   const real_t t = hemispheresRadius + dx;

   const AABB minAABB(domain.xMin(), domain.yMin(), domain.zMin(), domain.xMax(), domain.yMin() + t, domain.zMax());
   // const AABB maxAABB(domain.xMin(), domain.yMax() - t, domain.zMin(), domain.xMax(), domain.yMax(), domain.zMax());

   for (auto block = forest.begin(); block != forest.end(); ++block){
      const AABB& aabb = block->getAABB();

      // if (block->getLevel() < level && (minAABB.intersects(aabb) || maxAABB.intersects(aabb)) ){
      if (block->getLevel() < level && minAABB.intersects(aabb) ){
         block->setMarker(true);
      }
   }
}



int main(int argc, char** argv){
   Environment env( argc, argv );
   mpi::MPIManager::instance()->useWorldComm();
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   gpu::selectDeviceBasedOnMpiRank();
#endif

   auto config = env.config();

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                        SETUP AND CONFIGURATION                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto parameters        = config->getOneBlock("Parameters");
   auto loggingParameters = config->getOneBlock("Logging");
   auto boundariesConfig  = config->getBlock("Boundaries");

   Setup setup(config, infoMap);
   WALBERLA_LOG_INFO_ON_ROOT( "Simulation run data:"
                              "\n- simulation parameters:"   << std::setprecision(16) <<
                              "\n   + collision model:     " << setup.collisionModel <<
                              "\n   + stencil:             " << setup.stencil <<
                              "\n   + streaming:           " << setup.streamingPattern <<
                              "\n   + compressible:        " << ( StorageSpecification_T::compressible ? "yes" : "no" ) <<
                              "\n- simulation properties:  "
                              "\n   + kin. viscosity:      " << setup.kinViscosity << " [m^2/s] (on the coarsest grid)" <<
                              "\n   + viscosity LB:        " << setup.viscosity  << " [dx*dx/dt] (on the coarsest grid)" <<
                              "\n   + omega:               " << setup.omega << " (on the coarsest grid)" <<
                              "\n   + omega:               " << setup.omegaF << " (on the finest grid)" <<
                              "\n   + inflow velocity:     " << setup.referenceVelocity << " [m/s]" <<
                              "\n   + lattice velocity:    " << setup.latticeVelocity << " [dx/dt]" <<
                              "\n   + Reynolds number:     " << setup.reynoldsNumber <<
                              "\n   + Mach number:         " << setup.machNumber <<
                              "\n   + driving force:       " << setup.drivingForce <<
                              "\n   + dx (coarsest grid):  " << setup.dxC << " [m]" <<
                              "\n   + dx (finest grid):    " << setup.dxF << " [m]" <<
                              "\n   + dt (coarsest grid):  " << setup.dt << " [s]" <<
                              "\n   + conversion Faktor:   " << setup.conversionFactor <<
                              "\n   + #time steps:         " << setup.timesteps << " (on the coarsest grid, " << ( real_c(1.0) / setup.dt ) << " for 1s of real time)" <<
                              "\n   + simulation time:     " << ( real_c( setup.timesteps ) * setup.dt ) << " [s]" )

   bool writeSetupForestAndReturn = loggingParameters.getParameter< bool >("writeSetupForestAndReturn", false);
   if (uint_c(MPIManager::instance()->numProcesses()) > 1) writeSetupForestAndReturn = false;
   shared_ptr< BlockForest > bfs;

   auto refinementSelection = std::bind(refineWalls, _1, real_c(1.0), setup.dxF, setup.refinementLevels);

   createBlockForest(bfs, setup, refinementSelection);

   if (writeSetupForestAndReturn && mpi::MPIManager::instance()->numProcesses() == 1){
      WALBERLA_LOG_INFO_ON_ROOT("BlockForest has been created and writen to file. Returning program")
      logging::Logging::printFooterOnStream();
      return EXIT_SUCCESS;
   }

   auto blocks = std::make_shared< StructuredBlockForest >(bfs, setup.cellsPerBlock[0], setup.cellsPerBlock[1], setup.cellsPerBlock[2]);
   blocks->createCellBoundingBoxes();
   auto finalDomain = blocks->getDomain();

   const StorageSpecification_T StorageSpec = StorageSpecification_T();
   // Creating fields
   uint_t step = 0;
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Allocating data")
   IDs ids;
   ids.pdfField  = lbm_generated::addPdfFieldToStorage(blocks, "pdfs", StorageSpec, uint_c(1), field::fzyx);

   auto allocator = make_shared< gpu::HostFieldAllocator<macroFieldType> >(); // use pinned memory allocator for faster CPU-GPU memory transfers
   ids.velocityField = field::addToStorage< VelocityField_T >(blocks, "vel", macroFieldType(0.0), field::fzyx, uint_c(1), allocator);
   ids.meanVelocityField = field::addToStorage< VelocityField_T >(blocks, "mean velocity", macroFieldType(0.0), field::fzyx, uint_c(1), allocator);
   ids.densityField  = field::addToStorage< ScalarField_T >(blocks, "density", macroFieldType(1.0), field::fzyx, uint_c(1), allocator);
   ids.flagField     = field::addFlagFieldToStorage< FlagField_T >(blocks, "Boundary Flag Field");

   ids.pdfFieldGPU      = lbm_generated::addGPUPdfFieldToStorage< PdfField_T >(blocks, ids.pdfField, StorageSpec, "pdfs on GPU", true);
   ids.velocityFieldGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.velocityField, "velocity on GPU", true);
   ids.meanVelocityFieldGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.meanVelocityField, "mean velocity on GPU", true);
   ids.densityFieldGPU  = gpu::addGPUFieldToStorage< ScalarField_T >(blocks, ids.densityField, "velocity on GPU", true);


   std::function< Vector3< uint_t > (const shared_ptr< StructuredBlockStorage >&, IBlock * const) > getInflowSliceSize = inflowSliceSize;
   ids.velocityFieldInflowSlice    = field::addToStorage< VelocityField_T >(blocks, "velocity Inflow", getInflowSliceSize, macroFieldType(0.0), field::fzyx, uint_c(1));
   ids.velocityFieldInflowSliceGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.velocityFieldInflowSlice, "velocity Inflow on GPU", true);

   const Cell innerOuterSplit = Cell(parameters.getParameter< Vector3<cell_idx_t> >("innerOuterSplit", Vector3<cell_idx_t>(1, 1, 1)));
   SweepCollection_T sweepCollection(blocks, ids.pdfFieldGPU, ids.densityFieldGPU, ids.velocityFieldGPU, setup.drivingForce, setup.omega, innerOuterSplit);
   WelfordSweep_T welfordSweep(ids.meanVelocityFieldGPU, ids.velocityFieldGPU, real_c(0.0));

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Initialising data")
   for (auto& block : *blocks){
      sweepCollection.initialise(&block);

      auto velocityInflow = block.getData< VelocityField_T >(ids.velocityFieldInflowSlice);
      WALBERLA_FOR_ALL_CELLS_INCLUDING_GHOST_LAYER_XYZ(velocityInflow,
         velocityInflow->get(x, y, z, 0) = setup.latticeVelocity;
      )
   }
   sweepCollection.initialiseBlockPointer();
   gpu::fieldCpy< gpu::GPUField< VelocityField_T::value_type >, VelocityField_T >(blocks, ids.velocityFieldInflowSliceGPU, ids.velocityFieldInflowSlice);

   auto inflow = std::make_shared<Inflow>(blocks, setup, ids);
   auto probe = std::make_shared<Probe>(blocks, setup, ids, config->getOneBlock("Probes"));

   const real_t lenght = 8.0;
   const real_t res = 3.0;
   const real_t start = 0.0;

/*
   const uint_t nx = uint_c(std::floor(lenght / (setup.dxC * res)));
   for(uint_t i = 0; i < nx; ++i){
      probe->addPoint(Vector3<real_t>(start + real_c(i) * (setup.dxC * res), 0.5, 0.5 ));
   }
*/ 
   probe->initialise();

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                      LB SWEEPS AND BOUNDARY HANDLING                                       ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   // Boundaries
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setting boundary conditions")
   for (auto& block : *blocks){
      auto flagField = block.getData< FlagField_T >(ids.flagField);
      setup.inflowFlag          = flagField->registerFlag(setup.inflowUID);
      setup.turbulentInflowFlag = flagField->registerFlag(setup.turbulentInflowUID);
      setup.outflowFlag         = flagField->registerFlag(setup.outflowUID);
      setup.obstacleFlag        = flagField->registerFlag(setup.obstacleUID);
      setup.noSlipFlag          = flagField->registerFlag(setup.noSlipUID);
      setup.fluidFlag           = flagField->registerFlag(setup.fluidUID);
   }


   geometry::initBoundaryHandling< FlagField_T >(*blocks, ids.flagField, boundariesConfig);
   std::function< VelocityField_T::value_type (const Cell&, const Cell&, const shared_ptr< StructuredBlockForest >&, IBlock&) > wallDistanceFunctor = defaultWallDistance;
   if(setup.useGrid){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setting Grid after inflow to produce turbulence structure")
      initGrid(blocks, ids.flagField, setup);
   }

/*
   if(config->getNumBlocks("RoughWall") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Using RoughWall setup")
      auto wallParameters = config->getOneBlock("RoughWall");
      const real_t radius   = wallParameters.getParameter< real_t >("hemispheresRadius");
      const real_t hemispheresDistance = wallParameters.getParameter< real_t >("hemispheresDistance");
      initHemisphers(blocks, ids.flagField, setup, config->getOneBlock("RoughWall"));
      HemisphereWallDistance wallDist(radius, hemispheresDistance);
      wallDistanceFunctor = wallDist;
   }
*/
   if(config->getNumBlocks("TPMS") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Placing TPMS material")
      auto tpms = std::make_shared<TPMS>(config->getOneBlock("TPMS"), setup);
      tpms->setupBoundary(blocks, ids.flagField);
   }


   geometry::setNonBoundaryCellsToDomain< FlagField_T >(*blocks, ids.flagField, setup.fluidUID);
   // BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfFieldGPU, setup.fluidUID, ids.velocityFieldInflowSliceGPU, setup.drivingForce);
   BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfFieldGPU, setup.fluidUID, ids.velocityFieldInflowSliceGPU, setup.latticeVelocity, wallDistanceFunctor, ids.pdfField);
   // BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfFieldGPU, setup.fluidUID, ids.pdfField); 

   if(config->getNumBlocks("TPMS") > 0 && loggingParameters.getParameter< bool >("writeSurfaceMeshOfTPMS", false) ){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Writing TPMS as surface file")
      auto mesh = postprocessing::flagFieldToSurfaceMesh< FlagField_T >(blocks, ids.flagField, setup.obstacleFlag, true);
      WALBERLA_EXCLUSIVE_WORLD_SECTION(0){
         geometry::writeMesh("TPMS.obj", *mesh);
      }
   }

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                           COMMUNICATION SCHEME                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   const bool gpuEnabledMPI = parameters.getParameter< bool >("gpuEnabledMPI", false);
   auto nonUniformCommunication = std::make_shared< NonUniformGPUScheme< CommunicationStencil_T > >(blocks, gpuEnabledMPI);
   auto nonUniformPackInfo      = lbm_generated::setupNonuniformGPUPdfCommunication< GPUPdfField_T >(blocks, ids.pdfFieldGPU);
   nonUniformCommunication->addPackInfo(nonUniformPackInfo);

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                          TIME STEP DEFINITIONS                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   const int streamLowPriority = 0;
   auto defaultStream = gpu::StreamRAII::newPriorityStream(streamLowPriority);
   SweepTimeloop timeloop(blocks->getBlockStorage(), setup.timesteps);

   timeloop.addFuncBeforeTimeStep(inflow->updateGPUFunctor(), "Update inflow velocity");
   timeloop.addFuncBeforeTimeStep(probe->updateFunctor(), "Update Probe Data");

   lbm_generated::BasicRecursiveTimeStepGPU< GPUPdfField_T, SweepCollection_T, BoundaryCollection_T >
      LBMMeshRefinement(blocks, ids.pdfFieldGPU, sweepCollection, boundaryCollection, nonUniformCommunication, nonUniformPackInfo);

   // timeloop.addFuncBeforeTimeStep(LBMMeshRefinement, "Refinement timestep");
   LBMMeshRefinement.addRefinementToTimeLoop(timeloop, uint_c(0));

   shared_ptr< DragEvaluation > evaluation;
   if(config->getNumBlocks("RoughWall") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setup Drag Evaluation")
      evaluation = make_shared<DragEvaluation>(blocks, boundaryCollection, ids, setup, config);
      LBMMeshRefinement.addPostBoundaryHandlingBlockFunction(evaluation->forceCalculationFunctor());

   }


   const AABB sliceXY(finalDomain.xMin(), finalDomain.yMin(), finalDomain.center()[2] - setup.dxF,
                      finalDomain.xMax(), finalDomain.yMax(), finalDomain.center()[2] + setup.dxF);

/*                      
   CellInterval globalCellInterval;
   blocks->getCellBBFromAABB( globalCellInterval, sliceXY, uint_c(0) );

   std::map<IBlock*, CellInterval> sliceCellIntervals;

   for (auto& block : *blocks){
      CellInterval cellBB;
      blocks->getCellBBFromAABB( cellBB, sliceXY, blocks->getLevel(block) );
      blocks->transformGlobalToBlockLocalCellInterval( cellBB, block );

      Cell maxCell(cell_idx_c(0), cell_idx_c(0), cell_idx_c(0));
      Cell minCell(cell_idx_c( blocks->getNumberOfXCells( block ) ), cell_idx_c( blocks->getNumberOfYCells( block ) ), cell_idx_c( blocks->getNumberOfZCells( block ) ));

      for( cell_idx_t z = cell_idx_c(0); z != cell_idx_c( blocks->getNumberOfZCells( block ) ); ++z ){
         for( cell_idx_t y = cell_idx_c(0); y != cell_idx_c( blocks->getNumberOfYCells( block ) ); ++y ){
            for( cell_idx_t x = cell_idx_c(0); x != cell_idx_c( blocks->getNumberOfXCells( block ) ); ++x ){
               if( cellBB.contains(x,y,z) ){
                  if(Cell(x, y, z) < minCell){
                     minCell = Cell(x, y, z);
                  }
                  if(!(Cell(x, y, z) < maxCell)){
                     maxCell = Cell(x, y, z);
                  }
               }
            }
         }
      }
      CellInterval localCellInterval(minCell, maxCell);
      if(!localCellInterval.empty()){
         sliceCellIntervals[&block] = localCellInterval;
      }
   }

   timeloop.add() << BeforeFunction([&]()
   {
      for (auto const& sliceCellInterval : sliceCellIntervals){
         sweepCollection.calculateMacroscopicParametersCellInterval(sliceCellInterval.first, sliceCellInterval.second);
      }

      const uint_t velCtr = uint_c(welfordSweep.getCounter());
      welfordSweep.setCounter(welfordSweep.getCounter() + real_c(1.0));
   }) << Sweep(welfordSweep.getSweepOnCellInterval(blocks, globalCellInterval), "welford sweep");

*/

   /*SweepTimeloop timeloop(blocks->getBlockStorage(), setup.timesteps);
   timeloop.add() << BeforeFunction(communication.getCommunicateFunctor(), "communication")
                  << Sweep(boundaryCollection.getSweep(BoundaryCollection_T::ALL, defaultStream), "Boundary Conditions");
   timeloop.add() << Sweep(sweepCollection.streamCollide(SweepCollection_T::ALL, defaultStream), "LBM StreamCollide");*/

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                             VTK Output                                                     ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   // VTK
   auto VTKWriter = config->getOneBlock("VTKWriter");
   const uint_t vtkWriteFrequency       = VTKWriter.getParameter< uint_t >("vtkWriteFrequency", 0);
   const uint_t vtkGhostLayers          = VTKWriter.getParameter< uint_t >("vtkGhostLayers", 0);
   const bool amrFileFormat             = VTKWriter.getParameter< bool >("amrFileFormat", false);
   const bool oneFilePerProcess         = VTKWriter.getParameter< bool >("oneFilePerProcess", false);
   const real_t samplingResolution      = VTKWriter.getParameter< real_t >("samplingResolution", real_c(-1.0));
   const uint_t initialWriteCallsToSkip = VTKWriter.getParameter< uint_t >("initialWriteCallsToSkip", uint_c(0.0));
   const real_t sliceThickness          = VTKWriter.getParameter< real_t >("sliceThickness", real_c(1.0));

   if (vtkWriteFrequency > 0){
      const std::string vtkFolder = "VTKPorousMediaRE_" + std::to_string(uint64_c(setup.reynoldsNumber));
      auto vtkOutput =
         vtk::createVTKOutput_BlockData(*blocks, "vtk", vtkWriteFrequency, vtkGhostLayers, false, vtkFolder,
                                        "simulation_step", false, true, true, false, 0, amrFileFormat, oneFilePerProcess);

      vtkOutput->setInitialWriteCallsToSkip(initialWriteCallsToSkip);

      vtkOutput->addBeforeFunction([&]() {
         for (auto& block : *blocks){
            sweepCollection.calculateMacroscopicParameters(&block);
         }


         // for (auto const& sliceCellInterval : sliceCellIntervals){
         //    sweepCollection.calculateMacroscopicParametersCellInterval(sliceCellInterval.first, sliceCellInterval.second);
         // }

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
         gpu::fieldCpy< VelocityField_T, gpu::GPUField< VelocityField_T::value_type > >(blocks, ids.velocityField, ids.velocityFieldGPU);
         // gpu::fieldCpy< VelocityField_T, gpu::GPUField< VelocityField_T::value_type > >(blocks, ids.meanVelocityField, ids.meanVelocityFieldGPU);
         // gpu::fieldCpy< ScalarField_T, gpu::GPUField< ScalarField_T::value_type > >(blocks, ids.densityField, ids.densityFieldGPU);
#endif
      });

      vtkOutput->setSamplingResolution(samplingResolution);

      field::FlagFieldCellFilter<FlagField_T> fluidFilter( ids.flagField );
      fluidFilter.addFlag( setup.obstacleUID );
      vtkOutput->addCellExclusionFilter(fluidFilter);


      if (VTKWriter.getParameter< bool >("writeOnlySlice", true)){
         // const AABB sliceXY(finalDomain.xMin(), finalDomain.yMin(), finalDomain.center()[2] - sliceThickness * setup.dxF,
         //                    finalDomain.xMax(), finalDomain.yMax(), finalDomain.center()[2] + sliceThickness * setup.dxF);
         vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXY));

         if (VTKWriter.getParameter< bool >("writeXZSlice", false)){
            const AABB sliceXZ(finalDomain.xMin(), finalDomain.center()[1] - setup.dxF, finalDomain.zMin(),
                               finalDomain.xMax(), finalDomain.center()[1] + setup.dxF, finalDomain.zMax());
            vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXZ));
         }
      }

      if (VTKWriter.getParameter< bool >("velocity")){
         if (VTKWriter.getParameter< bool >("writeVelocityAsMagnitude")){
            auto velWriter = make_shared< field::VTKMagnitudeWriter< VelocityField_T, float32 > >(ids.velocityField, "velocityMagnitude");
            vtkOutput->addCellDataWriter(velWriter);
         }
         else{
            auto velWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(ids.velocityField, "velocity");
            vtkOutput->addCellDataWriter(velWriter);
         }
      }
      if (VTKWriter.getParameter< bool >("meanVelocity", false)){
         auto meanVelWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(ids.meanVelocityField, "meanVelocity");
         vtkOutput->addCellDataWriter(meanVelWriter);

      }
      if (VTKWriter.getParameter< bool >("density")){
         auto densityWriter = make_shared< field::VTKWriter< ScalarField_T, float32 > >(ids.densityField, "density");
         vtkOutput->addCellDataWriter(densityWriter);
      }
      if (VTKWriter.getParameter< bool >("flag")){
         auto flagWriter = make_shared< field::VTKWriter< FlagField_T > >(ids.flagField, "flag");
         vtkOutput->addCellDataWriter(flagWriter);
      }
      timeloop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");
   }

   // log remaining time
   if (uint_c(setup.remainingTimeLoggerFrequency) > 0){
      timeloop.addFuncAfterTimeStep(
         timing::RemainingTimeLogger(timeloop.getNrOfTimeSteps(), setup.remainingTimeLoggerFrequency),
         "remaining time logger");
   }

   auto CheckerParameters      = config->getOneBlock("StabilityChecker");
   const uint_t checkFrequency = CheckerParameters.getParameter< uint_t >("checkFrequency", uint_t(0));
   if (checkFrequency > 0){
      auto checkFunction = [](VelocityField_T::value_type value) {  return value < math::abs(VelocityField_T::value_type(10)); };
      timeloop.addFuncAfterTimeStep(
         makeSharedFunctor(field::makeStabilityChecker< VelocityField_T, FlagField_T >(
            config, blocks, ids.velocityField, ids.flagField, setup.fluidUID, checkFunction)),
         "Stability check");
   }

   WALBERLA_MPI_BARRIER()
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif

   const lbm_generated::PerformanceEvaluation< FlagField_T > performance(blocks, ids.flagField, setup.fluidUID);
   field::CellCounter< FlagField_T > fluidCells(blocks, ids.flagField, setup.fluidUID);
   fluidCells();

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Starting Simulation with " << setup.timesteps << " timesteps")
   WcTimingPool timeloopTiming;
   WcTimer simTimer;

   WALBERLA_MPI_BARRIER()
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif

   simTimer.start();
   timeloop.run(timeloopTiming);
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif
   WALBERLA_MPI_BARRIER()
   simTimer.end();
   WALBERLA_LOG_INFO_ON_ROOT("Simulation finished")
   double time = double_c(simTimer.max());
   WALBERLA_MPI_SECTION() { walberla::mpi::reduceInplace(time, walberla::mpi::MAX); }
   performance.logResultOnRoot(setup.timesteps, time);

   const auto reducedTimeloopTiming = timeloopTiming.getReduced();
   WALBERLA_LOG_RESULT_ON_ROOT("Time loop timing:\n" << *reducedTimeloopTiming)

   printResidentMemoryStatistics();
   return EXIT_SUCCESS;
}