//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file PorousMediaGPU.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================

#include "blockforest/Initialization.h"
#include "blockforest/SetupBlock.h"
#include "blockforest/SetupBlockForest.h"
#include "blockforest/loadbalancing/StaticCurve.h"
#include "blockforest/StructuredBlockForest.h"
#include "blockforest/communication/NonUniformBufferedScheme.h"

#include "core/Abort.h"
#include "core/DataTypes.h"
#include "core/SharedFunctor.h"
#include "core/debug/CheckFunctions.h"
#include "core/logging/Logging.h"
#include "core/math/Vector3.h"
#include "core/math/Constants.h"
#include "core/Environment.h"
#include "core/mpi/MPIManager.h"
#include "core/MemoryUsage.h"
#include "core/mpi/Reduce.h"
#include "core/timing/RemainingTimeLogger.h"
#include "core/timing/TimingPool.h"

#include "field/AddToStorage.h"
#include "field/CellCounter.h"
#include "field/FlagField.h"
#include "field/StabilityChecker.h"
#include "field/adaptors/AdaptorCreators.h"
#include "field/iterators/FieldIterator.h"
#include "field/vtk/VTKWriter.h"
#include "field/vtk/FlagFieldCellFilter.h"

#include "geometry/InitBoundaryHandling.h"
#include "geometry/mesh/TriangleMeshIO.h"

#include "lbm_generated/communication/NonuniformGeneratedPdfPackInfo.h"
#include "lbm_generated/evaluation/PerformanceEvaluation.h"
#include "lbm_generated/field/AddToStorage.h"
#include "lbm_generated/field/PdfField.h"
#include "lbm_generated/refinement/BasicRecursiveTimeStep.h"

#include "timeloop/SweepTimeloop.h"
#include "vtk/VTKOutput.h"

#include <cstdlib>
#include <iostream>
#include <memory>

#include "DebugStaticDefines.h"
#include "DebugInfoHeader.h"

using namespace walberla;

using StorageSpecification_T = lbm::DebugStorageSpecification;
using Stencil_T              = StorageSpecification_T::Stencil;
using CommunicationStencil_T = StorageSpecification_T::CommunicationStencil;

using PdfField_T           = lbm_generated::PdfField< StorageSpecification_T >;
using FlagField_T          = FlagField< uint8_t >;

using BoundaryCollection_T = lbm::DebugBoundaryCollection< FlagField_T >;
using SweepCollection_T = lbm::DebugSweepCollection;


using macroFieldType = VelocityField_T::value_type;
using RefinementSelectionFunction = std::function<void (SetupBlockForest &)>;

struct IDs
{
   BlockDataID pdfField;
   BlockDataID velocityField;
   BlockDataID meanVelocityField;
   BlockDataID velocityFieldInflowSlice;
   BlockDataID densityField;
   BlockDataID flagField;
};

void refineWalls(SetupBlockForest& forest, const real_t hemispheresRadius, const real_t dx, const uint_t level)
{
   auto domain = forest.getDomain();
   const real_t yMin = domain.yMin();
   const real_t yMax = domain.yMax();

   const real_t t = hemispheresRadius + dx;

   const AABB minAABB(domain.xMin(), domain.yMin(), domain.zMin(), domain.xMax(), domain.yMin() + t, domain.zMax());
   // const AABB maxAABB(domain.xMin(), domain.yMax() - t, domain.zMin(), domain.xMax(), domain.yMax(), domain.zMax());

   for (auto block = forest.begin(); block != forest.end(); ++block){
      const AABB& aabb = block->getAABB();

      // if (block->getLevel() < level && (minAABB.intersects(aabb) || maxAABB.intersects(aabb)) ){
      if (block->getLevel() < level && minAABB.intersects(aabb) ){
         block->setMarker(true);
      }
   }
}

void createSetupBlockForest(SetupBlockForest& setupBfs, std::shared_ptr<Config>& config, const RefinementSelectionFunction& refinementSelection, const bool useMPIManager=false)
{
   auto domainParameters  = config->getOneBlock("DomainSetup");

   WALBERLA_LOG_INFO_ON_ROOT("Generating SetupBlockForest...")

   uint_t numProcesses        = domainParameters.getParameter< uint_t >("numberProcesses");

   const std::string blockForestFilestem = "Debug";

   if(useMPIManager) {numProcesses = uint_c(mpi::MPIManager::instance()->numProcesses());}

   // TPMSRefinementSelection tpmsRefinementSelection( setup );
   setupBfs.addRefinementSelectionFunction(refinementSelection);

   // SphereBlockExclusion SphereBlockExclusion( Sphere );
   // setupBfs.addBlockExclusionFunction(SphereBlockExclusion);

   Vector3< real_t > domainSize = domainParameters.getParameter< Vector3< real_t > >("domainSize");
   Vector3< uint_t > rootBlocks          = domainParameters.getParameter< Vector3< uint_t > >("blocks");
   Vector3< bool >  periodic            = domainParameters.getParameter< Vector3< bool > >("periodic");
   uint_t refinementLevels    = domainParameters.getParameter< uint_t >("refinementLevels");

   const AABB domain(real_c(0.0), real_c(0.0), real_c(0.0), domainSize[0], domainSize[1], domainSize[2]);
   setupBfs.addWorkloadMemorySUIDAssignmentFunction(blockforest::uniformWorkloadAndMemoryAssignment);
   setupBfs.init(domain, rootBlocks[0], rootBlocks[1], rootBlocks[2],
                         periodic[0], periodic[1], periodic[2]);
   setupBfs.balanceLoad(blockforest::StaticLevelwiseCurveBalanceWeighted(), numProcesses);

   WALBERLA_LOG_INFO_ON_ROOT("===========================  BLOCK FOREST STATISTICS ============================");
   WALBERLA_LOG_INFO_ON_ROOT("Blocks created: " << setupBfs.getNumberOfBlocks())
   for (uint_t level = 0; level <= refinementLevels; level++){
      const uint_t numberOfBlocks = setupBfs.getNumberOfBlocks(level);
      WALBERLA_LOG_INFO_ON_ROOT("Level " << level << " Blocks: " << numberOfBlocks)
   }

   const real_t avgBlocksPerProc = real_c(setupBfs.getNumberOfBlocks()) / real_c(setupBfs.getNumberOfProcesses());
   WALBERLA_LOG_INFO_ON_ROOT("Average blocks per process: " << avgBlocksPerProc);

   Vector3< uint_t > c = domainParameters.getParameter< Vector3< uint_t > >("cellsPerBlock");

   const uint_t cellsPerBlock        = c[0] * c[1] * c[2];;
   const uint_t totalNumberCells     = setupBfs.getNumberOfBlocks() * cellsPerBlock;
   const real_t averageCellsPerGPU   = avgBlocksPerProc * real_c(cellsPerBlock);

   const uint_t PDFsPerCell            = StorageSpecification_T::inplace ? Stencil_T::Q : 2 * Stencil_T::Q;
   const uint_t valuesPerCell          = (PDFsPerCell + VelocityField_T::F_SIZE + ScalarField_T::F_SIZE);
   const uint_t sizePerValue           = sizeof(StorageSpecification_T::value_type);
   const double expectedMemory         = double_c(totalNumberCells * valuesPerCell * sizePerValue) * 1e-9;

   WALBERLA_LOG_INFO_ON_ROOT( "Total number of cells will be " << totalNumberCells << " fluid cells (in total on all levels)")
   WALBERLA_LOG_INFO_ON_ROOT( "Expected total memory demand will be " << expectedMemory << " GB")

   WALBERLA_LOG_INFO_ON_ROOT("=================================================================================")

   if(mpi::MPIManager::instance()->numProcesses() > 1)
      return;

   auto logging           = config->getOneBlock("Logging");
   bool writeSetupForestAndReturn    = logging.getParameter< bool >("writeSetupForestAndReturn", false);

   if(writeSetupForestAndReturn){
      std::ostringstream oss;
      oss << blockForestFilestem << ".bfs";
      setupBfs.saveToFile(oss.str().c_str());

      setupBfs.writeVTKOutput(blockForestFilestem);
   }
}

void createBlockForest(shared_ptr< BlockForest >& bfs, std::shared_ptr<Config>& config, const RefinementSelectionFunction& refinementSelection){
   if (mpi::MPIManager::instance()->numProcesses() > 1){
      // Load structured block forest from file
      std::ostringstream oss;
      oss << "debug" << ".bfs";
      const std::string setupBlockForestFilepath = oss.str();
      std::ifstream infile(setupBlockForestFilepath.c_str());
      if(!infile.good()){
         WALBERLA_LOG_WARNING_ON_ROOT("Blockforest was not created beforehand and thus needs to be created on the fly. For large simulation runs this can be a severe problem!")
         SetupBlockForest setupBfs;
         createSetupBlockForest(setupBfs, config, refinementSelection, true);
         bfs = std::make_shared< BlockForest >(uint_c(MPIManager::instance()->worldRank()), setupBfs);
      }
      else{
         bfs = std::make_shared< BlockForest >(uint_c(MPIManager::instance()->worldRank()),
                                               setupBlockForestFilepath.c_str(), false);
      }
   }
   else{
      SetupBlockForest setupBfs;
      createSetupBlockForest(setupBfs, config, refinementSelection);
      bfs = std::make_shared< BlockForest >(uint_c(MPIManager::instance()->worldRank()), setupBfs);
   }
}



int main(int argc, char** argv){
   Environment env( argc, argv );
   mpi::MPIManager::instance()->useWorldComm();

   auto config = env.config();

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                        SETUP AND CONFIGURATION                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto parameters        = config->getOneBlock("Parameters");
   auto loggingParameters = config->getOneBlock("Logging");
   auto boundariesConfig  = config->getBlock("Boundaries");
   auto domainParameters  = config->getOneBlock("DomainSetup");


   bool writeSetupForestAndReturn = loggingParameters.getParameter< bool >("writeSetupForestAndReturn", false);
   if (uint_c(MPIManager::instance()->numProcesses()) > 1) writeSetupForestAndReturn = false;
   shared_ptr< BlockForest > bfs;

   Vector3< real_t > domainSize          = domainParameters.getParameter< Vector3< real_t > >("domainSize");
   Vector3< uint_t > cellsPerBlock       = domainParameters.getParameter< Vector3< uint_t > >("cellsPerBlock");
   Vector3< uint_t > rootBlocks          = domainParameters.getParameter< Vector3< uint_t > >("blocks");
   uint_t refinementLevels    = domainParameters.getParameter< uint_t >("refinementLevels");
   uint_t timesteps = parameters.getParameter< uint_t >("timesteps");

   const real_t dx = domainSize[0] / real_c(rootBlocks[0] * cellsPerBlock[0]);
   WALBERLA_CHECK_FLOAT_EQUAL_EPSILON(dx, domainSize[1] / real_c(rootBlocks[1] * cellsPerBlock[1]), real_c(1e-12), "Resulting resolution is not equal in x and y direxction. This is not supported");
   WALBERLA_CHECK_FLOAT_EQUAL_EPSILON(dx, domainSize[2] / real_c(rootBlocks[2] * cellsPerBlock[2]), real_c(1e-12), "Resulting resolution is not equal in x and z direxction. This is not supported");

   auto dxF = dx / real_c( 1 << refinementLevels );
   auto refinementSelection = std::bind(refineWalls, std::placeholders::_1, real_c(1.0), dxF, refinementLevels);

   createBlockForest(bfs, config, refinementSelection);

   if (writeSetupForestAndReturn && mpi::MPIManager::instance()->numProcesses() == 1){
      WALBERLA_LOG_INFO_ON_ROOT("BlockForest has been created and writen to file. Returning program")
      logging::Logging::printFooterOnStream();
      return EXIT_SUCCESS;
   }

   auto blocks = std::make_shared< StructuredBlockForest >(bfs, cellsPerBlock[0], cellsPerBlock[1], cellsPerBlock[2]);
   blocks->createCellBoundingBoxes();
   auto finalDomain = blocks->getDomain();

   const StorageSpecification_T StorageSpec = StorageSpecification_T();
   // Creating fields
   uint_t step = 0;
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Allocating data")
   IDs ids;

   ids.pdfField  = lbm_generated::addPdfFieldToStorage(blocks, "pdfs", StorageSpec, uint_c(2), field::fzyx);
   ids.velocityField = field::addToStorage< VelocityField_T >(blocks, "vel", macroFieldType(0.0), field::fzyx, uint_c(2));
   ids.meanVelocityField = field::addToStorage< VelocityField_T >(blocks, "mean velocity", macroFieldType(0.0), field::fzyx, uint_c(2));
   ids.densityField  = field::addToStorage< ScalarField_T >(blocks, "density", macroFieldType(1.0), field::fzyx, uint_c(2));
   ids.flagField     = field::addFlagFieldToStorage< FlagField_T >(blocks, "Boundary Flag Field", uint_c(2));
 
   const real_t drivingForce = 1e-6;
   const real_t omega = 1.970443349753695;
   const real_t latticeVelocity = 0.05;

   const Cell innerOuterSplit = Cell(parameters.getParameter< Vector3< cell_idx_t > >("innerOuterSplit", Vector3< cell_idx_t >(1, 1, 1)));
   SweepCollection_T sweepCollection(blocks, ids.pdfField, ids.densityField, ids.velocityField, drivingForce, omega, innerOuterSplit);

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Initialising data")
   for (auto& block : *blocks){
      sweepCollection.initialise(&block, cell_idx_c(1));
   }

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                      LB SWEEPS AND BOUNDARY HANDLING                                       ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto fluidUID           = FlagUID("Fluid");
   auto obstacleUID        = FlagUID("Obstacle");

   // Boundaries
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setting boundary conditions")


   geometry::initBoundaryHandling< FlagField_T >(*blocks, ids.flagField, boundariesConfig);

   geometry::setNonBoundaryCellsToDomain< FlagField_T >(*blocks, ids.flagField, fluidUID, 2);
   BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfField, fluidUID, latticeVelocity);


   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                           COMMUNICATION SCHEME                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto communication = std::make_shared< NonUniformBufferedScheme< CommunicationStencil_T > >(blocks);
   auto packInfo      = lbm_generated::setupNonuniformPdfCommunication< PdfField_T >(blocks, ids.pdfField);
   communication->addPackInfo(packInfo);

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                          TIME STEP DEFINITIONS                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   SweepTimeloop timeloop(blocks->getBlockStorage(), timesteps);

   lbm_generated::BasicRecursiveTimeStep< PdfField_T, SweepCollection_T, BoundaryCollection_T > LBMMeshRefinement(
         blocks, ids.pdfField, sweepCollection, boundaryCollection, communication, packInfo);
/*
   for (auto& block : *blocks){
      const auto pdfs = block.getData< PdfField_T >(ids.pdfField);
      real_t level = real_c(blocks->getLevel(block));
      for( auto it = pdfs->beginWithGhostLayer(2); it != pdfs->end(); ++it ){
         for(uint_t i = 0; i < Stencil_T::Q; ++i){
            pdfs->get(it.x(), it.y(), it.z(), i) = real_c(Stencil_T::Q)*level + real_c(i);
         }
      }
   }

   return 0;

   for (auto& block : *blocks){
      const auto pdfs = block.getData< PdfField_T >(ids.pdfField);
      auto level = blocks->getLevel(block);

      if(level == 0){
         for( auto it = pdfs->beginGhostLayerOnly(stencil::Direction::S); it != pdfs->end(); ++it ){

            WALBERLA_LOG_DEVEL_VAR(it)
         }
      }

   }

   return 0;

// LBMMeshRefinement.addRefinementToTimeLoop(timeloop);
// timeloop.addFuncAfterTimeStep(LBMMeshRefinement, "test");
// timeloop.run();
// WALBERLA_LOG_INFO_ON_ROOT("test")

for(auto i = 0; i < 50; ++i){
   for (auto& block : *blocks){
      boundaryCollection(&block);
   }
}
*/
// for(auto i = 0; i < 50; ++i)
//    LBMMeshRefinement();

// return 0;

   // timeloop.addFuncBeforeTimeStep(LBMMeshRefinement, "Refinement timestep");
   LBMMeshRefinement.addRefinementToTimeLoop(timeloop);

   const AABB sliceXY(finalDomain.xMin(), finalDomain.yMin(), finalDomain.center()[2] - dxF,
                      finalDomain.xMax(), finalDomain.yMax(), finalDomain.center()[2] + dxF);


   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                             VTK Output                                                     ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   // VTtimeloop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");K
   auto VTKWriter = config->getOneBlock("VTKWriter");
   const uint_t vtkWriteFrequency       = VTKWriter.getParameter< uint_t >("vtkWriteFrequency", 0);
   const real_t sliceThickness          = VTKWriter.getParameter< real_t >("sliceThickness", real_c(1.0));

   if (vtkWriteFrequency > 0){
      const std::string vtkFolder = "debug";
      auto vtkOutput =
         vtk::createVTKOutput_BlockData(*blocks, "vtk", vtkWriteFrequency, 0, false, vtkFolder,
                                        "simulation_step", false, true, true, false, 0, false, false);


      vtkOutput->addBeforeFunction([&]() {
         for (auto& block : *blocks){
            sweepCollection.calculateMacroscopicParameters(&block);
         }
      });


      field::FlagFieldCellFilter<FlagField_T> fluidFilter( ids.flagField );
      fluidFilter.addFlag(obstacleUID );
      vtkOutput->addCellExclusionFilter(fluidFilter);


      if (VTKWriter.getParameter< bool >("writeOnlySlice", true)){
         // const AABB sliceXY(finalDomain.xMin(), finalDomain.yMin(), finalDomain.center()[2] - sliceThickness * setup.dxF,
         //                    finalDomain.xMax(), finalDomain.yMax(), finalDomain.center()[2] + sliceThickness * setup.dxF);
         vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXY));

         if (VTKWriter.getParameter< bool >("writeXZSlice", false)){
            const AABB sliceXZ(finalDomain.xMin(), finalDomain.center()[1] - dxF, finalDomain.zMin(),
                               finalDomain.xMax(), finalDomain.center()[1] + dxF, finalDomain.zMax());
            vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXZ));
         }
      }

      if (VTKWriter.getParameter< bool >("velocity")){
         if (VTKWriter.getParameter< bool >("writeVelocityAsMagnitude")){
            auto velWriter = make_shared< field::VTKMagnitudeWriter< VelocityField_T, float32 > >(ids.velocityField, "velocityMagnitude");
            vtkOutput->addCellDataWriter(velWriter);
         }
         else{
            auto velWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(ids.velocityField, "velocity");
            vtkOutput->addCellDataWriter(velWriter);
         }
      }
      if (VTKWriter.getParameter< bool >("density")){
         auto densityWriter = make_shared< field::VTKWriter< ScalarField_T, float32 > >(ids.densityField, "density");
         vtkOutput->addCellDataWriter(densityWriter);
      }
      if (VTKWriter.getParameter< bool >("flag")){
         auto flagWriter = make_shared< field::VTKWriter< FlagField_T > >(ids.flagField, "flag");
         vtkOutput->addCellDataWriter(flagWriter);
      }
      timeloop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");
   }

   // log remaining time
   timeloop.addFuncAfterTimeStep(
      timing::RemainingTimeLogger(timeloop.getNrOfTimeSteps(), 20),
      "remaining time logger");

   auto CheckerParameters      = config->getOneBlock("StabilityChecker");
   const uint_t checkFrequency = CheckerParameters.getParameter< uint_t >("checkFrequency", uint_t(0));
   if (checkFrequency > 0){
      auto checkFunction = [](VelocityField_T::value_type value) {  return value < math::abs(VelocityField_T::value_type(10)); };
      timeloop.addFuncAfterTimeStep(
         makeSharedFunctor(field::makeStabilityChecker< VelocityField_T, FlagField_T >(
            config, blocks, ids.velocityField, ids.flagField, fluidUID, checkFunction)),
         "Stability check");
   }

   WALBERLA_MPI_BARRIER()

   const lbm_generated::PerformanceEvaluation< FlagField_T > performance(blocks, ids.flagField, fluidUID);
   field::CellCounter< FlagField_T > fluidCells(blocks, ids.flagField, fluidUID);
   fluidCells();

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Starting Simulation with " << timesteps << " timesteps")
   WcTimingPool timeloopTiming;
   WcTimer simTimer;

   WALBERLA_MPI_BARRIER()

   simTimer.start();
   timeloop.run(timeloopTiming);
   WALBERLA_MPI_BARRIER()
   simTimer.end();
   WALBERLA_LOG_INFO_ON_ROOT("Simulation finished")
   double time = double_c(simTimer.max());
   WALBERLA_MPI_SECTION() { walberla::mpi::reduceInplace(time, walberla::mpi::MAX); }
   performance.logResultOnRoot(timesteps, time);

   const auto reducedTimeloopTiming = timeloopTiming.getReduced();
   WALBERLA_LOG_RESULT_ON_ROOT("Time loop timing:\n" << *reducedTimeloopTiming)

   printResidentMemoryStatistics();
   return EXIT_SUCCESS;
}