//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Evaluation.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#include "DragEvaluation.h"


namespace walberla
{

DragEvaluation::DragEvaluation(std::shared_ptr< StructuredBlockForest >& blocks, BoundaryCollection_T & boundaryCollection,
                               const IDs& ids, const Setup& setup, std::shared_ptr<Config>& config)
   : blocks_(blocks), boundaryCollection_(boundaryCollection), ids_(ids), setup_(setup)
{
   auto parameters     = config->getOneBlock("DragEvaluation");
   auto wallParameters = config->getOneBlock("RoughWall");

   // if(parameters.getParameter< uint_t >("evaluationFrequency") == uint_c(0))
   //    return;
   
   quantities_ = {{"timestep", 0}, {"Fx", 1}, {"Fy", 2}, {"Fz", 3}, {"CDReal", 4}, {"CLReal", 5}};

   WALBERLA_ASSERT_NOT_NULLPTR(blocks)
   maxLevel_ = blocks->getDepth();
   blocks->getBlocks(finestBlocks_, maxLevel_);

   checkFrequency_ = parameters.getParameter< uint_t >("evaluationFrequency");
   writeCounter_   = 1000 * checkFrequency_;
   dt_             = setup.dt / real_c( 1 << setup.refinementLevels );
   timesteps_      = setup.timesteps * ( 1 << setup.refinementLevels );
   totalExecutionCounter_ = 0;
   rampUpTime_ = parameters.getParameter< uint_t >("evaluationStart") * ( 1 << setup.refinementLevels );

   if(writeCounter_ == uint_c(0))
      writeCounter_++;

   const real_t radius   = wallParameters.getParameter< real_t >("hemispheresRadius");
   const real_t distance = wallParameters.getParameter< real_t >("hemispheresDistance");

   const auto outputFolder = setup_.outputFolder;
   const auto fileStem = setup_.fileStem;
   const auto fileName = outputFolder + "/" + fileStem + "Drag.h5"; 
   filesystem::path hdfFilePath( fileName.c_str() );

   WALBERLA_ROOT_SECTION()
   {
      if( filesystem::exists( hdfFilePath ) )
         std::remove( hdfFilePath.string().c_str() );

      extendSize[0] = 0;
      selectStart[0] = 0;
      H5::FileAccPropList fileAccProps;
      h5file_ = std::make_unique<H5::H5File>(hdfFilePath.c_str(), H5F_ACC_CREAT | H5F_ACC_RDWR, H5::FileCreatPropList::DEFAULT, fileAccProps);
      H5::Group dataGroup = h5file_->createGroup("Data");

      const hsize_t dsetSize[1]{ 1 };
      const hsize_t maxSize[1]{ H5S_UNLIMITED };
      const hsize_t chunkSize[1]{writeCounter_};

      H5::DataSpace arrayDataSpace(1, dsetSize, maxSize);
      H5::DSetCreatPropList props;
      props.setChunk(1, chunkSize);

      for (auto quantitie = quantities_.begin(); quantitie != quantities_.end(); quantitie++){
         dataGroup.createDataSet(quantitie->first, io::h5Float(), arrayDataSpace, props);
      }

   }
   WALBERLA_MPI_BARRIER()

   dragResults_.resize(quantities_.size() * writeCounter_);

   auto domain = blocks->getDomain();
   const real_t yMin = domain.yMin();
   const real_t yMax = domain.yMax();

   std::vector<Vector3<real_t>> hemisphereCenters;
   getAllHemisphereCenters(hemisphereCenters, radius, distance, domain);

   const double pi = double_c(3.14159265358979323846);
   surfaceAreaHemiSpheres_ = double_c(0.0);
   double totalCircleArea = double_c(0.0);

   for (auto center : hemisphereCenters){
      double surfaceAreaHemiSphere = double_c(2.0) * pi * double_c(radius * radius);
      double circleArea = pi * double_c(radius * radius);
      if(center[0] < (domain.xMin() + setup_.dxF) || center[0] > (domain.xMax() - setup_.dxF)){
         surfaceAreaHemiSphere *= double_c(0.5);
         circleArea *= double_c(0.5);
      }
      if(center[2] < (domain.zMin() + setup_.dxF) || center[2] > (domain.zMax() - setup_.dxF)){
         surfaceAreaHemiSphere *= double_c(0.5);
         circleArea *= double_c(0.5);
      }
      surfaceAreaHemiSpheres_ += surfaceAreaHemiSphere;
      totalCircleArea += circleArea;
   }

   surfaceAreaWall_ = double_c(2.0) * (domain.xMax() - domain.xMin()) * (domain.zMax() - domain.zMin());
   surfaceAreaWall_ -= totalCircleArea;
   WALBERLA_CHECK_GREATER_EQUAL(surfaceAreaWall_, real_c(0.0))

   meanVelocity_ = setup.latticeVelocity;
   totalSurface_ = surfaceAreaHemiSpheres_ + surfaceAreaWall_;

   WALBERLA_LOG_INFO_ON_ROOT(
      "Evaluation initialised:"
         "\n   + Total surface area of Hemispheres:     " << surfaceAreaHemiSpheres_
      << "\n   + Surface area of the wall:              " << surfaceAreaWall_
      << "\n   + Total Surface area:                    " << totalSurface_
   )

}

void DragEvaluation::forceCalculation(const uint_t level)
{
   if(level == maxLevel_){
      ++totalExecutionCounter_;
   }
   if (checkFrequency_ == uint_c(0) || rampUpTime_ > totalExecutionCounter_ || (totalExecutionCounter_ % checkFrequency_) != uint_c(0)){
      return;
   }
   
   if(level == maxLevel_){
      for (auto b : finestBlocks_){
         force_ += Vector3<double>(boundaryCollection_.ObstacleObject->getForce(b));
      }

      // Gather force on root rank and distrute result to all other processes
      mpi::reduceInplace(force_, mpi::SUM, 0, MPIManager::instance()->comm());
      mpi::broadcastObject(force_, 0, MPIManager::instance()->comm());

      WALBERLA_ROOT_SECTION(){
         const double t = double_c(totalExecutionCounter_ - 1) * dt_;
         const double meanU2 = double_c(meanVelocity_) * double_c(meanVelocity_);

         const double cDRealArea = (double_c(2.0) * force_[0]) / (meanU2 * double_c(totalSurface_));
         const double cLRealArea = (double_c(2.0) * force_[1]) / (meanU2 * double_c(totalSurface_));

         dragResults_[quantities_["timestep"]*writeCounter_ + executionCounter_] = float_c(t);
         dragResults_[quantities_["Fx"]*writeCounter_ + executionCounter_] = float_c(force_[0]);
         dragResults_[quantities_["Fy"]*writeCounter_ + executionCounter_] = float_c(force_[1]);
         dragResults_[quantities_["Fz"]*writeCounter_ + executionCounter_] = float_c(force_[2]);
         dragResults_[quantities_["CDReal"]*writeCounter_ + executionCounter_] = float_c(cDRealArea);
         dragResults_[quantities_["CLReal"]*writeCounter_ + executionCounter_] = float_c(cLRealArea);
      }

      sumFx_ += force_[0];

      force_[0] = double_c(0.0);
      force_[1] = double_c(0.0);
      force_[2] = double_c(0.0);

      ++meanQuantitiesCounter_;
      ++executionCounter_;
      if(executionCounter_ >= writeCounter_ || totalExecutionCounter_ == timesteps_){
         WALBERLA_ROOT_SECTION(){
            writeResults();
         }
      }
   }
}


void DragEvaluation::writeResults()
{
   extendSize[0] += executionCounter_;
   const hsize_t selectCount[1] { executionCounter_ };
   WALBERLA_ROOT_SECTION()
   {
      auto group = h5file_->openGroup("Data");
      for (auto quantitie = quantities_.begin(); quantitie != quantities_.end(); quantitie++){
         
         H5::DataSet dataset = group.openDataSet(quantitie->first);
         dataset.extend(extendSize);
         
         H5::DataSpace dataspace = dataset.getSpace();
         dataspace.selectHyperslab(H5S_SELECT_SET, selectCount, selectStart);
            
         H5::DataSpace memSpace(1, selectCount);
         dataset.write(&dragResults_[quantitie->second * writeCounter_], io::h5Float(), memSpace, dataspace);
      }
   }
   selectStart[0] += (executionCounter_);
   executionCounter_ = 0;
}

Plotter::Plotter(std::shared_ptr< StructuredBlockForest >& blocks, const Setup& setup, const IDs& ids)
   : blocks_(blocks), setup_(setup), ids_(ids)
{

   const auto outputFolder = setup_.outputFolder;
   const auto fileStem = setup_.fileStem;
   const auto fileName = outputFolder + "/" + fileStem + "MeanQuantities.h5"; 
   filesystem::path hdfFilePath( fileName.c_str() );

   WALBERLA_ROOT_SECTION()
   {
      if( filesystem::exists( hdfFilePath ) )
         std::remove( hdfFilePath.string().c_str() );

      H5::FileAccPropList fileAccProps;
      h5file_ = std::make_unique<H5::H5File>(hdfFilePath.c_str(), H5F_ACC_CREAT | H5F_ACC_RDWR, H5::FileCreatPropList::DEFAULT, fileAccProps);
      H5::Group dataGroup = h5file_->createGroup("Data");
   }
}

void Plotter::writeResult(const real_t frictionVelocity, const real_t welfordCounter){

   std::set<real_t> localYCoordinates;
   for (auto& block : *blocks_){
      const uint_t ny = blocks_->getNumberOfYCells(block);
      for(uint_t y = 0; y < ny; ++y){
         const Cell localCell(cell_idx_c(0), cell_idx_c(y), cell_idx_c(0));
         Vector3<real_t> globalCellCenter;
         blocks_->getBlockLocalCellCenter(block, localCell, globalCellCenter);
         localYCoordinates.insert(globalCellCenter[1]);
      }
   }

   std::vector<real_t> allYCoordinates(localYCoordinates.begin(), localYCoordinates.end());

   // gather all coordinates on the root rank and broadcast to all other ranks
   allYCoordinates = mpi::gatherv(allYCoordinates, 0, MPIManager::instance()->comm());
   mpi::broadcastObject(allYCoordinates, 0, MPIManager::instance()->comm());

   std::sort( allYCoordinates.begin(), allYCoordinates.end() );
   allYCoordinates.erase( std::unique( allYCoordinates.begin(), allYCoordinates.end() ), allYCoordinates.end() );

   std::map<real_t, uint_t> centerToIntex;
   for(uint_t idx = 0; idx < allYCoordinates.size(); ++idx){
      centerToIntex[allYCoordinates[idx]] = idx;
   }

   const real_t viscosity = setup_.kinViscosity;
   const real_t channelHalfHeight = setup_.referenceLength;

   WALBERLA_ROOT_SECTION(){
      addAttribute(viscosity, "viscosity");
      addAttribute(channelHalfHeight, "channelHalfHeight");
      addAttribute(frictionVelocity, "frictionVelocity");
      addAttribute(welfordCounter, "welfordCounter");
   }

   std::vector<real_t> yCellCenter(centerToIntex.size(), real_c(0.0));
   std::vector<real_t> yRel(centerToIntex.size(), real_c(0.0));
   std::vector<real_t> yPlus(centerToIntex.size(), real_c(0.0));
   std::vector<uint_t> cellCounterI(centerToIntex.size(), uint_c(0));
   std::vector<uint_t> cellCounterE(centerToIntex.size(), uint_c(0));

   std::vector<real_t> meanVelocityI(centerToIntex.size(), real_c(0.0));
   std::vector<real_t> meanVelocityE(centerToIntex.size(), real_c(0.0));
   std::vector<real_t> uPlusI(centerToIntex.size(), real_c(0.0));
   std::vector<real_t> uPlusE(centerToIntex.size(), real_c(0.0));
   std::vector<real_t> reynoldsStressVector(centerToIntex.size() * TensorField_T::F_SIZE, 0_r);

   for (auto const& it : centerToIntex){
      yCellCenter[it.second] = it.first;
      yRel[it.second] = real_c(it.second) / channelHalfHeight;
      yPlus[it.second] = real_c(it.second) * frictionVelocity / viscosity;
   }

   for( auto bIt = blocks_->begin(); bIt != blocks_->end(); ++bIt) {
      auto * meanVelocityField = bIt->template getData<VelocityField_T>(ids_.meanVelocityField);
      auto * sosField = bIt->template getData<TensorField_T>(ids_.sosField);
      auto * flagField = bIt->template getData<FlagField_T>(ids_.flagField);
      auto domainFlag = flagField->getFlag(setup_.fluidUID);
      const auto ci = meanVelocityField->xyzSize();
      for(auto cellIt = ci.begin(); cellIt != ci.end(); ++cellIt) {
         Vector3<real_t> globalCellCenter;
         blocks_->getBlockLocalCellCenter(*bIt, *cellIt, globalCellCenter);

         if (!centerToIntex.count(globalCellCenter[1])){
            WALBERLA_ABORT("y coordinate of center to index map was not contained")
         }

         const uint_t idx = centerToIntex[globalCellCenter[1]];
         cellCounterI[idx]++;

         if(isFlagSet(flagField->get(*cellIt), domainFlag)){
            cellCounterE[idx]++;
            meanVelocityE[idx] += meanVelocityField->get(*cellIt, 0);
            meanVelocityI[idx] += meanVelocityField->get(*cellIt, 0);

            for(uint_t i = 0; i < TensorField_T::F_SIZE; ++i) {
               reynoldsStressVector[i * centerToIntex.size() + idx] += sosField->get(*cellIt, i) / welfordCounter;
            }
         }
      }
   }

   mpi::reduceInplace(cellCounterI, mpi::SUM, 0, MPIManager::instance()->comm());
   mpi::reduceInplace(cellCounterE, mpi::SUM, 0, MPIManager::instance()->comm());
   mpi::reduceInplace(meanVelocityE, mpi::SUM, 0, MPIManager::instance()->comm());
   mpi::reduceInplace(meanVelocityI, mpi::SUM, 0, MPIManager::instance()->comm());
   mpi::reduceInplace(reynoldsStressVector, mpi::SUM, 0, MPIManager::instance()->comm());


   WALBERLA_ROOT_SECTION(){
      for(uint_t n = 0; n < centerToIntex.size(); ++n){
         meanVelocityI[n] /= real_c(cellCounterI[n]);
         meanVelocityE[n] /= real_c(cellCounterE[n]);

         for(uint_t i = 0; i < TensorField_T::F_SIZE; ++i) {
            reynoldsStressVector[i * centerToIntex.size() + n] /= real_c(cellCounterE[n]);
         }

         uPlusI[n] = meanVelocityI[n] / frictionVelocity;
         uPlusE[n] = meanVelocityE[n] / frictionVelocity;
      }

      writeData(yCellCenter.data(), centerToIntex.size(), "y");
      writeData(yRel.data(), centerToIntex.size(), "yRel");
      writeData(yPlus.data(), centerToIntex.size(), "yPlus");
      writeData(meanVelocityI.data(), centerToIntex.size(), "meanVelocityI");
      writeData(meanVelocityE.data(), centerToIntex.size(), "meanVelocityE");
      writeData(uPlusI.data(), centerToIntex.size(), "uPlusI");
      writeData(uPlusE.data(), centerToIntex.size(), "uPlusE");
      writeData(&reynoldsStressVector[0 * centerToIntex.size()], centerToIntex.size(), "uu_rms");
      writeData(&reynoldsStressVector[1 * centerToIntex.size()], centerToIntex.size(), "uv_rms");
      writeData(&reynoldsStressVector[2 * centerToIntex.size()], centerToIntex.size(), "uw_rms");
      writeData(&reynoldsStressVector[3 * centerToIntex.size()], centerToIntex.size(), "vu_rms");
      writeData(&reynoldsStressVector[4 * centerToIntex.size()], centerToIntex.size(), "vv_rms");
      writeData(&reynoldsStressVector[5 * centerToIntex.size()], centerToIntex.size(), "vw_rms");
      writeData(&reynoldsStressVector[6 * centerToIntex.size()], centerToIntex.size(), "wu_rms");
      writeData(&reynoldsStressVector[7 * centerToIntex.size()], centerToIntex.size(), "wv_rms");
      writeData(&reynoldsStressVector[8 * centerToIntex.size()], centerToIntex.size(), "ww_rms");

   }
}

void Plotter::writeData(const real_t* data, const uint_t size, const std::string& dataSetName){
   const hsize_t dsetSize[1]{ 1 };
   const hsize_t maxSize[1]{ size };
   const hsize_t chunkSize[1]{ size };

   H5::DataSpace arrayDataSpace(1, dsetSize, maxSize);
   H5::DSetCreatPropList props;
   props.setChunk(1, chunkSize);

   const hsize_t selectCount[1] { size };
   const hsize_t selectStart[1] { 0 };

   auto group = h5file_->openGroup("Data");
   H5::DataSet dataset = group.createDataSet(dataSetName, io::getH5ExportType<real_t>(), arrayDataSpace, props);
   dataset.extend( selectCount );

   H5::DataSpace dataspace = dataset.getSpace();
   dataspace.selectHyperslab(H5S_SELECT_SET, selectCount, selectStart);
      
   H5::DataSpace memSpace(1, selectCount);
   dataset.write(data, io::getH5ExportType<real_t>(), memSpace, dataspace);
}

void Plotter::addAttribute(const real_t value, const std::string& attributeName){
   hsize_t attributeShape[1]{ 1 };
   H5::DataSpace attributeShapeDspace(1, attributeShape, attributeShape);
   auto group = h5file_->openGroup("Data");
   H5::Attribute attr = group.createAttribute(attributeName, io::getH5ExportType<real_t>(), attributeShapeDspace);
   real_t valueArray[1]{value};
   attr.write(io::getH5ExportType<real_t>(), valueArray);
}

}