//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Evaluation.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
# pragma once

#include "core/config/Config.h"
#include "core/math/Constants.h"
#include "core/Filesystem.h"
#include "core/math/Sample.h"
#include "core/math/Vector3.h"
#include "core/mpi/Broadcast.h"
#include "core/mpi/Gatherv.h"
#include "core/mpi/Reduce.h"

#include "lbm_generated/field/PdfField.h"

#include "io/hdfWrapper.h"

#include "Types.h"
#include "Setup.h"
#include "util.h"

#include <deque>
#include <fstream>
#include <memory>
#include <string>
#include <utility>
#include <vector>

using namespace walberla;

namespace walberla
{
class DragEvaluation
{
 public:
   DragEvaluation(std::shared_ptr< StructuredBlockForest >& blocks, BoundaryCollection_T & boundaryCollection,
                  const IDs& ids, const Setup& setup, std::shared_ptr<Config>& config);

   void forceCalculation(const uint_t level); // for calculating the force
   void writeResults();

   std::function<void (const uint_t)> forceCalculationFunctor()
   {
      return [this](uint_t level) { forceCalculation(level); };
   }

   double getMeanFrictionVelocity(){
    if(meanQuantitiesCounter_ == 0){
      return double_c(0.0);
    }
    return (sumFx_ / double_c(meanQuantitiesCounter_)) / totalSurface_;
   }

  void setTotalExecutionCounter(const uint_t counter){
    totalExecutionCounter_ = counter * ( 1 << setup_.refinementLevels );
   }

 protected:
   bool initialized_{false};

   std::shared_ptr< StructuredBlockForest > blocks_;
   BoundaryCollection_T & boundaryCollection_;
   IDs ids_;
   Setup setup_;

   std::map<std::string, uint_t> quantities_;
   std::vector<Block *> finestBlocks_;
   uint_t maxLevel_;
   real_t dt_;
   uint_t timesteps_;

   uint_t checkFrequency_;
   uint_t writeCounter_;
   uint_t executionCounter_{0};
   uint_t totalExecutionCounter_{0};
   uint_t rampUpTime_{0};

   uint_t meanQuantitiesCounter_{0};

   std::unique_ptr<H5::H5File> h5file_;
   hsize_t extendSize[1];
   hsize_t selectStart[1];

   std::vector<float> dragResults_;

   double surfaceAreaHemiSpheres_;
   double surfaceAreaWall_;
   double totalSurface_;
   double meanVelocity_;

   Vector3< double > force_;

   double sumFx_{0.0};

}; // class DragEvaluation

class Plotter
{
 public:
   Plotter(std::shared_ptr< StructuredBlockForest >& blocks, const Setup& setup, const IDs& ids);

   // void forceCalculation(const uint_t level); // for calculating the force
   // void writeResults();

   void writeResult(const real_t frictionVelocity, const real_t welfordCounter);
   void writeData(const real_t* data, const uint_t size, const std::string& dataSetName);
   void addAttribute(const real_t value, const std::string& attributeName);

 protected:
   bool initialized_{false};

   std::shared_ptr< StructuredBlockForest > blocks_;
   Setup setup_;
   IDs ids_;

   std::unique_ptr<H5::H5File> h5file_;

  uint_t executionCounter_{0};
  uint_t totalExecutionCounter_{0};

  uint_t rampUpTime_{100000};
  uint_t checkFrequency_{10000};
}; // class Plotter


class PressureDropEvaluator
{
 public:
   PressureDropEvaluator(const shared_ptr< StructuredBlockForest >& blocks, const Setup& setup, const IDs& ids, const Config::BlockHandle& pressureParameters)
                         : blocks_(blocks), setup_(setup), ids_(ids)
      {

        evalAABB_     = pressureParameters.getParameter< AABB >("evaluationAABB");
        evalInterval_ = pressureParameters.getParameter< uint_t >("evaluationFrequency");

        pressureDrop_ = real_c(0.0);
        executionCounter_ = uint_c(0);

        fileName_ = setup_.outputFolder + "/" + setup_.fileStem + "PressureDrop.csv"; 
        filesystem::path pressureDropOuputFile( fileName_.c_str() );

        WALBERLA_ROOT_SECTION()
        {
            if(!setup_.loadSnapshot){
                if( filesystem::exists( pressureDropOuputFile ) )
                    std::remove( pressureDropOuputFile.string().c_str() );
            }

            if(!filesystem::exists( pressureDropOuputFile ) ){
                std::ofstream outfile( pressureDropOuputFile.c_str() );
                outfile << "SEP=," << "\n";

                outfile << "timestep," << "time," << "pressureDrop," << "x0," << "x1" << "\n";
                outfile.close();
            }
        }
      }

  std::function<void()>  updateFunctor() {return [this](){ updatePressureDrop(); };}
  void setInitialTimestep(const uint_t initialTimestep){
      executionCounter_ = initialTimestep;
  }



 private:
   const shared_ptr< StructuredBlockForest > blocks_;
   Setup setup_;
   IDs ids_;
   AABB evalAABB_;
   uint_t evalInterval_;
   real_t pressureDrop_;

   uint_t executionCounter_; // number of times operator() has been called
   std::string fileName_;

   void syncDensity(){
      gpu::fieldCpy< ScalarField_T , gpu::GPUField< ScalarField_T::value_type > >(blocks_, ids_.densityField, ids_.densityFieldGPU);
   }

   void updatePressureDrop(){

      ++executionCounter_;
      if (executionCounter_ % evalInterval_ != uint_c(0)) { return; }

      syncDensity();

      real_t lowerDensity       = real_c(0);
      real_t upperDensity       = real_c(0);
      uint_t numCellsLowerSlice = uint_c(0);
      uint_t numCellsUpperSlice = uint_c(0);

      auto domainFlag = setup_.fluidFlag;

      for (auto blockIterator = blocks_->begin(); blockIterator != blocks_->end(); ++blockIterator)
      {
         const uint_t level = blocks_->getLevel(*blockIterator);

         // global lower cell BB slice for density measurement (shifted 1 cell away from porous medium in
         // z-direction)
         CellInterval lowerCellBB = blocks_->getCellBBFromAABB(evalAABB_, level);
         lowerCellBB.xMax()       = lowerCellBB.xMin() + cell_idx_c(1);
         // lowerCellBB.xMin()       = lowerCellBB.xMin();
         numCellsLowerSlice       = lowerCellBB.numCells();

         // global upper cell BB slice for density measurement (shifted 1 cell away from porous medium in
         // z-direction)
         CellInterval upperCellBB = blocks_->getCellBBFromAABB(evalAABB_, level);
         upperCellBB.xMin()       = upperCellBB.xMax() - cell_idx_c(1);
         // upperCellBB.xMax()       = upperCellBB.xMin();
         numCellsUpperSlice       = upperCellBB.numCells();

         // block's cell BB that intersects the global slices (still in global coordinates)
         CellInterval lowerBlockCellBB = blocks_->getBlockCellBB(*blockIterator);
         lowerBlockCellBB.intersect(lowerCellBB);
         CellInterval upperBlockCellBB = blocks_->getBlockCellBB(*blockIterator);
         upperBlockCellBB.intersect(upperCellBB);

         // transform the global coordinates of relevant cells to block local coordinates
         CellInterval lowerBlockLocalCellBB;
         CellInterval upperBlockLocalCellBB;
         blocks_->transformGlobalToBlockLocalCellInterval(lowerBlockLocalCellBB, *blockIterator, lowerBlockCellBB);
         blocks_->transformGlobalToBlockLocalCellInterval(upperBlockLocalCellBB, *blockIterator, upperBlockCellBB);

         auto densityField = blockIterator->template getData< ScalarField_T >(ids_.densityField);
         auto flagField = blockIterator->template getData< FlagField_T >(ids_.flagField);

         // sum density of relevant cells
         // clang-format off
         WALBERLA_FOR_ALL_CELLS_IN_INTERVAL_XYZ(lowerBlockLocalCellBB,
                                                if (isFlagSet(flagField->get(x, y, z), domainFlag))
                                                {
                                                   lowerDensity += densityField->get(x, y, z);
                                                }
         ) // WALBERLA_FOR_ALL_CELLS_IN_INTERVAL_XYZ

         WALBERLA_FOR_ALL_CELLS_IN_INTERVAL_XYZ(upperBlockLocalCellBB,
                                                if (isFlagSet(flagField->get(x, y, z), domainFlag))
                                                {
                                                   upperDensity += densityField->get(x, y, z);
                                                }
         ) // WALBERLA_FOR_ALL_CELLS_IN_INTERVAL_XYZ
         // clang-format on
      }

      mpi::allReduceInplace< real_t >(lowerDensity, mpi::SUM);
      mpi::allReduceInplace< real_t >(upperDensity, mpi::SUM);

      // average density
      lowerDensity /= real_c(numCellsLowerSlice);
      upperDensity /= real_c(numCellsUpperSlice);

      pressureDrop_ = (upperDensity - lowerDensity) / real_c(3.0);
      const real_t timestep = real_c(executionCounter_) * setup_.dt;


      WALBERLA_ROOT_SECTION(){
          std::ofstream outfile( fileName_.c_str(), std::ios_base::app );
          outfile << std::setprecision(16);
          outfile << executionCounter_ << "," << timestep << "," << pressureDrop_ << "," << evalAABB_.xMin() << "," << evalAABB_.xMax() << "\n";
          outfile.close();
      }

   }
}; // class PressureDropEvaluator

}