//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Inflow.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#include "Inflow.h"

namespace walberla{

void Inflow::logging(){
    WALBERLA_LOG_INFO_ON_ROOT( "Using turbulent injection with:" << std::setprecision(16) <<
      "\n   + kin. viscosity:        " << nu_ << " [m^2/s]" <<
      "\n   + uRMS:                  " << urms_ <<
      "\n   + characteristicLength:  " << le_ <<
      "\n   + numberOfFourrierModes: " << n_ <<
      "\n   + alpha:                 " << alpha_ <<
      "\n   + kappa VKP:             " << kappaVKP_ <<
      "\n   + epsilon VKP:           " << epsilonVKP_ <<
      "\n   + kappa Kolmogorov:      " << kappaKolmogorov_ <<
      "\n   + firstMode:             " << firstMode_ <<
      "\n   + lastMode:              " << lastMode_)
}

void Inflow::updateCPU(){
    if(!setup_.turbulentInflow){
        return;
    }

    const real_t t = real_c(executionCounter_) * dt_;
    for (auto& block : *blocks_){
        auto velocityFieldInflowSlice = block.template getData< Field_T >(ids_.velocityFieldInflowSlice);
        const cell_idx_t zSize = cell_idx_c( velocityFieldInflowSlice->zSize() );
        const cell_idx_t ySize = cell_idx_c( velocityFieldInflowSlice->ySize() );
        const cell_idx_t gl = cell_idx_c( velocityFieldInflowSlice->nrOfGhostLayers() );

        WALBERLA_ASSERT_EQUAL(blocks_->getLevel( block ), uint_c(0), "Inflow must be located on coarsest resolution")

        for( cell_idx_t z = 0; z < zSize; ++z ) { 
            for( cell_idx_t y = 0; y < ySize; ++y ){
                auto minCorner = block.getAABB().minCorner();

                const real_t zCenter = (globalMinCorner_[2] + ( real_c( z ) + real_c(0.5) ) * dx_) + minCorner[2];
                const real_t yCenter = (globalMinCorner_[1] + ( real_c( y ) + real_c(0.5) ) * dx_) + minCorner[1];

                Vector3<real_t> p(real_c(0.0), yCenter, zCenter);

                const Vector3<real_t> v = summOfFourrierModes(p, uc_, kappa_, ut_, t);
                // velocityFieldInflowSlice->get(0, y, z, 0) = latticeVelocity_ + v[0];
                velocityFieldInflowSlice->get(0, y, z, 0) = v[0];
                velocityFieldInflowSlice->get(0, y, z, 1) = v[1];
                velocityFieldInflowSlice->get(0, y, z, 2) = v[2];
            } 
        }
    }
    gpu::fieldCpy< GPUField_T, Field_T >(blocks_, ids_.velocityFieldInflowSliceGPU, ids_.velocityFieldInflowSlice);
    ++executionCounter_;
}

void Inflow::updateGPU(){
    if(!setup_.turbulentInflow){
        return;
    }

    const real_t t = real_c(executionCounter_) * dt_;
    for (auto& block : *blocks_){
        
        auto blockMinCorner = block.getAABB().min();

        auto velocityGPU = block.template getData< GPUField_T >(ids_.velocityFieldInflowSliceGPU);
        updateVelocityPlane(velocityGPU, kappaGPU_, utGPU_, globalMinCorner_, blockMinCorner, uc_, 
                            dx_, t, n_, latticeVelocity_, nGhostLayers_, block_, grid_);

    }
    ++executionCounter_;
}

void Inflow::init(){
    nu_    = setup_.kinViscosity;
    urms_  = setup_.urms;
    le_    = setup_.characteristicLength;
    n_     = blockDimX_ * (1 + uint_c(std::floor(setup_.numberOfFourrierModes / blockDimX_)));
    dt_    = setup_.dt;
    dx_    = setup_.dxC;
    uc_[0] = setup_.uc[0];
    uc_[1] = setup_.uc[1];
    uc_[2] = setup_.uc[2];

    latticeVelocity_ = setup_.latticeVelocity;

    alpha_           = (real_c(55.0) * std::tgamma(real_c(5.0) / real_c(6.0))) / (real_c(9.0) * std::sqrt(math::pi) * std::tgamma(real_c(1.0) / real_c(3.0)));
    kappaVKP_        = (real_c(9.0) * math::pi * alpha_) / (real_c(55.0) * le_);
    // epsilonVKP_      = (std::pow(alpha_ * std::tgamma(real_c(2.0)/real_c(3.0)), (real_c(3.0)/real_c(2.0)))) * (real_c(0.5) * kappaVKP_ * std::pow(urms_, real_c(3.0)));
    epsilonVKP_      = std::pow(urms_, real_c(3.0)) / le_;
    kappaKolmogorov_ = std::pow(epsilonVKP_ / std::pow(nu_, real_c(3.0)), real_c(1.0)/real_c(4.0));

    firstMode_ = std::log10(kappaVKP_ / real_c(10.0));
    lastMode_  = std::log10(kappaKolmogorov_ * real_c(10.0));
    const real_t linDeltaKappa = (lastMode_ - firstMode_) / (real_c(n_) - real_c(1.0));

    real_t utSquaredSum = 0.0;
    for(uint_t i = 0; i < n_; ++i){
        const real_t kappaN  = std::pow(real_c(10.0), firstMode_ + (real_c(i) * linDeltaKappa));
        const real_t kappaN1 = std::pow(real_c(10.0), firstMode_ + (real_c(i + 1) * linDeltaKappa));
        const real_t deltaKappa = kappaN1 - kappaN;
        const real_t tmp = std::sqrt(getKarmanPaoEnergy(kappaN, alpha_, kappaVKP_, kappaKolmogorov_, urms_) * deltaKappa);

        kappa_.emplace_back(kappaN);
        ut_.emplace_back(tmp);
        utSquaredSum += tmp*tmp;
    }

    n_ = uint_c(kappa_.size());
    const real_t correctionTKE = std::sqrt(((real_c(3.0) / real_c(2.0)) * (urms_*urms_)) / utSquaredSum);

    for(uint_t i = 0; i < n_; ++i){
        ut_[i] *= correctionTKE;
    }

    globalMinCorner_ = blocks_->getDomain().min();

    const Vector3<uint_t> s(blockDimX_, setup_.cellsPerBlock[1] + 2 * nGhostLayers_, setup_.cellsPerBlock[2] + 2 * nGhostLayers_);
    getBlockAndGridSize(s, block_, grid_);

    WALBERLA_GPU_CHECK(gpuMalloc((void **)&kappaGPU_, sizeof(real_t) * kappa_.size() ));
    WALBERLA_GPU_CHECK(gpuMemcpy( kappaGPU_, &kappa_[0], sizeof(real_t) * kappa_.size(), gpuMemcpyHostToDevice ));

    WALBERLA_GPU_CHECK(gpuMalloc((void **)&utGPU_, sizeof(real_t) * ut_.size() ));
    WALBERLA_GPU_CHECK(gpuMemcpy( utGPU_, &ut_[0], sizeof(real_t) * ut_.size(), gpuMemcpyHostToDevice ));
}

} // namespace walberla
