//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Inflow.h
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#pragma once

#include "core/logging/Initialization.h"
#include "core/math/Constants.h"

#include "gpu/GPUWrapper.h"
#include "gpu/ErrorChecking.h"
#include "gpu/FieldCopy.h"

#include "Setup.h"
#include "util.h"
#include "utilGPU.h"

namespace walberla{

class Inflow
{
 public:
   Inflow(const std::shared_ptr< StructuredBlockForest >& blocks, const Setup& setup, const IDs& ids)
      : blocks_(blocks), setup_(setup), ids_(ids), executionCounter_(uint_c(0))
   {
       init();
       logging();
   }

   using Field_T = walberla::field::GhostLayerField<real_t, 3>;
   using GPUField_T = gpu::GPUField< real_t >;

   ~Inflow(){
    WALBERLA_GPU_CHECK(gpuFree(kappaGPU_))
    WALBERLA_GPU_CHECK(gpuFree(utGPU_))
   }

   void init();
   void logging();

   std::function<void()>  updateCPUFunctor() {return [this](){ Inflow::updateCPU();};}
   std::function<void()>  updateGPUFunctor() {return [this](){ Inflow::updateGPU();};}

   void updateCPU();
   void updateGPU();

 private:
   const std::shared_ptr< StructuredBlockForest > blocks_;
   const Setup setup_;
   const IDs ids_;

   real_t nu_;
   real_t urms_;
   real_t le_;
   uint_t n_;
   real_t alpha_;
   real_t kappaVKP_;
   real_t epsilonVKP_;
   real_t kappaKolmogorov_;

   real_t dt_;
   real_t dx_;
   Vector3<real_t> uc_;
   real_t latticeVelocity_;
   const int64_t nGhostLayers_{1};
   const uint_t blockDimX_{128};

   real_t firstMode_;
   real_t lastMode_;
   std::vector<real_t> kappa_;
   std::vector<real_t> ut_;

   uint_t executionCounter_; // number of times operator() has been called

   Vector3<real_t> globalMinCorner_;

   Vector3<uint32_t> block_;
   Vector3<uint32_t> grid_;
   real_t *kappaGPU_;
   real_t *utGPU_;

}; // class Inflow
} // namespace walberla
