//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file InitializerFunctions.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#include "InitializerFunctions.h"


namespace walberla
{
using FlagField_T = FlagField< uint8_t >;

void initGrid(const shared_ptr< StructuredBlockStorage >& blocks, const BlockDataID flagFieldID, const Setup& setup){

   int f = 1;
   for(uint_t r = 48; r <= 49; r+= 2){
      const real_t xLocation = 1.0;
      const real_t thickness = 1.0;
      const uint_t length = 16;

      const uint_t ySize = blocks->getNumberOfYCells();
      const uint_t zSize = blocks->getNumberOfZCells();

      const uint_t start = r; //uint_c(std::ceil((xLocation / setup.dxC) - thickness + real_c(0.5)));
      const uint_t end   = r + 1; //uint_c(std::floor((xLocation / setup.dxC) + thickness - real_c(0.5)));
      const uint_t stepSize = length*2;

      f = f == 1 ? 0:1;

      for(uint_t z = f; z < zSize; z+=stepSize){
         for(uint_t y = f; y < ySize; y+=stepSize){
            for(uint_t x = start; x <= end; ++x){
               for(uint_t z1 = z; z1 < z + length; ++z1){
                  for(uint_t y1 = y; y1 < y + length; ++y1){

                     Cell cell(x, y1, z1);
                     auto block = blocks->getBlock( cell );

                     if(block != nullptr){
                        const auto flagField    = block->template getData< FlagField_T >(flagFieldID);
                        const auto boundaryFlag = flagField->getOrRegisterFlag(setup.obstacleUID);
                        
                        blocks->transformGlobalToBlockLocalCell(cell, *block);
                        addFlag(flagField->get(cell), boundaryFlag);
                     }
                  }
               }
            }
         }
      }
   }
}

void initHemisphers(const shared_ptr< StructuredBlockStorage >& blocks, const BlockDataID flagFieldID, const Setup& setup,
                    const Config::BlockHandle& wallParameters){

   const real_t radius   = wallParameters.getParameter< real_t >("hemispheresRadius");
   const real_t hemispheresDistance = wallParameters.getParameter< real_t >("hemispheresDistance");

   auto domain = blocks->getDomain();
   const real_t yMin = domain.yMin();
   const real_t yMax = domain.yMax();

   for (auto bIt = blocks->begin(); bIt != blocks->end(); ++bIt){
      Block& b             = dynamic_cast< Block& >(*bIt);
      const auto flagField    = b.getData< FlagField_T >(flagFieldID);
      const auto boundaryFlag = flagField->getOrRegisterFlag(setup.obstacleUID);

      for( auto it = flagField->beginWithGhostLayer(1); it != flagField->end(); ++it ){
         Vector3< real_t > cellCenter = blocks->getBlockLocalCellCenter( b, it.cell() );
         blocks->mapToPeriodicDomain(cellCenter);
         if(cellCenter[1] > radius && cellCenter[1] < yMax - radius){
            continue;
         }
         auto center = getCenterOfHemisphere(cellCenter, hemispheresDistance, yMin, yMax);
         if(hemispheresContains(cellCenter, center, radius)) {
            addFlag(flagField->get(it.x(), it.y(), it.z()), boundaryFlag);
         }
      }
   }
}

void initGenericPorousMatrix(const shared_ptr< StructuredBlockStorage >& blocks, const BlockDataID flagFieldID, const Setup& setup,
                             const Config::BlockHandle& gpmParameters){

   const real_t poreSize     = gpmParameters.getParameter< real_t >("poreSize");
   const real_t barSize      = gpmParameters.getParameter< real_t >("barSize");
   const real_t PoreSize2    = poreSize * real_c(2.0);
   const real_t halfbarSize  = barSize * real_c(0.5);

   auto domain = blocks->getDomain();
   const real_t zMin = domain.zMin();
   const real_t zMax = domain.zMax();

   std::vector<AABB> bars;

   const uint_t xn = uint_c(std::ceil(domain.xMax() / poreSize));
   const uint_t yn = uint_c(std::ceil(domain.xMax() / PoreSize2));

   for(uint_t i = 0; i <= xn; ++i){
      for(uint_t j = 0; j <= yn; ++j){
         const real_t xc = real_c(i) * poreSize + barSize;
         const real_t yc = i % 2 != 0 ? real_c(j) * PoreSize2 + barSize : real_c(j) * PoreSize2 + barSize + poreSize;

         const AABB barAABB(xc - halfbarSize, yc - halfbarSize, zMin, xc + halfbarSize, yc + halfbarSize, zMax);
         bars.emplace_back(barAABB);
      }
   }

   for (auto bIt = blocks->begin(); bIt != blocks->end(); ++bIt){
      Block& b             = dynamic_cast< Block& >(*bIt);
      const auto flagField    = b.getData< FlagField_T >(flagFieldID);
      const auto boundaryFlag = flagField->getOrRegisterFlag(setup.noSlipUID);

      for( auto it = flagField->beginWithGhostLayer(0); it != flagField->end(); ++it ){
         Vector3< real_t > cellCenter = blocks->getBlockLocalCellCenter( b, it.cell() );
         blocks->mapToPeriodicDomain(cellCenter);

         for(const auto& barAABB: bars){
            if(barAABB.contains(cellCenter)) {
               addFlag(flagField->get(it.x(), it.y(), it.z()), boundaryFlag);
            }
         }

         // WALBERLA_LOG_DEVEL_VAR(cellCenter[0] / poreSize)
         // WALBERLA_LOG_DEVEL_VAR(std::round( (cellCenter[0]) / poreSize))

         // const real_t xc = xMap[std::round( (cellCenter[0]) / (poreSize + barSize))];

         // const real_t xc = std::round( (cellCenter[0]) / poreSize) * poreSize;
         //const real_t ys = real_c((uint_c(std::round(cellCenter[0] / poreSize)) % 2)) * poreSize;
         //const real_t yc = (std::round((cellCenter[1] - ys) / PoreSize2) * (PoreSize2)) + ys;

         //const AABB barAABB(xc - halfbarSize, yc - halfbarSize, zMin, xc + halfbarSize, yc + halfbarSize, zMax);
         //if(barAABB.contains(cellCenter)) {
         //   addFlag(flagField->get(it.x(), it.y(), it.z()), boundaryFlag);
        // }
         
      }
   }
}

void setVelocityFieldsAsmuth(const shared_ptr< StructuredBlockStorage >& blocks,
                             const Setup& setup, const IDs& ids) {

   const real_t B = real_c(5.5);
   const real_t kappa = real_c(0.41);

   auto domainSize = blocks->getDomain().max();
   const real_t delta = domainSize[1] / real_c(2.0);

   domainSize /= setup.dxC;

   const real_t frictionVelocity = setup.frictionVelocity;
   const real_t dx = setup.dxC;
   const real_t viscosity = setup.viscosity;

   for( auto block = blocks->begin(); block != blocks->end(); ++block ) {
      const uint_t level = blocks->getLevel(*block.get());
      auto * velocityField = block->template getData<VelocityField_T>(ids.velocityField);

      const auto ci = velocityField->xyzSizeWithGhostLayer();
      for(auto cellIt = ci.begin(); cellIt != ci.end(); ++cellIt) {

         Cell globalCell(*cellIt);
         blocks->transformBlockLocalToGlobalCell(globalCell, *block);
         Vector3<real_t> cellCenter;
         blocks->getCellCenter(cellCenter, globalCell, level);

         const auto y = cellCenter[1];
         const auto rel_x = cellCenter[0] / domainSize[0];
         // const auto rel_y = cellCenter[1] / domainSize[1];
         const auto rel_z = cellCenter[2] / domainSize[2];

         const real_t pos = std::max(delta - std::abs(y - delta - 0_r), 0.05_r);
         const auto rel_y = pos / delta;

         auto initialVel = frictionVelocity * (std::log(frictionVelocity * pos / viscosity) / kappa + B);

         Vector3<real_t> vel;
         vel[0] = initialVel;

         vel[2] = 2_r * frictionVelocity / kappa * std::sin(math::pi * 16_r/dx * rel_x) *
                        std::sin(math::pi * 8_r  * rel_y) / (std::pow(rel_y, 2_r) + 1_r);

         vel[1] = 8_r * frictionVelocity / kappa *
                           (std::sin(math::pi * 8_r/dx * rel_z) * std::sin(math::pi * 8_r * rel_y) +
                           std::sin(math::pi * 8_r/dx * rel_x)) / (std::pow(0.5_r * delta - pos, 2_r) + 1_r);


/*
         const real_t pos = std::max(delta - std::abs(y - delta), real_c(2.0));
         const real_t t1 = std::sin(math::pi * 8.0 / dx * rel_z) * std::sin(math::pi * 8.0 / dx * rel_y) + std::sin(math::pi * 8.0 / dx * rel_x);
         const real_t t2 = std::pow(1.0_r * delta - pos, 2_r) + 1_r;
         const real_t t3 = std::sin(math::pi * 16.0_r / dx * rel_x) * std::sin(math::pi * 8.0_r / dx * rel_y);

         Vector3<real_t> vel;
         
         const real_t intVelX = frictionVelocity * (std::log(frictionVelocity * pos / viscosity) / kappa + B);
         vel[0] = setup.latticeVelocity; // intVelX > setup.latticeVelocity ? setup.latticeVelocity : intVelX;
         vel[1] = 1.0_r * frictionVelocity / kappa * t1 / t2;
         vel[2] = 2.0_r * frictionVelocity / kappa * t3 / (std::pow(rel_y, 2.0_r) + 1.0_r);
         
*/
         velocityField->get(*cellIt, 0) = vel[0];
         velocityField->get(*cellIt, 1) = vel[1];
         velocityField->get(*cellIt, 2) = vel[2];

      }
   }
} // function setVelocityFieldsHenrik


} // namespace walberla
