//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file PorousMediaGPU.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================

#include "blockforest/Initialization.h"
#include "blockforest/StructuredBlockForest.h"
#include "blockforest/communication/UniformBufferedScheme.h"

#include "core/Abort.h"
#include "core/DataTypes.h"
#include "core/SharedFunctor.h"
#include "core/debug/CheckFunctions.h"
#include "core/logging/Logging.h"
#include "core/math/Vector3.h"
#include "core/math/Constants.h"
#include "core/Environment.h"
#include "core/mpi/MPIManager.h"
#include "core/MemoryUsage.h"
#include "core/mpi/Reduce.h"
#include "core/timing/RemainingTimeLogger.h"
#include "core/timing/TimingPool.h"

#include "field/AddToStorage.h"
#include "field/CellCounter.h"
#include "field/FlagField.h"
#include "field/StabilityChecker.h"
#include "field/adaptors/AdaptorCreators.h"
#include "field/iterators/FieldIterator.h"
#include "field/vtk/VTKWriter.h"
#include "field/vtk/FlagFieldCellFilter.h"
#include "field/blockforest/BlockDataHandling.h"

#include "geometry/InitBoundaryHandling.h"
#include "geometry/mesh/TriangleMeshIO.h"

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
#   include "gpu/AddGPUFieldToStorage.h"
#   include "gpu/DeviceSelectMPI.h"
#   include "gpu/ErrorChecking.h"
#   include "gpu/FieldCopy.h"
#   include "gpu/HostFieldAllocator.h"
#   include "gpu/ParallelStreams.h"
#   include "gpu/communication/NonUniformGPUScheme.h"
#endif

#include "lbm_generated/communication/UniformGeneratedPdfPackInfo.h"
#include "lbm_generated/evaluation/PerformanceEvaluation.h"
#include "lbm_generated/field/AddToStorage.h"
#include "lbm_generated/field/PdfField.h"

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
#   include "lbm_generated/gpu/AddToStorage.h"
#   include "lbm_generated/gpu/GPUPdfField.h"
#   include "lbm_generated/gpu/NonuniformGeneratedGPUPdfPackInfo.h"
#   include "lbm_generated/gpu/BasicRecursiveTimeStepGPU.h"
#endif

#include "postprocessing/FieldToSurfaceMesh.h"

#include "timeloop/SweepTimeloop.h"
#include "vtk/VTKOutput.h"

#include <cstdlib>
#include <iostream>
#include <memory>

#include "DragEvaluation.h"
#include "Inflow.h"
#include "InitializerFunctions.h"
#include "GridGeneration.h"
#include "Setup.h"
#include "Probe.h"
#include "TPMS.h"
#include "Types.h"
#include "util.h"

#include "PorousMediaGPUStaticDefines.h"

using namespace walberla;

#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
using gpu::communication::NonUniformGPUScheme;
using lbm_generated::NonuniformGeneratedGPUPdfPackInfo;
#else
using blockforest::communication::UniformBufferedScheme;
using lbm_generated::UniformGeneratedPdfPackInfo;
#endif

using VelocityFieldDH_T = field::DefaultBlockDataHandling< VelocityField_T >;
using TensorFieldDH_T = field::DefaultBlockDataHandling< TensorField_T >;

class HemisphereWallDistance
{
 public:
   HemisphereWallDistance(const real_t radius, const real_t hemispheresDistance) : radius_(radius), hemispheresDistance_(hemispheresDistance){} 

   real_t operator()(const Cell& fluidCell, const Cell& boundaryCell,
                     const shared_ptr< StructuredBlockForest >& SbF, IBlock& block){

      auto domain = SbF->getDomain();
      const real_t yMin = domain.yMin();
      const real_t yMax = domain.yMax();

      // WALBERLA_CHECK(!hemispheresContains(fluid, center, radius_), "fluid cell is in contained in Hemisphere (" << fluid[0] << ", " << fluid[1] << ", " << fluid[2] << "). The block local cell is " << fluidCell)
      // WALBERLA_CHECK(hemispheresContains(boundary, center, radius_), "boundary cell is not in contained in Hemisphere (" << boundary[0] << ", " << boundary[1] << ", " << boundary[2] << "). The block local cell is " << boundaryCell)

      Vector3< real_t > boundary = SbF->getBlockLocalCellCenter( block, boundaryCell );
      Vector3< real_t > fluid    = SbF->getBlockLocalCellCenter( block, fluidCell );

      auto center = getCenterOfHemisphere(boundary, hemispheresDistance_, yMin, yMax);

      if(hemispheresContains(fluid, center, radius_) || !hemispheresContains(boundary, center, radius_)){
         return real_c(0.5);
      }

      return delta(fluid, boundary, center);
   }

   real_t delta(const Vector3< real_t >& fluid, const Vector3< real_t >& boundary, const Vector3< real_t >& center) const
   {
      // http://devmag.org.za/2009/04/17/basic-collision-detection-in-2d-part-2/
      const Vector3< real_t > f = fluid - center;
      const Vector3< real_t > d = (boundary - center) - f;

      const real_t dDotd = d[0] * d[0] + d[1] * d[1] + d[2] * d[2];
      const real_t fDotf = f[0] * f[0] + f[1] * f[1] + f[2] * f[2];
      const real_t dDotf = d[0] * f[0] + d[1] * f[1] + d[2] * f[2];

      const real_t b = real_c(2.0) * dDotf;
      const real_t c = fDotf - (radius_ * radius_);

      const real_t bb4ac = b * b - (real_c(4.0) * dDotd * c);
      WALBERLA_CHECK_GREATER_EQUAL(bb4ac, real_c(0.0))

      const real_t sqrtbb4ac = std::sqrt(bb4ac);
      const real_t alpha = std::min((-b + sqrtbb4ac) / (real_c(2.0) * dDotd), (-b - sqrtbb4ac) / (real_c(2.0) * dDotd));

      WALBERLA_CHECK_GREATER_EQUAL(alpha, real_c(0.0))
      WALBERLA_CHECK_LESS_EQUAL(alpha, real_c(1.0))

      return alpha;
   }


private:
   const real_t radius_;
   const real_t hemispheresDistance_;
};

real_t defaultWallDistance(const Cell& /*fluidCell*/, const Cell& /*boundaryCell*/, const shared_ptr< StructuredBlockForest >& /*SbF*/, IBlock& /*block*/){
   return real_c(0.5);
}

void refineWalls(SetupBlockForest& forest, const real_t hemispheresRadius, const real_t dx, const uint_t level)
{
   auto domain = forest.getDomain();
   const real_t yMin = domain.yMin();
   const real_t yMax = domain.yMax();

   const real_t t = hemispheresRadius + dx;

   const AABB minAABB(domain.xMin(), domain.yMin(), domain.zMin(), domain.xMax(), domain.yMin() + t, domain.zMax());
   const AABB maxAABB(domain.xMin(), domain.yMax() - t, domain.zMin(), domain.xMax(), domain.yMax(), domain.zMax());

   for (auto block = forest.begin(); block != forest.end(); ++block){
      const AABB& aabb = block->getAABB();

      if (block->getLevel() < level && (minAABB.intersects(aabb) || maxAABB.intersects(aabb)) ){
      //if (block->getLevel() < level && minAABB.intersects(aabb) ){
         block->setMarker(true);
      }
   }
}

real_t calculateBulkVelocity(const shared_ptr< StructuredBlockStorage >& blocks, const IDs& ids, const FlagUID domainFlagUID) {
   real_t bulkVelocity = real_c(0.0);
   real_t numCells = real_c(0);

   for( auto block = blocks->begin(); block != blocks->end(); ++block) {
      const uint_t level = blocks->getLevel(*block);
      const real_t scaleFactor = real_c(1.0) / real_c( 1 << level );
      const real_t cellVolume = scaleFactor * scaleFactor * scaleFactor;
      auto * velocityField = block->template getData<VelocityField_T>(ids.velocityField);
      auto * flagField = block->template getData<FlagField_T>(ids.flagField);
      auto domainFlag = flagField->getFlag(domainFlagUID);
      const auto ci = velocityField->xyzSize();
      for(auto cellIt = ci.begin(); cellIt != ci.end(); ++cellIt) {
         if(isFlagSet(flagField->get(*cellIt), domainFlag)){
            bulkVelocity += (velocityField->get(*cellIt, 0) * cellVolume);
            numCells += cellVolume;
         }
      }
   }

   mpi::allReduceInplace< real_t >(bulkVelocity, mpi::SUM);
   bulkVelocity /= real_c(numCells);
   return bulkVelocity;
}



int main(int argc, char** argv){
   Environment env( argc, argv );
   mpi::MPIManager::instance()->useWorldComm();
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   gpu::selectDeviceBasedOnMpiRank();
#endif

   auto config = env.config();

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                        SETUP AND CONFIGURATION                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto parameters        = config->getOneBlock("Parameters");
   auto loggingParameters = config->getOneBlock("Logging");
   auto boundariesConfig  = config->getBlock("Boundaries");

   const uint_t welfordSamplingStart = parameters.getParameter< uint_t >("welfordSamplingStart", 0);
   const uint_t welfordSamplingFrequency = parameters.getParameter< uint_t >("welfordSamplingFrequency", 0);

   Setup setup(config, infoMap);
   WALBERLA_LOG_INFO_ON_ROOT( "Simulation run data:"
                              "\n- simulation parameters:"   << std::setprecision(16) <<
                              "\n   + collision model:       " << setup.collisionModel <<
                              "\n   + stencil:               " << setup.stencil <<
                              "\n   + streaming:             " << setup.streamingPattern <<
                              "\n   + forceModel:            " << infoMap["forceModel"] <<
                              "\n   + FourthOrderCorrection: " << infoMap["FourthOrderCorrection"] <<
                              "\n   + compressible:          " << ( StorageSpecification_T::compressible ? "yes" : "no" ) <<
                              "\n- simulation properties:    "
                              "\n   + kin. viscosity:        " << setup.kinViscosity << " [m^2/s] (on the coarsest grid)" <<
                              "\n   + viscosity LB:          " << setup.viscosity  << " [dx*dx/dt] (on the coarsest grid)" <<
                              "\n   + omega:                 " << setup.omega << " (on the coarsest grid)" <<
                              "\n   + omega:                 " << setup.omegaF << " (on the finest grid)" <<
                              "\n   + inflow velocity:       " << setup.referenceVelocity << " [m/s]" <<
                              "\n   + lattice velocity:      " << setup.latticeVelocity << " [dx/dt]" <<
                              "\n   + friction velocity:     " << setup.frictionVelocity << " [dx/dt]" <<
                              "\n   + Reynolds number:       " << setup.reynoldsNumber <<
                              "\n   + Mach number:           " << setup.machNumber <<
                              "\n   + driving force:         " << setup.drivingForce <<
                              "\n   + dx (coarsest grid):    " << setup.dxC << " [m]" <<
                              "\n   + dx (finest grid):      " << setup.dxF << " [m]" <<
                              "\n   + dt (coarsest grid):    " << setup.dt << " [s]" <<
                              "\n   + conversion Faktor:     " << setup.conversionFactor <<
                              "\n   + #time steps:           " << setup.timesteps << " (on the coarsest grid, " << ( real_c(1.0) / setup.dt ) << " for 1s of real time)" <<
                              "\n   + simulation time:       " << ( real_c( setup.timesteps ) * setup.dt ) << " [s]"
                              "\n   + flow through cycles:   " << ( setup.flowThroughCycles ))

   bool writeSetupForestAndReturn = loggingParameters.getParameter< bool >("writeSetupForestAndReturn", false);
   if (uint_c(MPIManager::instance()->numProcesses()) > 1) writeSetupForestAndReturn = false;
   shared_ptr< BlockForest > bfs;

   auto refinementSelection = std::bind(refineWalls, _1, real_c(1.0), setup.dxF, setup.refinementLevels);

   createBlockForest(bfs, setup, refinementSelection);

   if (writeSetupForestAndReturn && mpi::MPIManager::instance()->numProcesses() == 1){
      WALBERLA_LOG_INFO_ON_ROOT("BlockForest has been created and writen to file. Returning program")
      logging::Logging::printFooterOnStream();
      return EXIT_SUCCESS;
   }

   auto blocks = std::make_shared< StructuredBlockForest >(bfs, setup.cellsPerBlock[0], setup.cellsPerBlock[1], setup.cellsPerBlock[2]);
   blocks->createCellBoundingBoxes();
   auto finalDomain = blocks->getDomain();

   uint_t beginTimeStep = 0;
   uint_t welfordCtr = 0;
   if (setup.loadSnapshot){
      WALBERLA_ROOT_SECTION(){
         std::ifstream file;
         file.open(setup.checkpointConfigFile);
         if (file.fail()) WALBERLA_ABORT("Error: " << setup.checkpointConfigFile << " could not be opened!");
         file >> beginTimeStep;
         file >> welfordCtr;
         file.close();
      }
      mpi::broadcastObject(beginTimeStep);
      mpi::broadcastObject(welfordCtr);

      WALBERLA_LOG_INFO_ON_ROOT("Successfully read config parameters from checkpoint config file:")
      WALBERLA_LOG_INFO_ON_ROOT(" - beginTimeStep = " << beginTimeStep)
   }

   setup.timesteps += beginTimeStep;

   const StorageSpecification_T StorageSpec = StorageSpecification_T();
   // Creating fields
   uint_t step = 0;
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Allocating data")
   IDs ids;
   auto allocator = make_shared< gpu::HostFieldAllocator<macroFieldType> >(); // use pinned memory allocator for faster CPU-GPU memory transfers

   auto defaultFieldSize = [](const shared_ptr< StructuredBlockStorage > & sbs, IBlock * const block){
      return Vector3<uint_t>( sbs->getNumberOfXCells( *block ), sbs->getNumberOfYCells( *block ), sbs->getNumberOfZCells( *block ) );
   };

   auto inflowFieldSize = [](const shared_ptr< StructuredBlockStorage > & sbs, IBlock * const block){
      return Vector3<uint_t>( uint_c(1), sbs->getNumberOfYCells( *block ), sbs->getNumberOfZCells( *block ) );
   };

   if (setup.loadSnapshot){
      shared_ptr< lbm_generated::internal::PdfFieldHandling< StorageSpecification_T > > pdfFieldDataHandling =
         make_shared< lbm_generated::internal::PdfFieldHandling< StorageSpecification_T > >( blocks, StorageSpec, setup.numGhostLayers, field::fzyx);

      ids.pdfField = blocks->getBlockStorage().loadBlockData(setup.pdfFieldFile, pdfFieldDataHandling, "pdfs");

      if(welfordSamplingFrequency > 0){

         std::shared_ptr< VelocityFieldDH_T > meanVelocityDataHandling =
         std::make_shared< VelocityFieldDH_T >(blocks, setup.numGhostLayers, macroFieldType(0.0), field::fzyx, defaultFieldSize, allocator);

         std::shared_ptr< TensorFieldDH_T > sosDataHandling =
         std::make_shared< TensorFieldDH_T >(blocks, setup.numGhostLayers, macroFieldType(0.0), field::fzyx, defaultFieldSize, allocator);

         ids.meanVelocityField = blocks->getBlockStorage().loadBlockData(setup.meanVelocityFile, meanVelocityDataHandling, "mean velocity");
         ids.sosField = blocks->getBlockStorage().loadBlockData(setup.sosFile, sosDataHandling, "sos velocity");

      }

      std::shared_ptr< VelocityFieldDH_T > inflowSliceDataHandling =
      std::make_shared< VelocityFieldDH_T >(blocks, setup.numGhostLayers, macroFieldType(0.0), field::fzyx, inflowFieldSize);
      ids.velocityFieldInflowSlice = blocks->getBlockStorage().loadBlockData(setup.inflowFile, inflowSliceDataHandling, "velocity Inflow");
   }
   else{
      ids.pdfField  = lbm_generated::addPdfFieldToStorage(blocks, "pdfs", StorageSpec, setup.numGhostLayers, field::fzyx);
      ids.velocityFieldInflowSlice    = field::addToStorage< VelocityField_T >(blocks, "velocity Inflow", inflowFieldSize, macroFieldType(0.0), field::fzyx, setup.numGhostLayers);
   }

   ids.velocityField = field::addToStorage< VelocityField_T >(blocks, "vel", macroFieldType(0.0), field::fzyx, setup.numGhostLayers, allocator);
   ids.densityField  = field::addToStorage< ScalarField_T >(blocks, "density", macroFieldType(1.0), field::fzyx, setup.numGhostLayers, allocator);
   ids.flagField     = field::addFlagFieldToStorage< FlagField_T >(blocks, "Boundary Flag Field", setup.numGhostLayers);

   ids.pdfFieldGPU      = lbm_generated::addGPUPdfFieldToStorage< PdfField_T >(blocks, ids.pdfField, StorageSpec, "pdfs on GPU", true);
   ids.velocityFieldGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.velocityField, "velocity on GPU", true);
   ids.densityFieldGPU  = gpu::addGPUFieldToStorage< ScalarField_T >(blocks, ids.densityField, "velocity on GPU", true);
   ids.flagFieldGPU     = gpu::addGPUFieldToStorage< FlagField_T >(blocks, ids.flagField, "flag on GPU", true);

   ids.velocityFieldInflowSliceGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.velocityFieldInflowSlice, "velocity Inflow on GPU", true);

   const Cell innerOuterSplit = Cell(parameters.getParameter< Vector3<cell_idx_t> >("innerOuterSplit", Vector3<cell_idx_t>(1, 1, 1)));
   SweepCollection_T sweepCollection(blocks, ids.densityFieldGPU, ids.pdfFieldGPU, ids.velocityFieldGPU, setup.drivingForce, setup.omega, innerOuterSplit);
   sweepCollection.initialiseBlockPointer();
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Initialising data")

   if (!setup.loadSnapshot){
      if(setup.initialiseWithlatticeVelocity){
         for (auto& block : *blocks){
            auto velocity = block.getData< VelocityField_T >(ids.velocityField);
            WALBERLA_FOR_ALL_CELLS_INCLUDING_GHOST_LAYER_XYZ(velocity,
               velocity->get(x, y, z, 0) = setup.latticeVelocity;
            )
         }
      }

      if(parameters.getParameter< bool >("initialiseWithTurbulence", false)){
         setVelocityFieldsAsmuth(blocks, setup, ids);
      }

      gpu::fieldCpy< gpu::GPUField< VelocityField_T::value_type >, VelocityField_T >(blocks, ids.velocityFieldGPU, ids.velocityField);
      for (auto& block : *blocks){
         sweepCollection.initialise(&block, cell_idx_c(setup.numGhostLayers - 1));
         auto velocityInflow = block.getData< VelocityField_T >(ids.velocityFieldInflowSlice);
         WALBERLA_FOR_ALL_CELLS_INCLUDING_GHOST_LAYER_XYZ(velocityInflow,
            velocityInflow->get(x, y, z, 0) = setup.latticeVelocity;
         )
      }

      gpu::fieldCpy< gpu::GPUField< VelocityField_T::value_type >, VelocityField_T >(blocks, ids.velocityFieldInflowSliceGPU, ids.velocityFieldInflowSlice);
   
   }
   // else{
   //    for (auto& block : *blocks)
   //       sweepCollection.calculateMacroscopicParameters(&block, cell_idx_c(setup.numGhostLayers - 1));
   // }

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                      LB SWEEPS AND BOUNDARY HANDLING                                       ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   // Boundaries
   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setting boundary conditions")
   for (auto& block : *blocks){
      auto flagField = block.getData< FlagField_T >(ids.flagField);
      setup.inflowFlag          = flagField->registerFlag(setup.inflowUID);
      setup.turbulentInflowFlag = flagField->registerFlag(setup.turbulentInflowUID);
      setup.outflowFlag         = flagField->registerFlag(setup.outflowUID);
      setup.obstacleFlag        = flagField->registerFlag(setup.obstacleUID);
      setup.noSlipFlag          = flagField->registerFlag(setup.noSlipUID);
      setup.fluidFlag           = flagField->registerFlag(setup.fluidUID);
   }

   geometry::initBoundaryHandling< FlagField_T >(*blocks, ids.flagField, boundariesConfig);
   std::function< VelocityField_T::value_type (const Cell&, const Cell&, const shared_ptr< StructuredBlockForest >&, IBlock&) > wallDistanceFunctor = defaultWallDistance;
   if(setup.useGrid){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setting Grid after inflow to produce turbulence structure")
      initGrid(blocks, ids.flagField, setup);
   }

   if(config->getNumBlocks("RoughWall") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Using RoughWall setup")
      auto wallParameters = config->getOneBlock("RoughWall");
      const real_t radius   = wallParameters.getParameter< real_t >("hemispheresRadius");
      const real_t hemispheresDistance = wallParameters.getParameter< real_t >("hemispheresDistance");
      initHemisphers(blocks, ids.flagField, setup, config->getOneBlock("RoughWall"));
      HemisphereWallDistance wallDist(radius, hemispheresDistance);
      wallDistanceFunctor = wallDist;
   }

   if(config->getNumBlocks("genericPorousMatrix") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Using Generic Porous Matrix setup")
      initGenericPorousMatrix(blocks, ids.flagField, setup, config->getOneBlock("genericPorousMatrix"));
   }

   if(config->getNumBlocks("TPMS") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Placing TPMS material")
      auto tpms = std::make_shared<TPMS>(config->getOneBlock("TPMS"), setup);
      tpms->setupBoundary(blocks, ids.flagField);
   }

   geometry::setNonBoundaryCellsToDomain< FlagField_T >(*blocks, ids.flagField, setup.fluidUID, cell_idx_c(0));
   BoundaryCollection_T boundaryCollection(blocks, ids.flagField, ids.pdfFieldGPU, setup.fluidUID,
                                           ids.velocityFieldInflowSliceGPU, setup.latticeVelocity, wallDistanceFunctor, ids.pdfField);

   if(config->getNumBlocks("TPMS") > 0 && loggingParameters.getParameter< bool >("writeSurfaceMeshOfTPMS", false) ){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Writing TPMS as surface file")
      auto mesh = postprocessing::flagFieldToSurfaceMesh< FlagField_T >(blocks, ids.flagField, setup.obstacleFlag, true);
      WALBERLA_EXCLUSIVE_WORLD_SECTION(0){
         geometry::writeMesh("output/TPMS.obj", *mesh);
      }
   }
   gpu::fieldCpy< GPUFlagField_T, FlagField_T >(blocks, ids.flagFieldGPU, ids.flagField);

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                           COMMUNICATION SCHEME                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   const bool gpuEnabledMPI = parameters.getParameter< bool >("gpuEnabledMPI", false);
   auto nonUniformCommunication = std::make_shared< NonUniformGPUScheme< CommunicationStencil_T > >(blocks, gpuEnabledMPI);
   auto nonUniformPackInfo      = lbm_generated::setupNonuniformGPUPdfCommunication< GPUPdfField_T >(blocks, ids.pdfFieldGPU);
   nonUniformCommunication->addPackInfo(nonUniformPackInfo);

   if(setup.loadSnapshot){
      const uint_t maxLevel = blocks->getDepth();
      for (uint_t level = 0; level <= maxLevel; level++){
         nonUniformCommunication->communicateEqualLevel(level);
         if(level + 1 <= maxLevel){
            nonUniformCommunication->communicateCoarseToFine(level + 1);
            nonUniformCommunication->communicateFineToCoarse(level + 1);
         }
      }
   }

   WALBERLA_MPI_BARRIER()
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                          TIME STEP DEFINITIONS                                             ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   const int streamLowPriority = 0;
   auto defaultStream = gpu::StreamRAII::newPriorityStream(streamLowPriority);
   SweepTimeloop timeloop(blocks->getBlockStorage(), setup.timesteps);
   timeloop.setCurrentTimeStep(beginTimeStep);

   lbm_generated::BasicRecursiveTimeStepGPU< GPUPdfField_T, SweepCollection_T, BoundaryCollection_T >
      LBMMeshRefinement(blocks, ids.pdfFieldGPU, sweepCollection, boundaryCollection, nonUniformCommunication, nonUniformPackInfo);

   // timeloop.addFuncBeforeTimeStep(LBMMeshRefinement, "Refinement timestep");
   LBMMeshRefinement.addRefinementToTimeLoop(timeloop, uint_c(0));

   shared_ptr< DragEvaluation > evaluation;
   if(config->getNumBlocks("RoughWall") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Setup Drag Evaluation")
      evaluation = make_shared<DragEvaluation>(blocks, boundaryCollection, ids, setup, config);
      evaluation->setTotalExecutionCounter(beginTimeStep);
      LBMMeshRefinement.addPostBoundaryHandlingBlockFunction(evaluation->forceCalculationFunctor());
   }

   shared_ptr< BulkVelocityCalculator > bvc;
   if(config->getNumBlocks("ForcingParameters") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Add bulkvelocity calculator")
      bvc = std::make_shared<BulkVelocityCalculator>(blocks, ids, setup, config->getOneBlock("ForcingParameters"));

      if(bvc->forceUpdateFrequency() > 0){
         auto updateForce = [&](){
         if(timeloop.getCurrentTimeStep() % bvc->forceUpdateFrequency() == 0) {
               const real_t correctedDrivingForce = bvc->getDrivingForce();
               sweepCollection.setFx(correctedDrivingForce);
            }
         };
         timeloop.addFuncBeforeTimeStep(updateForce, "Update force");
      }
   }

   shared_ptr< PressureDropEvaluator > pde;
   if(config->getNumBlocks("PressureEvaluation") > 0){
      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Add pressure drop calculator")

      pde = std::make_shared<PressureDropEvaluator>(blocks, setup, ids, config->getOneBlock("PressureEvaluation"));
      pde->setInitialTimestep(beginTimeStep);
      timeloop.addFuncBeforeTimeStep(pde->updateFunctor(), "Pressure Drop Evaluation");

   }


   std::function<void(IBlock*)> copyPDFs = gpu::fieldCpyFunctor< PdfField_T, gpu::GPUField< PdfField_T::value_type > >(ids.pdfField, ids.pdfFieldGPU);
   std::function<void(IBlock*)> copyVelocityInflow = gpu::fieldCpyFunctor< VelocityField_T, gpu::GPUField< VelocityField_T::value_type > >(ids.velocityFieldInflowSlice, ids.velocityFieldInflowSliceGPU);
   std::function<void(IBlock*)> copyVelocity = gpu::fieldCpyFunctor< VelocityField_T, gpu::GPUField< VelocityField_T::value_type > >(ids.velocityField, ids.velocityFieldGPU);
   std::function<void(IBlock*)> copyDensity  = gpu::fieldCpyFunctor< ScalarField_T, gpu::GPUField< ScalarField_T::value_type > >(ids.densityField, ids.densityFieldGPU);
   std::function<void(IBlock*)> copyMeanVelocity;
   std::function<void(IBlock*)> copySOSField;

   std::shared_ptr<WelfordSweep_T> welfordSweep;

   if(welfordSamplingFrequency > 0){

      WALBERLA_LOG_INFO_ON_ROOT(++step << ". Adding welford algorithm to evaluate mean field quantities")

      if (!setup.loadSnapshot){
         ids.meanVelocityField = field::addToStorage< VelocityField_T >(blocks, "mean velocity", macroFieldType(0.0), field::fzyx, setup.numGhostLayers, allocator);
         ids.sosField = field::addToStorage< TensorField_T >(blocks, "sos velocity", macroFieldType(0.0), field::fzyx, setup.numGhostLayers, allocator);
      }
      
      ids.meanVelocityFieldGPU = gpu::addGPUFieldToStorage< VelocityField_T >(blocks, ids.meanVelocityField, "mean velocity on GPU", true);
      ids.sosFieldGPU = gpu::addGPUFieldToStorage< TensorField_T >(blocks, ids.sosField, "sos field on GPU", true);

      copyMeanVelocity = gpu::fieldCpyFunctor< VelocityField_T, gpu::GPUField< VelocityField_T::value_type > >(ids.meanVelocityField, ids.meanVelocityFieldGPU);
      copySOSField     = gpu::fieldCpyFunctor< TensorField_T, gpu::GPUField< TensorField_T::value_type > >(ids.sosField, ids.sosFieldGPU);

      welfordSweep = make_shared<WelfordSweep_T>(ids.meanVelocityFieldGPU, ids.sosFieldGPU, ids.velocityFieldGPU, real_c(welfordCtr));

      auto welfordUpdate = [&](){
         if(timeloop.getCurrentTimeStep() > welfordSamplingStart  && (timeloop.getCurrentTimeStep() % welfordSamplingFrequency) == 0) {
            const uint_t velCtr = uint_c(welfordSweep->getCounter());
            welfordSweep->setCounter(welfordSweep->getCounter() + real_c(1.0));
            for (auto& block : *blocks){
               welfordSweep->run(&block);
            }
         };
      };
      timeloop.addFuncBeforeTimeStep(welfordUpdate, "Welford Sampling");
   }

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                          Post Processing                                                   ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   auto inflow = std::make_shared<Inflow>(blocks, setup, ids);
   auto probe = std::make_shared<Probe>(blocks, setup, ids, config->getOneBlock("Probes"));
   probe->setInitialTimestep(beginTimeStep);


   probe->addPoint(finalDomain.center());
   const real_t xStart = finalDomain.xMin();
   const uint_t nx = uint_c(finalDomain.xSize() / setup.dxC);
   for(uint_t i = 0; i <= nx; i+=4){
      probe->addPoint(Vector3<real_t>(xStart + real_c(i) * setup.dxC, finalDomain.center()[1], finalDomain.center()[2] ));
   }


/*
   const real_t xStart = finalDomain.xMin();
   const uint_t spacing = uint_c(4);
   const uint_t nx = uint_c(finalDomain.xSize() / setup.dxC);

   for(uint_t j = 0; j < 3; ++j){
      const Vector3<real_t> lineCenter(finalDomain.center()[0], real_c(2.0) + real_c(j) * real_c(2.0), finalDomain.center()[2]);
      probe->addPoint(lineCenter);
      for(uint_t i = 0; i <= nx; i+=spacing){
         probe->addPoint(Vector3<real_t>(xStart + real_c(i) * setup.dxC, real_c(2.0) + real_c(j) * real_c(2.0), finalDomain.center()[2] ));
      }
   }

   const real_t yStart = finalDomain.yMin();
   const uint_t ny = uint_c(finalDomain.ySize() / setup.dxC);

   for(uint_t j = 0; j < 5; ++j){
      const Vector3<real_t> lineCenter(real_c(2.0) + real_c(j) * real_c(2.0), finalDomain.center()[1], finalDomain.center()[2]);
      probe->addPoint(lineCenter);
      for(uint_t i = 0; i <= ny; i+=spacing){
         probe->addPoint(Vector3<real_t>(real_c(2.0) + real_c(j) * real_c(2.0), yStart + real_c(i) * setup.dxC, finalDomain.center()[2] ));
      }
   }

*/

   probe->initialise();

   
   timeloop.addFuncBeforeTimeStep(inflow->updateGPUFunctor(), "Update inflow velocity");
   timeloop.addFuncBeforeTimeStep(probe->updateFunctor(), "Update Probe Data");

   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
   ///                                             VTK Output                                                     ///
   //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   // VTK
   auto VTKWriter = config->getOneBlock("VTKWriter");
   const uint_t vtkWriteFrequency       = VTKWriter.getParameter< uint_t >("vtkWriteFrequency", 0);
   const uint_t vtkGhostLayers          = VTKWriter.getParameter< uint_t >("vtkGhostLayers", 0);
   const bool amrFileFormat             = VTKWriter.getParameter< bool >("amrFileFormat", false);
   const bool oneFilePerProcess         = VTKWriter.getParameter< bool >("oneFilePerProcess", false);
   const real_t samplingResolution      = VTKWriter.getParameter< real_t >("samplingResolution", real_c(-1.0));
   const uint_t initialWriteCallsToSkip = VTKWriter.getParameter< uint_t >("initialWriteCallsToSkip", uint_c(0.0));
   const real_t sliceThickness          = VTKWriter.getParameter< real_t >("sliceThickness", real_c(1.0));
   const bool lastStepAsVolumeFile      = VTKWriter.getParameter< bool >("lastStepAsVolumeFile", false);

   if (vtkWriteFrequency > 0){
      const std::string identifier = setup.fileStem;
      auto vtkOutput =
         vtk::createVTKOutput_BlockData(*blocks, identifier, vtkWriteFrequency, vtkGhostLayers, false, setup.outputFolder,
                                        "simulation_step", false, true, true, false, 0, amrFileFormat, oneFilePerProcess);

      vtkOutput->setInitialWriteCallsToSkip(initialWriteCallsToSkip);
      vtkOutput->addBeforeFunction([&]() {

         for (auto& block : *blocks){
            if (VTKWriter.getParameter< bool >("velocity", false)){copyVelocity(&block);}
            if (VTKWriter.getParameter< bool >("density", false)){copyDensity(&block);}
            if (VTKWriter.getParameter< bool >("meanVelocity", false) && copyMeanVelocity){copyMeanVelocity(&block);}
            if (VTKWriter.getParameter< bool >("reynoldsstress", false) && copySOSField){copySOSField(&block);}
         }

         if(lastStepAsVolumeFile && timeloop.getCurrentTimeStep() == setup.timesteps - 1){
            vtkOutput->removeCellInclusionFilters();
         }

      });

      vtkOutput->setSamplingResolution(samplingResolution);

      field::FlagFieldCellFilter<FlagField_T> fluidFilter( ids.flagField );
      fluidFilter.addFlag( setup.obstacleUID );
      fluidFilter.addFlag( setup.noSlipUID );
      vtkOutput->addCellExclusionFilter(fluidFilter);


      if (VTKWriter.getParameter< bool >("writeOnlySlice", true)){
         const AABB sliceXY(finalDomain.xMin(), finalDomain.yMin(), finalDomain.center()[2] - sliceThickness * setup.dxF,
                            finalDomain.xMax(), finalDomain.yMax(), finalDomain.center()[2] + sliceThickness * setup.dxF);
         vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXY));

         if (VTKWriter.getParameter< bool >("writeXZSlice", false)){
            const AABB sliceXZ(finalDomain.xMin(), finalDomain.center()[1] - setup.dxF, finalDomain.zMin(),
                               finalDomain.xMax(), finalDomain.center()[1] + setup.dxF, finalDomain.zMax());
            vtkOutput->addCellInclusionFilter(vtk::AABBCellFilter(sliceXZ));
         }
      }

      if (VTKWriter.getParameter< bool >("velocity", false)){
         if (VTKWriter.getParameter< bool >("writeVelocityAsMagnitude")){
            auto velWriter = make_shared< field::VTKMagnitudeWriter< VelocityField_T, float32 > >(ids.velocityField, "velocityMagnitude", setup.conversionFactor);
            vtkOutput->addCellDataWriter(velWriter);
         }
         else{
            auto velWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(ids.velocityField, "velocity", setup.conversionFactor);
            vtkOutput->addCellDataWriter(velWriter);
         }
      }
      if (VTKWriter.getParameter< bool >("meanVelocity", false) && ids.meanVelocityField){
         auto meanVelWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(ids.meanVelocityField, "meanVelocity");
         vtkOutput->addCellDataWriter(meanVelWriter);
      }
      if (VTKWriter.getParameter< bool >("reynoldsstress", false) && ids.sosField){
         auto sosWriter = make_shared< field::VTKWriter< TensorField_T, float32 > >(ids.sosField, "RR");
         vtkOutput->addCellDataWriter(sosWriter);
      }
      if (VTKWriter.getParameter< bool >("density", false)){
         auto densityWriter = make_shared< field::VTKWriter< ScalarField_T, float32 > >(ids.densityField, "density");
         vtkOutput->addCellDataWriter(densityWriter);
      }
      if (VTKWriter.getParameter< bool >("flag", false)){
         auto flagWriter = make_shared< field::VTKWriter< FlagField_T > >(ids.flagField, "flag");
         vtkOutput->addCellDataWriter(flagWriter);
      }
      timeloop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");
   }

   // log remaining time
   if (uint_c(setup.remainingTimeLoggerFrequency) > 0){
      timeloop.addFuncAfterTimeStep(
         timing::RemainingTimeLogger(timeloop.getNrOfTimeSteps(), setup.remainingTimeLoggerFrequency),
         "remaining time logger");
   }

   auto CheckerParameters      = config->getOneBlock("StabilityChecker");
   const uint_t checkFrequency = CheckerParameters.getParameter< uint_t >("checkFrequency", uint_t(0));
   if (checkFrequency > 0){
      auto checkFunction = [](VelocityField_T::value_type value) {  return value < math::abs(VelocityField_T::value_type(10)); };
      timeloop.addFuncAfterTimeStep(
         makeSharedFunctor(field::makeStabilityChecker< VelocityField_T, FlagField_T >(
            config, blocks, ids.velocityField, ids.flagField, setup.fluidUID, checkFunction)),
         "Stability check");
   }

   if(setup.storeSnapshot){
      auto snapShotFunction = [&](){
         if(timeloop.getCurrentTimeStep() > uint_c(0)  && ( (timeloop.getCurrentTimeStep() + 1) % setup.snapshotFrequency) == 0) {
            WALBERLA_LOG_INFO_ON_ROOT("Writing Snapshot")

            for (auto& block : *blocks){
               copyPDFs(&block);
               copyVelocityInflow(&block);
               if(copyMeanVelocity){copyMeanVelocity(&block);}
               if(copySOSField){copySOSField(&block);}
            }

            blocks->saveBlockData(setup.pdfFieldFile + "_tmp", ids.pdfField);
            blocks->saveBlockData(setup.inflowFile + "_tmp", ids.velocityFieldInflowSlice);

            if(copyMeanVelocity){blocks->saveBlockData(setup.meanVelocityFile + "_tmp", ids.meanVelocityField);}
            if(copySOSField){blocks->saveBlockData(setup.sosFile + "_tmp", ids.sosField);}

            WALBERLA_ROOT_SECTION(){
               uint_t welfordCounter = uint_c(0);
               if(welfordSweep)
                  welfordCounter = uint_c(welfordSweep->getCounter());

               std::string tmpCheckpointConfigFile = setup.checkpointConfigFile + "_tmp";
               std::ofstream file;
               file.open(tmpCheckpointConfigFile.c_str());

               file << std::setprecision(16);
               file << timeloop.getCurrentTimeStep() + 1 << "\n";
               file << welfordCounter << "\n";
               file.close();

               if(bvc){
                  bvc->writeStateToFile();
               }
            }

            WALBERLA_MPI_BARRIER();

            WALBERLA_ROOT_SECTION(){
               renameFile(setup.pdfFieldFile + "_tmp", setup.pdfFieldFile);
               renameFile(setup.inflowFile + "_tmp", setup.inflowFile);
               if(copyMeanVelocity){renameFile(setup.meanVelocityFile + "_tmp", setup.meanVelocityFile);}
               if(copySOSField){renameFile(setup.sosFile + "_tmp", setup.sosFile);}
               renameFile(setup.checkpointConfigFile + "_tmp", setup.checkpointConfigFile);
               }
         };
      };
      timeloop.addFuncAfterTimeStep(snapShotFunction, "Snapshot writing");
   }


   WALBERLA_MPI_BARRIER()
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif

   const lbm_generated::PerformanceEvaluation< FlagField_T > performance(blocks, ids.flagField, setup.fluidUID);
   field::CellCounter< FlagField_T > fluidCells(blocks, ids.flagField, setup.fluidUID);
   fluidCells();

   WALBERLA_LOG_INFO_ON_ROOT(++step << ". Starting Simulation with " << setup.timesteps << " timesteps")
   WcTimingPool timeloopTiming;
   WcTimer simTimer;

   WALBERLA_MPI_BARRIER()
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif

   simTimer.start();
   timeloop.run(timeloopTiming);
#if defined(WALBERLA_BUILD_WITH_GPU_SUPPORT)
   WALBERLA_GPU_CHECK(gpuDeviceSynchronize())
   WALBERLA_GPU_CHECK(gpuPeekAtLastError())
#endif
   WALBERLA_MPI_BARRIER()
   simTimer.end();
   WALBERLA_LOG_INFO_ON_ROOT("Simulation finished")
   double time = double_c(simTimer.max());

   if(copySOSField && evaluation){
      WALBERLA_CHECK_NOT_NULLPTR(welfordSweep)
      WALBERLA_LOG_INFO_ON_ROOT("Evaluating mean field")
      auto plotter = std::make_shared<Plotter>(blocks, setup, ids);
      for (auto& block : *blocks){
         copySOSField(&block);
      }

      const bool adaptForce = parameters.getParameter< bool >("adaptDrivingForce", false);

      const double ut = adaptForce ? evaluation->getMeanFrictionVelocity() : setup.frictionVelocity;
      const double welfordCounter = welfordSweep->getCounter();

      plotter->writeResult(ut, welfordCounter);
   }

   WALBERLA_MPI_SECTION() { walberla::mpi::reduceInplace(time, walberla::mpi::MAX); }
   performance.logResultOnRoot(setup.timesteps, time);

   const auto reducedTimeloopTiming = timeloopTiming.getReduced();
   WALBERLA_LOG_RESULT_ON_ROOT("Time loop timing:\n" << *reducedTimeloopTiming)

   printResidentMemoryStatistics();
   return EXIT_SUCCESS;
}