import sympy as sp
from pystencils import Target

from pystencils.field import fields
from pystencils.simp import insert_aliases, insert_constants

from lbmpy import LBStencil, LBMConfig, LBMOptimisation
from lbmpy.boundaries.boundaryconditions import NoSlip, NoSlipLinearBouzidi, ExtrapolationOutflow, UBB, FixedDensity
from lbmpy.creationfunctions import create_lb_collision_rule
from lbmpy.enums import Method, Stencil, ForceModel, SubgridScaleModel
from lbmpy.flow_statistics import welford_assignments

from pystencils_walberla import CodeGeneration, generate_info_header, generate_sweep
from lbmpy_walberla import generate_lbm_package, lbm_boundary_generator
from lbmpy_walberla import RefinementScaling

import pkg_resources

info_header = """
#pragma once
#include <map>
std::map<std::string, std::string> infoMap{{{{"stencil", "{stencil}"}},
                                           {{"streamingPattern", "{streaming_pattern}"}}, 
                                           {{"collisionOperator", "{collision_operator}"}},
                                           {{"forceModel", "{force_model}"}},
                                           {{"FourthOrderCorrection", "{fourth_order_correction}"}},
                                           {{"PythonEnv", "{python_env}"}}}};
"""

omega = sp.symbols("omega")
inlet_velocity = sp.symbols("u_x")
fx = sp.Symbol("Fx")
force_vector = tuple([fx, 0.0, 0.0])
max_threads = 256
 
with CodeGeneration() as ctx:
    target = Target.GPU if ctx.gpu else Target.CPU
    name_prefix = "PorousMediaGPU"
    sweep_params = {'block_size': (128, 1, 1)} if ctx.gpu else {}

    # The application is meant to be compiled with double precision. For single precision, the pdf_dtype can be switched
    # to float32. In this way the calculations are still performed in double precision while the application can profit
    # from enhanced performance due to the lower precision of the PDF field
    dtype = 'float64' if ctx.double_accuracy else 'float32'
    pdf_dtype = 'float64'

    stencil = LBStencil(Stencil.D3Q19)
    q = stencil.Q
    dim = stencil.D
 
    streaming_pattern = 'esopull'

    pdfs = fields(f"pdfs({stencil.Q}): {pdf_dtype}[3D]", layout='fzyx')
    velocity_field, density_field = fields(f"velocity({dim}), density(1) : {dtype}[{dim}D]", layout='fzyx')
    mean_velocity = fields(f"mean_velocity({dim}): {dtype}[{stencil.D}D]", layout='fzyx')
    sum_of_squares_field = fields(f"sum_of_squares_field({stencil.D**2}): {dtype}[{stencil.D}D]", layout='fzyx')

    macroscopic_fields = {'density': density_field, 'velocity': velocity_field}

    # method_enum = Method.CENTRAL_MOMENT
    method_enum = Method.CUMULANT
    force_model_enum = ForceModel.GUO
    lbm_config = LBMConfig(
        method=method_enum,
        stencil=stencil,
        relaxation_rate=omega,
        compressible=True,
        force=force_vector,
        force_model=force_model_enum,
        # fourth_order_correction=0.1 if method_enum == Method.CUMULANT and stencil.Q == 27 else False,
        streaming_pattern=streaming_pattern,
        output={'density': density_field, 'velocity': velocity_field}
    )

    lbm_opt = LBMOptimisation(cse_global=False, cse_pdfs=False, field_layout="fzyx")

    collision_rule = create_lb_collision_rule(lbm_config=lbm_config, lbm_optimisation=lbm_opt)
    if lbm_config.method == Method.CUMULANT:
        collision_rule = insert_constants(collision_rule)
        collision_rule = insert_aliases(collision_rule)
    lb_method = collision_rule.method

    no_slip = lbm_boundary_generator(class_name='NoSlip', flag_uid='NoSlip',
                                     boundary_object=NoSlip(), field_data_type=pdf_dtype)
 
    quadratic_bounce_back = NoSlipLinearBouzidi(calculate_force_on_boundary=True, data_type=dtype)
    no_slip_interpolated = lbm_boundary_generator(class_name='Obstacle', flag_uid='Obstacle',
                                                  boundary_object=quadratic_bounce_back,
                                                  field_data_type=pdf_dtype)


    turbulentubb = lbm_boundary_generator(class_name='TurbulentInflow', flag_uid='TurbulentInflow',
                                          boundary_object=UBB(velocity=velocity_field.center_vector, density=1.0, data_type=dtype, dim=dim),
                                          field_data_type=pdf_dtype)

    ubb = lbm_boundary_generator(class_name='UBB', flag_uid='UBB',
                                 boundary_object=UBB(velocity=(inlet_velocity, 0.0, 0.0), density=1.0, data_type=dtype, dim=dim),
                                 field_data_type=pdf_dtype)

    # outflow_boundary = ExtrapolationOutflow(stencil[4], lb_method, data_type=pdf_dtype, density=1.0)
    outflow_boundary = ExtrapolationOutflow(stencil[4], lb_method, data_type=pdf_dtype)
    outflow = lbm_boundary_generator(class_name='Outflow', flag_uid='Outflow',
                                     boundary_object=outflow_boundary,
                                     field_data_type=pdf_dtype)

    # outflow = lbm_boundary_generator(class_name='Outflow', flag_uid='Outflow',
    #                                 boundary_object=FixedDensity(1.0),
    #                                  field_data_type=pdf_dtype)

    refinement_scaling = RefinementScaling()
    refinement_scaling.add_standard_relaxation_rate_scaling(omega)
    refinement_scaling.add_force_scaling(fx)

    generate_lbm_package(ctx, name=name_prefix, collision_rule=collision_rule,
                         lbm_config=lbm_config, lbm_optimisation=lbm_opt,
                         nonuniform=True, refinement_scaling=refinement_scaling, boundaries=[no_slip, no_slip_interpolated, ubb, turbulentubb, outflow],
                         macroscopic_fields=macroscopic_fields, gpu_indexing_params=sweep_params,
                         target=target, data_type=dtype, pdfs_data_type=pdf_dtype,
                         max_threads=max_threads)

    welford_update = welford_assignments(field=velocity_field, mean_field=mean_velocity, sum_of_squares_field=sum_of_squares_field)
    generate_sweep(ctx, "Welford", welford_update, target=target)

    field_typedefs = {'VelocityField_T': velocity_field,
                      'ScalarField_T': density_field,
                      'TensorField_T': sum_of_squares_field}

    # Info header containing correct template definitions for stencil and field
    generate_info_header(ctx, f'{name_prefix}InfoHeader',
                         field_typedefs=field_typedefs)
    
    installed_packages = pkg_resources.working_set
    installed_packages_list = sorted([f"{i.key}=={i.version}" for i in installed_packages])

    infoHeaderParams = {
        'stencil': stencil.name.lower(),
        'streaming_pattern': streaming_pattern,
        'collision_operator': lbm_config.method.name.lower(),
        'force_model': force_model_enum.name.lower(),
        'fourth_order_correction': lbm_config.fourth_order_correction,
        'python_env': installed_packages_list,
    }

    ctx.write_file(f"{name_prefix}StaticDefines.h", info_header.format(**infoHeaderParams))