//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Probe.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#include "Probe.h"


namespace walberla{

void Probe::initialise(){
    if(probePoints_.empty()){return;}

    uint_t idx = 0;
    for(auto p: probePoints_){
        IBlock* block = blocks_->getBlock(p);
        if(block != nullptr){
            const uint_t level = blocks_->getLevel(*block);
            auto * flagField = block->template getData<FlagField_T>(ids_.flagField);
            auto domainFlag = flagField->getFlag(setup_.fluidUID);
            const Cell localCell = blocks_->getBlockLocalCell(*block, p);
            if(isFlagSet(flagField->get(localCell), domainFlag)){
                Cell globalCell;
                blocks_->transformBlockLocalToGlobalCell(globalCell, *block, localCell);
                Vector3<real_t> gloablCellCenter = blocks_->getCellCenter(globalCell, level);

                auto element = ProbeInfo(gloablCellCenter, localCell, idx);
                if(std::find(probeData_[block].begin(), probeData_[block].end(), element) == probeData_[block].end()){
                    probeData_[block].emplace_back(element);
                    idx++;
                }
            }
        }
    }

    for (auto const& it : probeData_){
        IBlock* block = it.first;
        auto probeInfos = it.second;

        auto * dataVector = block->getData< DataVector > ( probeDataID );
        auto & cpuCellVector = dataVector->cpuCellVector();
        auto & cpuDataVector = dataVector->cpuDataVector();

        const uint_t n = probeInfos.size();
        cpuCellVector.resize(3 * n);

        uint_t i = 0;
        for(const auto probeInfo: probeInfos){

            cpuCellVector[i] = probeInfo.localCell[0];
            cpuCellVector[i + n] = probeInfo.localCell[1];
            cpuCellVector[i + 2*n] = probeInfo.localCell[2];
            ++i;
        }

        cpuDataVector.resize(dataSetNames.size() * n * writeCounter_);
        dataVector->syncGPU();

        const Vector3<uint_t> s(n, uint_c(1), uint_c(1));
        Vector3<uint32_t> threadPerBlock;
        Vector3<uint32_t> numberBlocks;

        getBlockAndGridSize(s, threadPerBlock, numberBlocks);
        gpuBlocksAndThreads_[block] = std::make_pair(threadPerBlock, numberBlocks);
    }
    writeHeaderFiles();
}

void Probe::writeHeaderFiles(){
    const auto outputFolder = setup_.outputFolder;
    const auto fileStem = setup_.fileStem;
    const auto probeFolderString = outputFolder + "/" + fileStem + "Probes";
    filesystem::path probeFolder( probeFolderString );

    WALBERLA_ROOT_SECTION()
    {
        if(!setup_.loadSnapshot){
            if( filesystem::exists( probeFolder ))
                filesystem::remove_all( probeFolder );

            if ( !filesystem::exists( probeFolder ) )
                filesystem::create_directories( probeFolder );
        }
    }

    WALBERLA_MPI_BARRIER()
    if(probeData_.empty()){return;}

    if(!setup_.loadSnapshot){
        const auto fileName = probeFolderString + "/" + "ProbeRank_" + std::to_string(MPIManager::instance()->rank()) + ".h5";
        H5::FileAccPropList fileAccProps;

        const hsize_t dsetSize[1]{ 1 };
        const hsize_t maxSize[1]{ H5S_UNLIMITED };
        const hsize_t chunkSize[1]{writeCounter_};

        H5::DataSpace arrayDataSpace(1, dsetSize, maxSize);
        H5::DSetCreatPropList props;
        props.setChunk(1, chunkSize);

        auto h5file = H5::H5File(fileName.c_str(), H5F_ACC_CREAT | H5F_ACC_RDWR, H5::FileCreatPropList::DEFAULT, fileAccProps);
        H5::Group timeGroup = h5file.createGroup("time");
        timeGroup.createDataSet("t", io::h5Float(), arrayDataSpace, props);

        for (auto const& it : probeData_){
            for(uint_t i = 0; i < it.second.size(); ++i){
                const auto p = it.second[i].globalPoint;
                const auto idx = it.second[i].idx;
                std::string probeName = "Probe_" + std::to_string(idx);
                if (h5file.exists(probeName)) { continue; }
                H5::Group probeGroup = h5file.createGroup(probeName);

                hsize_t locationShape[1]{ 3 };
                H5::DataSpace spacingAttrDspace(1, locationShape, locationShape);
                H5::Attribute locationAttr = probeGroup.createAttribute("Location", io::h5FileFloat(), spacingAttrDspace);

                float location[3]{float(p[0]), float(p[1]), float(p[2])};
                locationAttr.write(io::h5Float(), location);

                for(auto name: dataSetNames){
                    probeGroup.createDataSet(name, io::h5Float(), arrayDataSpace, props);
                }
            }
        }
    }
}

void Probe::writeProbes(){
    if(probeData_.empty()){return;}

    std::vector<float> timeData(executionCounter_);
    const real_t t = (real_c(totalExecutionCounter_ + frequency_ - uint_c(1)) * setup_.dt) - ( real_c(executionCounter_ * frequency_) * setup_.dt);
    for(uint_t i = 0; i < executionCounter_; ++i){
        timeData[i] = float_c(t) + float_c(i)*float_c(setup_.dt);
    }

    writeData(timeData, "time", "t");

    for (auto const& it : probeData_){
        IBlock* block = it.first;
        auto* dataVector = block->getData< DataVector > ( probeDataID );
        dataVector->syncCPU();
        dataVector->clear();
    }

    const uint_t nDatasets = dataSetNames.size();

    std::vector<float> tmp(executionCounter_);
    for (auto const& it : probeData_){
        IBlock* block = it.first;

        auto* dataVector = block->getData< DataVector > ( probeDataID );
        auto & cpuDataVector = dataVector->cpuDataVector();

        const uint_t n = uint_c(dataVector->size());
        WALBERLA_CHECK_EQUAL(n, it.second.size())
        for(uint_t i = 0; i < n; ++i){
            const auto idx = it.second[i].idx;

            const std::string probeName = "Probe_" + std::to_string(idx);
            for(uint_t j = 0; j < nDatasets; ++j){
                auto dataSetName = dataSetNames[j];
                for(uint_t k = 0; k < executionCounter_; ++k){
                    tmp[k] = cpuDataVector[k*nDatasets*n + j*n + i];
                }
                writeData(tmp, probeName, dataSetName);
            }
        }
    }
    executionCounter_ = 0;
}

void Probe::writeData(const std::vector<float>& data, const std::string& groupName, const std::string& dataSetName){
    const auto outputFolder = setup_.outputFolder;
    const auto fileStem = setup_.fileStem;
    const auto probeFolderString = outputFolder + "/" + fileStem + "Probes";
    const auto fileName = probeFolderString + "/" + "ProbeRank_" + std::to_string(MPIManager::instance()->rank()) + ".h5";
    H5::FileAccPropList fileAccProps;
    auto h5file = H5::H5File(fileName.c_str(), H5F_ACC_RDWR, H5::FileCreatPropList::DEFAULT, fileAccProps);

    auto group = h5file.openGroup(groupName);
    H5::DataSet dataset = group.openDataSet(dataSetName);

    H5::DataSpace dataspace = dataset.getSpace();
    std::vector<hsize_t> dims(1);
    dataspace.getSimpleExtentDims(dims.data(), NULL);

    const hsize_t selectCount[1] { data.size() };
    const hsize_t extendSize[1] { dims[0] + data.size() };

    dataset.extend(extendSize);
    H5::DataSpace newDataspace = dataset.getSpace();

    newDataspace.selectHyperslab(H5S_SELECT_SET, selectCount, dims.data());
        
    H5::DataSpace memSpace(1, selectCount);
    dataset.write(data.data(), io::h5Float(), memSpace, newDataspace);
}

void Probe::update(){    
    ++totalExecutionCounter_;
    if (probePoints_.empty() || rampUpTime_ > totalExecutionCounter_ || (totalExecutionCounter_ - uint_c(1)) % frequency_ != 0) return;

    const float velocityScalingFactor = float_c(setup_.conversionFactor);

    for (auto const& it : probeData_){
        IBlock* block = it.first;
        auto density = block->getData< gpu::GPUField<double> >(ids_.densityFieldGPU);
        auto velocity = block->getData< gpu::GPUField<double> >(ids_.velocityFieldGPU);
        auto* dataVector = block->getData< DataVector > ( probeDataID );

        int64_t* gpuCellVector = dataVector->gpuCellVector();
        float* gpuDataVector = dataVector->advanceGPUDataVector();
        const int64_t n = dataVector->size();

        updateProbe(density, velocity, gpuCellVector, gpuDataVector, n, velocityScalingFactor, gpuBlocksAndThreads_[block].first, gpuBlocksAndThreads_[block].second);
    }

    ++executionCounter_;
    if(executionCounter_ >= writeCounter_ || totalExecutionCounter_ == setup_.timesteps){
        writeProbes();
    }
    
}



} // namespace walberla
