//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Probe.h
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#pragma once

#include "domain_decomposition/IBlock.h"

#include "core/logging/Initialization.h"
#include "core/math/Constants.h"
#include "core/Filesystem.h"
#include "core/mpi/Broadcast.h"
#include "core/mpi/Gatherv.h"

#include "gpu/GPUWrapper.h"
#include "gpu/ErrorChecking.h"
#include "gpu/FieldCopy.h"

#include "io/hdfWrapper.h"

#include "Setup.h"
#include "utilGPU.h"

namespace walberla{

class Probe
{
  public:

  struct ProbeInfo{
    Vector3<real_t> globalPoint;
    Cell localCell;
    uint_t idx;
    ProbeInfo(Vector3<real_t> globalPoint_, Cell localCell_, uint_t idx_) : globalPoint(globalPoint_), localCell(localCell_), idx(idx_) {}
    bool operator==(const ProbeInfo & o) const {return globalPoint == o.globalPoint;}
  };


  class DataVector{
    public:
        DataVector() = default;
        bool operator==(DataVector const &other) const { return other.cpuCellVector_ == cpuCellVector_ && other.cpuDataVector_ == cpuDataVector_; }

        ~DataVector() {
            if(synced){
              if(!cpuCellVector_.empty()){WALBERLA_GPU_CHECK(gpuFree(gpuCellVector_))}
              if(!cpuDataVector_.empty()){WALBERLA_GPU_CHECK(gpuFree(gpuDataVector_))}
            }
        }
        
        std::vector<int64_t>& cpuCellVector() { return cpuCellVector_; }
        std::vector<float>& cpuDataVector() { return cpuDataVector_; }

        int64_t* gpuCellVector() {return gpuCellVector_;}
        float* gpuDataVector() {return gpuDataVector_;}

        float* advanceGPUDataVector() {
          float* result = cur_;
          cur_ += int64_c(4) * size();
          return result;
        }

        inline void clear() { cur_ = begin_; }

        int64_t size() const {return int64_c(cpuCellVector_.size() / 3);}

        void syncGPU(){
          WALBERLA_GPU_CHECK(gpuMalloc((void **)&gpuCellVector_, sizeof(int64_t) * cpuCellVector_.size() ));
          WALBERLA_GPU_CHECK(gpuMemcpy( gpuCellVector_, &cpuCellVector_[0], sizeof(int64_t) * cpuCellVector_.size(), gpuMemcpyHostToDevice ));

          WALBERLA_GPU_CHECK(gpuMalloc((void **)&gpuDataVector_, sizeof(float) * cpuDataVector_.size() ));
          WALBERLA_GPU_CHECK(gpuMemcpy( gpuDataVector_, &cpuDataVector_[0], sizeof(float) * cpuDataVector_.size(), gpuMemcpyHostToDevice ));

          begin_ = gpuDataVector_;
          cur_ = gpuDataVector_;

          synced = true;
        }

        void syncCPU(){
          WALBERLA_GPU_CHECK(gpuMemcpy(&cpuDataVector_[0], gpuDataVector_, sizeof(float) * cpuDataVector_.size(), gpuMemcpyDeviceToHost ));
        }
        
    private:
        std::vector<int64_t> cpuCellVector_;
        std::vector<float> cpuDataVector_;

        int64_t *gpuCellVector_;
        float *gpuDataVector_;

        float *begin_ = nullptr;
        float *cur_ = nullptr;

        bool synced{false};
  };



  Probe(const std::shared_ptr< StructuredBlockForest >& blocks, const Setup& setup, const IDs& ids, const Config::BlockHandle& probeParameters)
      : blocks_(blocks), setup_(setup), ids_(ids), executionCounter_(uint_c(0)), totalExecutionCounter_(uint_c(0))
    {
        frequency_    = probeParameters.getParameter< uint_t >("evaluationFrequency");
        writeCounter_ = probeParameters.getParameter< uint_t >("writeCounter");
        rampUpTime_   = probeParameters.getParameter< uint_t >("rampUpTime", 0);

        auto createProbeData = []( IBlock * const , StructuredBlockStorage * const ) { return new DataVector(); };
        probeDataID = blocks_->addStructuredBlockData< DataVector >( createProbeData, "Probe Data");
    }
 
  void addPoint(const Vector3<real_t>& point) {
    probePoints_.emplace_back(point);

  }

  std::function<void()>  updateFunctor() {return [this](){ Probe::update();};}

  void initialise();
  void update();
  void setInitialTimestep(const uint_t initialTimestep){
    totalExecutionCounter_ = initialTimestep;
    rampUpTime_ += initialTimestep;
  }


  private:
  const std::shared_ptr< StructuredBlockForest > blocks_;
  const Setup setup_;
  const IDs ids_;
  BlockDataID probeDataID;

  uint_t frequency_;
  uint_t writeCounter_;
  uint_t executionCounter_;
  uint_t totalExecutionCounter_;
  uint_t rampUpTime_;

  const std::vector<std::string> dataSetNames{"rho", "ux", "uy", "uz"};
  std::vector<Vector3<real_t>> probePoints_;
  std::map<IBlock*, std::vector< ProbeInfo > > probeData_;
  std::map<IBlock*, std::pair<Vector3<uint32_t>, Vector3<uint32_t>>> gpuBlocksAndThreads_;

  void writeHeaderFiles();
  void writeProbes();
  void writeData(const std::vector<float>& data, const std::string& groupName, const std::string& dataSetName);

}; // class Probe
} // namespace walberla
