//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file Setup.h
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#pragma once

#include "core/config/Config.h"
#include "core/DataTypes.h"
#include "core/math/Vector3.h"
#include "core/waLBerlaBuildInfo.h"

#include "lbm_generated/refinement/RefinementScaling.h"

#include <string>

#include "Types.h"
#include "util.h"
#include "core/Filesystem.h"

namespace walberla
{

struct Setup
{

   Setup(std::shared_ptr<Config>& config, std::map<std::string, std::string>& infoMap)
   {

      auto parameters        = config->getOneBlock("Parameters");
      auto domainParameters  = config->getOneBlock("DomainSetup");
      auto logging           = config->getOneBlock("Logging");
      auto turbulence        = config->getOneBlock("Turbulence");
      auto snapshot          = config->getOneBlock("Snapshot");

      blockForestFilestem = domainParameters.getParameter< std::string >("blockForestFilestem", "blockforest");
      numProcesses        = domainParameters.getParameter< uint_t >("numberProcesses");
      domainSize          = domainParameters.getParameter< Vector3< real_t > >("domainSize");
      cellsPerBlock       = domainParameters.getParameter< Vector3< uint_t > >("cellsPerBlock");
      rootBlocks          = domainParameters.getParameter< Vector3< uint_t > >("blocks");
      cells               = Vector3<uint_t>(rootBlocks[0] * cellsPerBlock[0], rootBlocks[1] * cellsPerBlock[1], rootBlocks[2] * cellsPerBlock[2]);
      periodic            = domainParameters.getParameter< Vector3< bool > >("periodic");
      refinementLevels    = domainParameters.getParameter< uint_t >("refinementLevels");

      const real_t coarseMeshSize = domainSize[0] / real_c(rootBlocks[0] * cellsPerBlock[0]);
      WALBERLA_CHECK_FLOAT_EQUAL_EPSILON(coarseMeshSize, domainSize[1] / real_c(rootBlocks[1] * cellsPerBlock[1]), real_c(1e-12), "Resulting resolution is not equal in x and y direxction. This is not supported");
      WALBERLA_CHECK_FLOAT_EQUAL_EPSILON(coarseMeshSize, domainSize[2] / real_c(rootBlocks[2] * cellsPerBlock[2]), real_c(1e-12), "Resulting resolution is not equal in x and z direxction. This is not supported");

      referenceLength               = domainSize[1];
      reynoldsNumber                = parameters.getParameter< real_t >("reynoldsNumber");
      latticeVelocity               = parameters.getParameter< real_t >("latticeVelocity");
      initialiseWithlatticeVelocity = parameters.getParameter< bool >("initialiseWithlatticeVelocity");
      frictionVelocity              = real_c(0.0);

      useGrid               = turbulence.getParameter< bool >("useGrid");
      turbulentInflow       = turbulence.getParameter< bool >("turbulentInflow");
      urms                  = turbulence.getParameter< real_t >("U_rms");
      characteristicLength  = turbulence.getParameter< real_t >("characteristicLength");
      numberOfFourrierModes = turbulence.getParameter< uint_t >("numberOfFourrierModes");
      uc                    = turbulence.getParameter< Vector3< real_t > >("U_c");

      const real_t speedOfSound = real_c(real_c(1.0) / std::sqrt( real_c(3.0) ));
      drivingForce = real_c(0.0);

      if(config->getNumBlocks("RoughWall") > 0){
         // kinViscosity                    = parameters.getParameter< real_t >("viscosity");
         referenceLength *= real_c(0.5);
         const real_t latticeReferenceLength = referenceLength / coarseMeshSize;
         const real_t bulkReynoldsNumber = parameters.getParameter< real_t >("bulkReynoldsNumber");

         viscosity = latticeReferenceLength * latticeVelocity / bulkReynoldsNumber;
         referenceVelocity = latticeVelocity;
         dt  = latticeVelocity / referenceVelocity * coarseMeshSize;

         kinViscosity = referenceLength * referenceVelocity / bulkReynoldsNumber;

         machNumber = latticeVelocity / speedOfSound;
         omega      = real_c(real_c(1.0) / (real_c(3.0) * viscosity + real_c(0.5)));

         frictionVelocity = (reynoldsNumber * viscosity) / latticeReferenceLength;
         if(periodic[0])
            drivingForce = calculateDrivingForce(frictionVelocity, latticeReferenceLength, latticeVelocity, latticeVelocity);
      }
      else if(config->getNumBlocks("genericPorousMatrix") > 0){
         auto gpm = config->getOneBlock("genericPorousMatrix");

         referenceLength = gpm.getParameter< real_t >("barSize");
         const real_t latticeReferenceLength = referenceLength / coarseMeshSize;
         viscosity = latticeReferenceLength * latticeVelocity / reynoldsNumber;
         referenceVelocity = latticeVelocity;
         dt  = latticeVelocity / referenceVelocity * coarseMeshSize;
         kinViscosity = referenceLength * referenceVelocity / reynoldsNumber;
         machNumber = latticeVelocity / speedOfSound;
         omega      = real_c(real_c(1.0) / (real_c(3.0) * viscosity + real_c(0.5)));

         const real_t pressureGradient = parameters.getParameter< real_t >("pressureGradient");
         const real_t XYslice = real_c(rootBlocks[0] * cellsPerBlock[0] * rootBlocks[1] * cellsPerBlock[1]);
         if(periodic[0])
            drivingForce = pressureGradient / real_c(cells[0] * cells[1]);
      }
      else{
         referenceVelocity = parameters.getParameter< real_t >("referenceVelocity");
         const real_t latticeReferenceLength = referenceLength / coarseMeshSize;
         
         machNumber = latticeVelocity / speedOfSound;
         viscosity  = real_c((latticeVelocity * latticeReferenceLength) / reynoldsNumber);
         omega      = real_c(real_c(1.0) / (real_c(3.0) * viscosity + real_c(0.5)));

         dt  = latticeVelocity / referenceVelocity * coarseMeshSize;
         kinViscosity = viscosity * coarseMeshSize * coarseMeshSize / dt;
      }

      if( parameters.isDefined("drivingForce") )
         drivingForce = parameters.getParameter< real_t >("drivingForce");

      omegaF = lbm_generated::relaxationRateScaling(omega, refinementLevels);

      rho = real_c(1.0);
      dxC = coarseMeshSize;
      dxF = real_c(coarseMeshSize) / real_c( 1 << refinementLevels );
      
      conversionFactor = dxC / dt;

      timesteps = parameters.getParameter< uint_t >("timesteps");
      numGhostLayers = refinementLevels == 0 ? uint_c(1) : uint_c(2);

      flowThroughCycles = (real_c(timesteps) * dt) / (domainSize[0] / (latticeVelocity * conversionFactor));

      valuesPerCell = (Stencil_T::Q + VelocityField_T::F_SIZE + uint_c(1) * ScalarField_T::F_SIZE);
      memoryPerCell = memory_t(valuesPerCell * sizeof(PdfField_T::value_type) + uint_c(1));
      processMemoryLimit = parameters.getParameter< memory_t >( "processMemoryLimit", memory_t( 512 ) ) * memory_t( 1024 * 1024 );

      stencil          = infoMap["stencil"];
      streamingPattern = infoMap["streamingPattern"];
      collisionModel   = infoMap["collisionOperator"];

      fluidUID           = FlagUID("Fluid");
      noSlipUID          = FlagUID("NoSlip");
      obstacleUID        = FlagUID("Obstacle");
      inflowUID          = FlagUID("UBB");
      turbulentInflowUID = FlagUID("TurbulentInflow");
      outflowUID         = FlagUID("Outflow");

      writeSetupForestAndReturn    = logging.getParameter< bool >("writeSetupForestAndReturn", false);
      remainingTimeLoggerFrequency = logging.getParameter< real_t >("remainingTimeLoggerFrequency", 3.0);
      outputFolder                 = logging.getParameter< std::string >("outputFolder", "output");
      fileStem                     = logging.getParameter< std::string >("filestem", "data");

      loadSnapshot        = snapshot.getParameter< bool >("loadSnapshot", false);
      storeSnapshot       = snapshot.getParameter< bool >("storeSnapshot", false);
      recreateBlockForest = snapshot.getParameter< bool >("recreateBlockForest", false);
      snapshotFrequency   = snapshot.getParameter< uint_t >("snapshotFrequency", 0);
      snapshotBaseFolder  = snapshot.getParameter< std::string >("snapshotBaseFolder", "snapshots");

      checkpointConfigFile = snapshotBaseFolder + "/" + fileStem + "CheckpointConfig.file";
      pdfFieldFile = snapshotBaseFolder + "/" + fileStem + "pdfs.file";
      meanVelocityFile = snapshotBaseFolder + "/" + fileStem + "meanVelocity.file";
      sosFile = snapshotBaseFolder + "/" + fileStem + "sos.file";
      inflowFile = snapshotBaseFolder + "/" + fileStem + "inflow.file";

      WALBERLA_ROOT_SECTION(){
         if (!filesystem::is_directory(outputFolder) || !filesystem::exists(outputFolder)) { // Check if output folder exists
            filesystem::create_directory(outputFolder); // create output folder
         }

         if (!filesystem::is_directory(snapshotBaseFolder) || !filesystem::exists(snapshotBaseFolder)) {
            filesystem::create_directory(snapshotBaseFolder); 
         }
         
         writeInfoFile(config, infoMap);
      }
      WALBERLA_MPI_BARRIER()
   }

   void writeInfoFile(std::shared_ptr<Config>& config, std::map<std::string, std::string>& infoMap){
      const auto fileName = outputFolder + "/" + fileStem + "MetaData.txt";
      filesystem::path filePath( fileName.c_str() );
      if( filesystem::exists( filePath ) )
         std::remove( filePath.string().c_str() );

      WALBERLA_CHECK_NOT_NULLPTR(config)
      std::ofstream outfile( filePath.c_str() );

      outfile << "- simulation parameters:"   << std::setprecision(16) <<
      "\n   + collision model:       " << collisionModel <<
      "\n   + stencil:               " << stencil <<
      "\n   + streaming:             " << streamingPattern <<
      "\n   + forceModel:            " << infoMap["forceModel"] <<
      "\n   + FourthOrderCorrection: " << infoMap["FourthOrderCorrection"] <<
      "\n   + compressible:          " << ( StorageSpecification_T::compressible ? "yes" : "no" ) <<
      "\n- simulation properties:    "
      "\n   + kin. viscosity:        " << kinViscosity << " [m^2/s] (on the coarsest grid)" <<
      "\n   + viscosity LB:          " << viscosity  << " [dx*dx/dt] (on the coarsest grid)" <<
      "\n   + omega:                 " << omega << " (on the coarsest grid)" <<
      "\n   + omega:                 " << omegaF << " (on the finest grid)" <<
      "\n   + inflow velocity:       " << referenceVelocity << " [m/s]" <<
      "\n   + lattice velocity:      " << latticeVelocity << " [dx/dt]" <<
      "\n   + friction velocity:     " << frictionVelocity << " [dx/dt]" <<
      "\n   + Reynolds number:       " << reynoldsNumber <<
      "\n   + Mach number:           " << machNumber <<
      "\n   + driving force:         " << drivingForce <<
      "\n   + dx (coarsest grid):    " << dxC << " [m]" <<
      "\n   + dx (finest grid):      " << dxF << " [m]" <<
      "\n   + dt (coarsest grid):    " << dt << " [s]" <<
      "\n   + conversion Faktor:     " << conversionFactor <<
      "\n   + #time steps:           " << timesteps << " (on the coarsest grid, " << ( real_c(1.0) / dt ) << " for 1s of real time)" <<
      "\n   + simulation time:       " << ( real_c( timesteps ) * dt ) << " [s]"
      "\n   + flow through cycles:   " << ( flowThroughCycles ) << "\n\n";

      outfile << *config << "\n \n";

      outfile << "- waLBerla Build Info:"  << 
      "\n   + git SHA:          " << walberla::core::buildinfo::gitSHA1() <<
      "\n   + build Type:       " << walberla::core::buildinfo::buildType() <<
      "\n   + compiler Flags:   " << walberla::core::buildinfo::compilerFlags() <<
      "\n   + build Machine:    " << walberla::core::buildinfo::buildMachine() <<
      "\n   + source directory: " << walberla::core::buildinfo::sourceDir() <<
      "\n   + binary directory: " << walberla::core::buildinfo::binaryDir() << "\n\n";

      for (auto const& it : infoMap){
         outfile << it.first << " " << it.second << "\n";
      }

      outfile.close();
   }

   std::string blockForestFilestem;
   uint_t numProcesses;

   Vector3<uint_t> rootBlocks;
   Vector3<uint_t> cells;
   Vector3<real_t> domainSize;

   Vector3< bool > periodic;
   uint_t refinementLevels;

   Vector3< uint_t > cellsPerBlock;
   uint_t numGhostLayers;

   real_t referenceVelocity;
   real_t referenceLength;
   real_t latticeVelocity;
   real_t frictionVelocity;
   bool initialiseWithlatticeVelocity;

   bool useGrid;
   bool turbulentInflow;
   real_t urms;
   real_t characteristicLength;
   uint_t numberOfFourrierModes;
   Vector3<real_t> uc;

   real_t machNumber;
   real_t reynoldsNumber;

   real_t viscosity;
   real_t kinViscosity;
   real_t omega;
   real_t omegaF;
   real_t rho;
   real_t dxC;
   real_t dxF;
   real_t dt;
   real_t flowThroughCycles;

   real_t drivingForce;
   real_t conversionFactor;
   uint_t timesteps;

   uint_t valuesPerCell;
   memory_t memoryPerCell;
   memory_t processMemoryLimit;

   std::string stencil;
   std::string streamingPattern;
   std::string collisionModel;

   FlagUID fluidUID;
   FlagUID noSlipUID;
   FlagUID obstacleUID;
   FlagUID inflowUID;
   FlagUID turbulentInflowUID;
   FlagUID outflowUID;

   uint8_t inflowFlag;
   uint8_t turbulentInflowFlag;
   uint8_t outflowFlag;
   uint8_t noSlipFlag;
   uint8_t obstacleFlag;
   uint8_t fluidFlag;

   bool writeSetupForestAndReturn;
   real_t remainingTimeLoggerFrequency;
   std::string outputFolder;
   std::string fileStem;

   bool loadSnapshot;
   bool storeSnapshot;
   bool recreateBlockForest;
   uint_t snapshotFrequency;
   std::string snapshotBaseFolder;

   std::string checkpointConfigFile;
   std::string pdfFieldFile;
   std::string meanVelocityFile;
   std::string sosFile;
   std::string inflowFile;
};

}