//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file TPMS.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================

#include "TPMS.h"

namespace walberla
{

real_t TPMS::getDistance(const real_t x, const real_t y, const real_t z, const real_t scaling) const {
   const real_t kx = waveNumber_ * scaling;
   const real_t ky = waveNumber_ * scaling;
   const real_t kz = waveNumber_ * scaling;

   if(tpmsType_ == GYROID){
      return cos(kx * x) * sin(ky * y) + cos(ky * y) * sin(kz * z) + cos(kz * z) * sin(kx * x);
   }
   else{
      return real_c(0.0);
   }
}

bool TPMS::contains(const Vector3< real_t >& point) const
{
   const real_t dist1 = getDistance(point[0], point[1], point[2], real_c(1.0));
   const real_t dist2 = getDistance(point[0], point[1], point[2], scaling_);
   return (dist1 * dist2) >= strut_;
}

bool TPMS::contains(const AABB& aabb) const
{
   Vector3< real_t > p[8];
   p[0].set(aabb.xMin(), aabb.yMin(), aabb.zMin());
   p[1].set(aabb.xMax(), aabb.yMin(), aabb.zMin());
   p[2].set(aabb.xMin(), aabb.yMax(), aabb.zMin());
   p[3].set(aabb.xMax(), aabb.yMax(), aabb.zMin());
   p[4].set(aabb.xMin(), aabb.yMin(), aabb.zMax());
   p[5].set(aabb.xMax(), aabb.yMin(), aabb.zMax());
   p[6].set(aabb.xMin(), aabb.yMax(), aabb.zMax());
   p[7].set(aabb.xMax(), aabb.yMax(), aabb.zMax());
   return contains(p[0]) && contains(p[1]) && contains(p[2]) && contains(p[3]) && contains(p[4]) && contains(p[5]) &&
          contains(p[6]) && contains(p[7]);
}

void TPMS::setupBoundary(const std::shared_ptr< StructuredBlockForest >& sbfs, const BlockDataID flagFieldID){
   const real_t start = tpmsBeginn_ / setup_.dxC;
   const real_t end   = tpmsEnd_ / setup_.dxC;
   for (auto& block : *sbfs){
      const auto flagField = block.template getData< FlagField_T >(flagFieldID);
      const auto boundaryFlag = flagField->getOrRegisterFlag(setup_.obstacleUID);
      WALBERLA_FOR_ALL_CELLS_INCLUDING_GHOST_LAYER_XYZ(flagField, Cell globalCell;
      sbfs->transformBlockLocalToGlobalCell(globalCell, block, Cell(x, y, z));
      if (globalCell[0] > start && globalCell[0] <= end){
         Vector3<real_t> p;
         sbfs->getCellCenter(p, globalCell, 0);
         if(contains(p)){
            addFlag(flagField->get(x, y, z), boundaryFlag);
         }
      }
      )
   }
}

} // namespace walberla