//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file TPMS.h
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#pragma once

#include "blockforest/SetupBlock.h"
#include "blockforest/SetupBlockForest.h"
#include "blockforest/StructuredBlockForest.h"

#include "domain_decomposition/IBlock.h"

#include "field/FlagUID.h"

#include "core/DataTypes.h"
#include "core/math/AABB.h"
#include "core/math/Vector3.h"
#include "core/cell/Cell.h"
#include "core/config/Config.h"

#include "stencil/D3Q7.h"
#include "stencil/D3Q27.h"

#include "Types.h"
#include "Setup.h"

namespace walberla
{

class TPMS
{
 public:

   enum TPMSType {GYROID = 0};

   TPMS(const Config::BlockHandle& tpmsParameters, const Setup& setup) : setup_(setup){

      waveNumber_ = tpmsParameters.getParameter< real_t >("waveNumber");
      scaling_    = tpmsParameters.getParameter< real_t >("scaling");
      strut_      = tpmsParameters.getParameter< real_t >("strut");
      tpmsBeginn_ = tpmsParameters.getParameter< real_t >("start");
      tpmsEnd_    = tpmsParameters.getParameter< real_t >("end");

      const std::string tpmsString = tpmsParameters.getParameter< std::string >("type");

      if (tpmsString.compare("gyroid") == 0){
         tpmsType_ = GYROID;
      }
      else{
         WALBERLA_ABORT("Unsoprted TPMS Type: " << tpmsString);
      }
         
   }

   real_t getDistance(const real_t x, const real_t y, const real_t z, const real_t scaling) const;
   bool operator()(const Vector3< real_t >& point) const { return contains(point); }

   bool contains(const Vector3< real_t >& point) const;
   bool contains(const AABB& aabb) const;

   Setup getSetup(){ return setup_; }
   void setupBoundary(const std::shared_ptr< StructuredBlockForest >& sbfs, const BlockDataID flagFieldID);

 private:
   real_t waveNumber_;
   real_t scaling_;
   real_t strut_;
   real_t tpmsBeginn_;
   real_t tpmsEnd_;
   TPMSType tpmsType_;
   Setup setup_;
}; // class TPMS

/*
class TPMSRefinementSelection
{
 public:
   TPMSRefinementSelection(const TPMS& tpms){
      const auto setup = tpms.getSetup();

      const real_t start = tpms.tpmsBeginn;
      const real_t end   = tpms.tpmsEnd;
      boundingBox_ = AABB(start, real_c(0.0), real_c(0.0),
                          end  , setup.domainSize[1], setup.domainSize[2]);

      level_ = setup.refinementLevels;
   }

   void operator()(SetupBlockForest& forest)
   {
      if(level_ == 0)
         return;
      for (auto block = forest.begin(); block != forest.end(); ++block){
         const AABB& aabb = block->getAABB();

         if (block->getLevel() < level_ && aabb.intersects(boundingBox_) )
            block->setMarker(true);
      }
   }

 private:
   uint_t level_;
   AABB boundingBox_;

}; // class TPMSRefinementSelection
*/

} // namespace walberla