//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file util.cpp
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#include "core/logging/Logging.h"
#include "util.h"

namespace walberla
{

Vector3<real_t> summOfFourrierModes(const Vector3<real_t>& p, const Vector3<real_t>& uc,
                                    const std::vector<real_t>& kappa, const std::vector<real_t>& ut, const real_t t){

    Vector3<real_t> result(real_c(0.0), real_c(0.0), real_c(0.0));
    const real_t pi2 = real_c(6.28318530717958647693);
    const uint_t n = uint_c(kappa.size());

    for(uint_t i = 0; i < n; ++i){
        const real_t a     = math::realRandom(real_c(0.0), real_c(1.0));
        const real_t phi   = math::realRandom(real_c(0.0), real_c(1.0)) * pi2;
        const real_t psi   = math::realRandom(real_c(0.0), real_c(1.0)) * pi2;
        const real_t zetaX = math::realRandom(real_c(-1.0), real_c(1.0));
        const real_t zetaY = math::realRandom(real_c(-1.0), real_c(1.0));
        const real_t zetaZ = math::realRandom(real_c(-1.0), real_c(1.0));

        const real_t theta = std::acos(real_c(2.0) * a - real_c(1.0));
        
        const real_t kappaN = kappa[i];
        const real_t kappaX = std::cos(phi)   * std::sin(theta) * kappaN;
        const real_t kappaY = std::sin(theta) * std::sin(phi)   * kappaN;
        const real_t kappaZ = std::cos(theta) * kappaN;
        // WALBERLA_LOG_INFO_ON_ROOT("Kappa: [" << kappaX << ", " << kappaY << ", " << kappaZ << "]")

        const real_t zettaLength = std::sqrt(zetaX*zetaX + zetaY*zetaY + zetaZ*zetaZ);
        // WALBERLA_LOG_DEVEL_VAR(zettaLength)
        
        const real_t zetaXNormalized = zetaX / zettaLength;
        const real_t zetaYNormalized = zetaY / zettaLength;
        const real_t zetaZNormalized = zetaZ / zettaLength;
        // WALBERLA_LOG_INFO_ON_ROOT("zeta: [" << zetaXNormalized << ", " << zetaYNormalized << ", " << zetaZNormalized << "]")
        
        const real_t sigmaX = kappaY * zetaZNormalized - kappaZ * zetaYNormalized;
        const real_t sigmaY = kappaZ * zetaXNormalized - kappaX * zetaZNormalized;
        const real_t sigmaZ = kappaX * zetaYNormalized - kappaY * zetaXNormalized;

        const real_t sigmaLength = std::sqrt(sigmaX*sigmaX + sigmaY*sigmaY + sigmaZ*sigmaZ);
        // WALBERLA_LOG_DEVEL_VAR(sigmaLength)

        const real_t sigmaXNormalized = sigmaX / sigmaLength;
        const real_t sigmaYNormalized = sigmaY / sigmaLength;
        const real_t sigmaZNormalized = sigmaZ / sigmaLength;
        // WALBERLA_LOG_INFO_ON_ROOT("Sigma: [" << sigmaXNormalized << ", " << sigmaYNormalized << ", " << sigmaZNormalized << "]")

        const real_t pX = p[0] - (t * uc[0]);
        const real_t pY = p[1] - (t * uc[1]);
        const real_t pZ = p[2] - (t * uc[2]);

        const real_t arg =  real_c(2.0) * ut[i] * std::cos(kappaX * pX + kappaY * pY + kappaZ * pZ + psi);

        result[0] += arg * sigmaXNormalized;
        result[1] += arg * sigmaYNormalized;
        result[2] += arg * sigmaZNormalized;
    }
    return result;
}

real_t getKarmanPaoEnergy(const real_t k, const real_t alpha, const real_t kappaVKP, const real_t kappaKolmogorov, const real_t uRMS){
    const real_t a1 = alpha * std::pow(uRMS, real_c(2.0)) / kappaVKP;
    const real_t a2 = std::pow(k / kappaVKP, real_c(4.0));
    const real_t a3 = std::pow(real_c(1.0) + std::pow(k / kappaVKP, real_c(2.0)), real_c(17.0) / real_c(6.0));
    const real_t a4 = std::exp(real_c(-2.0) * std::pow(k / kappaKolmogorov, real_c(2.0)));
    return a1 * a2 / a3 * a4;
}

Vector3<real_t> getCenterOfHemisphere(const Vector3<real_t>& p, const real_t hemispheresDistance,
                                      const real_t domainYMin, const real_t domainYMax){

    const real_t spaceX     = hemispheresDistance / real_c(2.0) * math::root_three;
    const real_t zShift     = hemispheresDistance / real_c(2.0);
    const real_t domainHalf = (domainYMax - domainYMin) / real_c(2.0);

    const real_t xc = std::round(p[0] / spaceX) * spaceX;
    const real_t yc = p[1] > domainHalf ? real_c(domainYMax) : real_c(domainYMin);
    const real_t zs = real_c((uint_c(std::round(p[0] / spaceX)) % 2)) * zShift;
    const real_t zc = (std::round((p[2] - zs) / hemispheresDistance) * (hemispheresDistance)) + zs;

    return Vector3<real_t>(xc, yc, zc);
}

bool hemispheresContains(const Vector3<real_t>& point, const Vector3<real_t>& center, const real_t radius){
   return (point - center).sqrLength() <= (radius * radius);
}

void getAllHemisphereCenters(std::vector<Vector3<real_t>>& centers, const real_t hemispheresRadius, const real_t hemispheresDistance, const math::AABB& domain){

    const real_t spaceX     = hemispheresDistance / real_c(2.0) * math::root_three;
    const real_t zShift     = hemispheresDistance / real_c(2.0);

    const uint_t nx = uint_c((domain.xMax() - domain.xMin()) / spaceX) + 1;
    const uint_t nz = uint_c((domain.zMax() - domain.zMin()) / hemispheresDistance) + 1;

    for(uint_t x = 0; x <= nx; ++x){
        for(uint_t z = 0; z <= nz; ++z){
            const real_t cx = real_c(x) * spaceX;
            real_t cz = real_c(z) * hemispheresDistance;
            if(x%2 != 0){cz += zShift;}

            if(cx > (domain.xMax() + hemispheresRadius) || cz > (domain.zMax() + hemispheresRadius )){
                continue;
            }

            centers.emplace_back(Vector3<real_t>(cx, domain.yMin(), cz));
            centers.emplace_back(Vector3<real_t>(cx, domain.yMax(), cz));
        }
    }
}

real_t calculateDrivingForce(const real_t targetFrictionVelocity, const real_t referenceLenght, const real_t targetBulkVelocity, const real_t bulkVelocity){
   return targetFrictionVelocity * targetFrictionVelocity / referenceLenght + (targetBulkVelocity - bulkVelocity) * targetBulkVelocity / referenceLenght;
}

void renameFile(const std::string& oldName, const std::string& newName)
{
   int result = std::rename(oldName.c_str(), newName.c_str());
   if (result != 0)
      WALBERLA_LOG_WARNING_ON_ROOT("Could not rename file " << oldName << " to " << newName << " with error code "
                                                            << result);
}

} // namespace walberla