//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file utilGPU.cu
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#include "utilGPU.h"
#include <curand_kernel.h>

namespace walberla
{

namespace internal{

#if __CUDA_ARCH__ < 600
__device__ double atomicAdd(double* address, double val)
{
    unsigned long long int* address_as_ull = (unsigned long long int*)address;
    unsigned long long int old = *address_as_ull, assumed;

    do {
        assumed = old;
        old = atomicCAS(address_as_ull, assumed, __double_as_longlong(val + __longlong_as_double(assumed)));

    // Note: uses integer comparison to avoid hang in case of NaN (since NaN != NaN)
    } while (assumed != old);

    return __longlong_as_double(old);
}
#endif

static __device__ void singleMode(double* rX, double* rY, double* rZ, const double kappaN, const double ut,
                                  const double pX, const double pY, const double pZ, const int64_t linear_index)
{

   const real_t pi2 = 6.28318530717958647693;

   curandStatePhilox4_32_10_t state;
   curand_init(42, linear_index, 0, &state);

   const float4 r1 = curand_uniform4(&state);
   const float4 r2 = curand_uniform4(&state);

   const real_t a   = real_t(r1.x);
   const real_t phi = real_t(r1.y) * pi2;
   const real_t psi = real_t(r1.z) * pi2;
   const real_t zetaX = real_t(2.0) * real_t(r1.w) - real_t(1.0);
   const real_t zetaY = real_t(2.0) * real_t(r2.x) - real_t(1.0);
   const real_t zetaZ = real_t(2.0) * real_t(r2.y) - real_t(1.0);

/*
   // For debugging
   const real_t a     = 0.1; // math::realRandom(real_c(0.0), real_c(1.0));
   const real_t phi   = 0.2 * pi2; //math::realRandom(real_c(0.0), real_c(1.0)) * pi2;
   const real_t psi   = 0.3 * pi2; //math::realRandom(real_c(0.0), real_c(1.0)) * pi2;
   const real_t zetaX = -0.4; //math::realRandom(real_c(-1.0), real_c(1.0));
   const real_t zetaY = -0.6; //math::realRandom(real_c(-1.0), real_c(1.0));
   const real_t zetaZ = 0.8; // math::realRandom(real_c(-1.0), real_c(1.0));
*/

   const real_t theta = acos(real_t(2.0) * a - real_t(1.0));

   const real_t kappaX = cos(phi) * sin(theta) * kappaN;
   const real_t kappaY = sin(theta) * sin(phi) * kappaN;
   const real_t kappaZ = cos(theta) * kappaN;

   const real_t zettaLength = rsqrt(zetaX*zetaX + zetaY*zetaY + zetaZ*zetaZ);
   
   const real_t zetaXNormalized = zetaX * zettaLength;
   const real_t zetaYNormalized = zetaY * zettaLength;
   const real_t zetaZNormalized = zetaZ * zettaLength;

   const real_t sigmaX = kappaY * zetaZNormalized - kappaZ * zetaYNormalized;
   const real_t sigmaY = kappaZ * zetaXNormalized - kappaX * zetaZNormalized;
   const real_t sigmaZ = kappaX * zetaYNormalized - kappaY * zetaXNormalized;

   const real_t sigmaLength = rsqrt(sigmaX*sigmaX + sigmaY*sigmaY + sigmaZ*sigmaZ);

   const real_t sigmaXNormalized = sigmaX * sigmaLength;
   const real_t sigmaYNormalized = sigmaY * sigmaLength;
   const real_t sigmaZNormalized = sigmaZ * sigmaLength;

   const real_t arg =  real_t(2.0) * ut * cos(kappaX * pX + kappaY * pY + kappaZ * pZ + psi);

   *rX = arg * sigmaXNormalized;
   *rY = arg * sigmaYNormalized;
   *rZ = arg * sigmaZNormalized;
   
}


static __global__ void summOfFourrierModesKernel(double * velocity, double * const kappa, double * const ut,
                                                 const real_t yMin, const real_t zMin, const real_t minCornerY, const real_t minCornerZ,
                                                 const real_t dx, const real_t ucX, const real_t ucY, const real_t ucZ,
                                                 const real_t t, const uint_t n, const real_t latticeVelocity, const int64_t nGhostLayer,
                                                 const int64_t ySize, const int64_t zSize,
                                                 const int64_t yStride, const int64_t zStride, const int64_t fStride)
{
   if(blockDim.y*blockIdx.y + threadIdx.y < ySize && blockDim.z*blockIdx.z + threadIdx.z < zSize)
   {
      extern __shared__ double sharedMemory[];

      double* uPrimeX = &sharedMemory[0];
      double* uPrimeY = &sharedMemory[blockDim.x];
      double* uPrimeZ = &sharedMemory[2*blockDim.x];

      const int64_t thx = threadIdx.x;
      const int64_t y   = blockDim.y*blockIdx.y + threadIdx.y;
      const int64_t z   = blockDim.z*blockIdx.z + threadIdx.z;

      uPrimeX[thx] = 0.0;
      uPrimeY[thx] = 0.0;
      uPrimeZ[thx] = 0.0;

      const real_t xCenter = real_t(0.0);
      const real_t yCenter = (yMin + ( real_t( y - nGhostLayer ) + real_t(0.5) ) * dx) + minCornerY;
      const real_t zCenter = (zMin + ( real_t( z - nGhostLayer ) + real_t(0.5) ) * dx) + minCornerZ;

      const real_t pX = xCenter - (t * ucX);
      const real_t pY = yCenter - (t * ucY);
      const real_t pZ = zCenter - (t * ucZ);

      for(uint_t i = 0; i < n; ++i){
         const int64_t x = thx + i * blockDim.x;
         const int64_t linear_index = x + (y * n * blockDim.x) + (z * n * blockDim.x * ySize);

         double rX = 0.0;
         double rY = 0.0;
         double rZ = 0.0;

         singleMode(&rX, &rY, &rZ, kappa[x], ut[x], pX, pY, pZ, linear_index);

         uPrimeX[thx] += rX;
         uPrimeY[thx] += rY;
         uPrimeZ[thx] += rZ;
      }

      __syncthreads();

      for (uint_t s = 1; s < blockDim.x; s *= 2) {
         int index = 2 * s * thx;
         if (index < blockDim.x) {
            uPrimeX[index] += uPrimeX[index + s];
            uPrimeY[index] += uPrimeY[index + s];
            uPrimeZ[index] += uPrimeZ[index + s];
         }
      __syncthreads();
      }

      velocity[yStride*y + zStride*z + 0*fStride] = latticeVelocity + uPrimeX[0];
      velocity[yStride*y + zStride*z + 1*fStride] = uPrimeY[0];
      velocity[yStride*y + zStride*z + 2*fStride] = uPrimeZ[0];
   }
}


static __global__ void evaluateProbePoints(double* const density, double* const velocity, int64_t* cellData, float* probeData,
                                           const int64_t n, const float velocityScalingFactor,
                                           const int64_t xStrideDensity, const int64_t yStrideDensity, const int64_t zStrideDensity,
                                           const int64_t xStrideVelocity, const int64_t yStrideVelocity, const int64_t zStrideVelocity, const int64_t fStrideVelocity)
{
   if(blockDim.x*blockIdx.x + threadIdx.x < n)
   {
      const int64_t ctr = blockDim.x*blockIdx.x + threadIdx.x;

      const int64_t x = cellData[ctr];
      const int64_t y = cellData[ctr + n];
      const int64_t z = cellData[ctr + 2*n];

      const float rho = ((float)(density[xStrideDensity*x + yStrideDensity*y + zStrideDensity*z]));
      const float uX = ((float)(velocity[xStrideVelocity*x + yStrideVelocity*y + zStrideVelocity*z]));
      const float uY = ((float)(velocity[xStrideVelocity*x + yStrideVelocity*y + zStrideVelocity*z + fStrideVelocity]));
      const float uZ = ((float)(velocity[xStrideVelocity*x + yStrideVelocity*y + zStrideVelocity*z + 2*fStrideVelocity]));

      probeData[ctr] = rho;
      probeData[ctr + n]   = velocityScalingFactor * uX;
      probeData[ctr + 2*n] = velocityScalingFactor * uY;
      probeData[ctr + 3*n] = velocityScalingFactor * uZ;
   }
}

static __global__ void sumVelocity(double* result, double* const velocity, uint8_t* const flagField, const int64_t n, const double cellVolume, const uint_t fluidBitNr,
                                   const int64_t xSize, const int64_t ySize, const int64_t zSize,
                                   const int64_t xStrideVelocity, const int64_t yStrideVelocity, const int64_t zStrideVelocity, const int64_t fStrideVelocity,
                                   const int64_t xStrideFlag, const int64_t yStrideFlag, const int64_t zStrideFlag)
{
   if(blockDim.y*blockIdx.y + threadIdx.y < ySize && blockDim.z*blockIdx.z + threadIdx.z < zSize)
   {

      extern __shared__ double sdata[];

      const int64_t thx = threadIdx.x;
      const int64_t x = blockIdx.x * (blockDim.x * 2) + threadIdx.x;
      const int64_t y = blockDim.y*blockIdx.y + threadIdx.y;
      const int64_t z = blockDim.z*blockIdx.z + threadIdx.z;

      // const double ux = ((flagField[xStrideFlag*x + yStrideFlag*y + zStrideFlag*z]) >> (fluidBitNr) & 1) * cellVolume * velocity[xStrideVelocity*x + yStrideVelocity*y + zStrideVelocity*z + 0*fStrideVelocity];
      // atomicAdd(&result[n], ux);

      
      double firstSum = (x < xSize) ? ((flagField[xStrideFlag*x + yStrideFlag*y + zStrideFlag*z]) >> (fluidBitNr) & 1) * cellVolume * velocity[xStrideVelocity*x + yStrideVelocity*y + zStrideVelocity*z + 0*fStrideVelocity] : 0;
      if (x + blockDim.x < xSize) firstSum += ((flagField[xStrideFlag*(x + blockDim.x) + yStrideFlag*y + zStrideFlag*z]) >> (fluidBitNr) & 1) * cellVolume * velocity[xStrideVelocity*(x + blockDim.x) + yStrideVelocity*y + zStrideVelocity*z + 0*fStrideVelocity];
      sdata[thx] = firstSum;
      __syncthreads();

      for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1) {
         if (thx < s) {
            sdata[thx] += sdata[thx + s];
         }
         __syncthreads();
      }

      if (thx == 0) atomicAdd(&result[n], sdata[0]);
      

   }
}

}


void getBlockAndGridSize(Vector3<uint_t> s, Vector3<uint32_t>& block, Vector3<uint32_t>& grid){
   
   const int64_t b0 = 128;
   const int64_t b1 = 1;
   const int64_t b2 = 1;

   const int64_t x0 = int64_c(s[0]);
   const int64_t x1 = int64_c(s[1]);
   const int64_t x2 = int64_c(s[2]);

   block[0] = uint32_c((1024 < ((b0 < x0) ? b0 : x0)) ? 1024 : ((b0 < x0) ? b0 : x0));
   block[1] = uint32_c((1024 < ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))) ? 1024 : ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))));
   block[2] = uint32_c((64 < ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))))) ? 64 : ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))));
   grid[0] =  uint32_c((x0) % (((1024 < ((b0 < x0) ? b0 : x0)) ? 1024 : ((b0 < x0) ? b0 : x0))) == 0 ? (int64_t)(x0) / (int64_t)(((1024 < ((b0 < x0) ? b0 : x0)) ? 1024 : ((b0 < x0) ? b0 : x0))) : ( (int64_t)(x0) / (int64_t)(((1024 < ((b0 < x0) ? b0 : x0)) ? 1024 : ((b0 < x0) ? b0 : x0))) ) +1 );
   grid[1] =  uint32_c((x1) % (((1024 < ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))) ? 1024 : ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))) == 0 ? (int64_t)(x1) / (int64_t)(((1024 < ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))) ? 1024 : ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))) : ( (int64_t)(x1) / (int64_t)(((1024 < ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))) ? 1024 : ((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))) ) +1 );
   grid[2] =  uint32_c((x2) % (((64 < ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))))) ? 64 : ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))))) == 0 ? (int64_t)(x2) / (int64_t)(((64 < ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))))) ? 64 : ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))))) : ( (int64_t)(x2) / (int64_t)(((64 < ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))))))) ? 64 : ((x2 < b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))) ? x2 : b2*((int64_t)(b0*b1) / (int64_t)(((b0 < x0) ? b0 : x0)*((x1 < b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0)))) ? x1 : b1*((int64_t)(b0) / (int64_t)(((b0 < x0) ? b0 : x0))))))))) ) +1 );   
}

void updateVelocityPlane(gpu::GPUField<real_t>* velocityGPU, real_t* kappa, real_t* ut, const Vector3<real_t>& globalMinCorner, const Vector3<real_t>& blockMinCorner, const Vector3<real_t>& uc,
                         const real_t dx, const real_t t, const uint_t n, const real_t latticeVelocity, const int64_t nGhostLayers,
                         const Vector3<uint32_t>& block, const Vector3<uint32_t>& grid){

   dim3 block_(block[0], block[1], block[2]);
   dim3 grid_(grid[0], grid[1], grid[2]);

   WALBERLA_ASSERT_EQUAL(grid[0], 0, "The summOfFourrierModesKernel kernel contains a reduction which is solved using a loop inside the block of threads. This does not work with multiple threads.")

   // In shared memory we store the result of the u_prime (x, y, z)
   const uint_t sharedMemorySize = 3 * block[0] * sizeof(real_t);
   const uint_t factor = n / block[0];
 
   double * data_velocity = velocityGPU->dataAt(-nGhostLayers, -nGhostLayers, -nGhostLayers, 0);
   const int64_t size_velocity_1 = int64_c(velocityGPU->ySize()) + 2 * nGhostLayers;
   const int64_t size_velocity_2 = int64_c(velocityGPU->zSize()) + 2 * nGhostLayers;
   const int64_t stride_velocity_1 = int64_c(velocityGPU->yStride());
   const int64_t stride_velocity_2 = int64_c(velocityGPU->zStride());
   const int64_t stride_velocity_3 = int64_c(velocityGPU->fStride());
   internal::summOfFourrierModesKernel<<<grid_, block_, sharedMemorySize, nullptr>>>(data_velocity, kappa, ut,
                                                                      globalMinCorner[1], globalMinCorner[2], blockMinCorner[1], blockMinCorner[2],
                                                                      dx, uc[0], uc[1], uc[2], t, factor, latticeVelocity, nGhostLayers,
                                                                      size_velocity_1, size_velocity_2,
                                                                      stride_velocity_1, stride_velocity_2, stride_velocity_3);

}

void updateProbe(gpu::GPUField<real_t>* density, gpu::GPUField<real_t>* velocity, int64_t* cellData, float* probeData, const int64_t n, const float velocityScalingFactor,
                 const Vector3<uint32_t>& block, const Vector3<uint32_t>& grid){

   dim3 block_(block[0], block[1], block[2]);
   dim3 grid_(grid[0], grid[1], grid[2]);

   double * data_density = density->dataAt(0, 0, 0, 0);
   double * data_velocity = velocity->dataAt(0, 0, 0, 0);
  
   const int64_t xStrideDensity = int64_c(density->xStride());
   const int64_t yStrideDensity = int64_c(density->yStride());
   const int64_t zStrideDensity = int64_c(density->zStride());

   const int64_t xStrideVelocity = int64_c(velocity->xStride());
   const int64_t yStrideVelocity = int64_c(velocity->yStride());
   const int64_t zStrideVelocity = int64_c(velocity->zStride());
   const int64_t fStrideVelocity = int64_c(velocity->fStride());
   internal::evaluateProbePoints<<<grid_, block_, 0, nullptr>>>(data_density, data_velocity, cellData, probeData, n, velocityScalingFactor,
                                                                xStrideDensity, yStrideDensity, zStrideDensity, xStrideVelocity, yStrideVelocity, zStrideVelocity, fStrideVelocity);

}

void BulkVelocityCalculator::calculateBulkVelocity() {
   dim3 b(block_[0], block_[1], block_[2]);
   dim3 g(grid_[0], grid_[1], grid_[2]);

   const uint_t sharedMemorySize = block_[0] * sizeof(double);

   int64_t counter = 0;
   for( auto it = blocks_->begin(); it != blocks_->end(); ++it ){
      auto* block = dynamic_cast< Block* >(it.get());
      const uint_t level = block->getLevel();
      const double scaleFactor = double_c(1.0) / double_c( 1 << level );
      const double cellVolume = scaleFactor * scaleFactor * scaleFactor;
      GPUField_T* velocityField = block->template getData<GPUField_T>(ids_.velocityFieldGPU);
      GPUFlagField_T* flagField = block->template getData<GPUFlagField_T>(ids_.flagFieldGPU);
      
      double* const data_velocity = velocityField->dataAt(0, 0, 0, 0);
      uint8_t* const data_flagField = flagField->dataAt(0, 0, 0, 0);

      const int64_t xSize = int64_c(velocityField->xSize());
      const int64_t ySize = int64_c(velocityField->ySize());
      const int64_t zSize = int64_c(velocityField->zSize());

      const int64_t xStrideVelocity = int64_c(velocityField->xStride());
      const int64_t yStrideVelocity = int64_c(velocityField->yStride());
      const int64_t zStrideVelocity = int64_c(velocityField->zStride());
      const int64_t fStrideVelocity = int64_c(velocityField->fStride());

      const int64_t xStrideFlag = int64_c(flagField->xStride());
      const int64_t yStrideFlag = int64_c(flagField->yStride());
      const int64_t zStrideFlag = int64_c(flagField->zStride());

      internal::sumVelocity<<<g, b, sharedMemorySize, nullptr>>>(resultGPU_, data_velocity, data_flagField, counter, cellVolume, fluidBitNr_,
                                                                 xSize, ySize, zSize,
                                                                 xStrideVelocity, yStrideVelocity, zStrideVelocity, fStrideVelocity,
                                                                 xStrideFlag, yStrideFlag, zStrideFlag);
      counter++;

      
   }
   WALBERLA_GPU_CHECK(gpuMemcpy(&resultCPU_[0], resultGPU_, sizeof(double) * resultCPU_.size(), gpuMemcpyDeviceToHost ));

   double bulkVelocity = double_c(0.0);
   for (auto& bulkVelocityPerBlock : resultCPU_){
      bulkVelocity += bulkVelocityPerBlock;
   }

   mpi::allReduceInplace< double >(bulkVelocity, mpi::SUM, MPIManager::instance()->comm());
   bulkVelocity_ = double_c(bulkVelocity) / volume_;

   WALBERLA_GPU_CHECK(gpuMemset(resultGPU_, 0.0, sizeof(double) * resultCPU_.size()));
}


} // namespace walberla