//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file utilGPU.h
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================
#pragma once

#include "blockforest/Block.h"

#include "core/DataTypes.h"
#include "core/math/Constants.h"
#include "core/Filesystem.h"
#include "core/mpi/Reduce.h"

#include "gpu/GPUWrapper.h"
#include "gpu/ErrorChecking.h"
#include "gpu/FieldCopy.h"
#include "gpu/GPUField.h"

#include "Types.h"
#include "Setup.h"

namespace walberla{

void getBlockAndGridSize(Vector3<uint_t> s, Vector3<uint32_t>& block, Vector3<uint32_t>& grid);

void updateVelocityPlane(gpu::GPUField<real_t>* velocityGPU, real_t* kappa, real_t* ut, const Vector3<real_t>& globalMinCorner, const Vector3<real_t>& blockMinCorner, const Vector3<real_t>& uc,
                         const real_t dx, const real_t t, const uint_t n, const real_t latticeVelocity, const int64_t nGhostLayers,
                         const Vector3<uint32_t>& block, const Vector3<uint32_t>& grid);

void updateProbe(gpu::GPUField<real_t>* density, gpu::GPUField<real_t>* velocity, int64_t* cellData, float* probeData, const int64_t n, const float velocityScalingFactor,
                 const Vector3<uint32_t>& block, const Vector3<uint32_t>& grid);


class BulkVelocityCalculator {
public:
    BulkVelocityCalculator(const std::shared_ptr<StructuredBlockForest> & blocks, const IDs& ids, const Setup& setup, const Config::BlockHandle& forceParameters) 
    : blocks_(blocks), setup_(setup), ids_(ids) {

        bool snapshotExists = false;
        const std::string snapshotFilename = setup_.snapshotBaseFolder + "/" + setup_.fileStem + "DrivingForceSnapshot.file";
        filesystem::path snapshotFilePath( snapshotFilename.c_str() );
        if( filesystem::exists( snapshotFilePath ) )
            snapshotExists = true;

        if(snapshotExists && setup_.loadSnapshot){
            WALBERLA_LOG_INFO_ON_ROOT("Intialise PID Controller from Snapshot")
            forceUpdateFrequency_ = forceParameters.getParameter< uint_t >("forceUpdateFrequency", 0);
            adaptDrivingForce_ = forceParameters.getParameter< bool >("adaptDrivingForce", false);
            writeFrequency_ = forceParameters.getParameter< uint_t >("writeFrequency", 100);
            targetBulkVelocity_ = forceParameters.isDefined("targetBulkVelocity") ? forceParameters.getParameter< real_t >("targetBulkVelocity") : setup_.latticeVelocity;
            readStateFromFile();

            if(!adaptDrivingForce_){
                drivingForce_ = setup_.drivingForce;
            }
        }
        else{
            forceUpdateFrequency_ = forceParameters.getParameter< uint_t >("forceUpdateFrequency", 0);
            adaptDrivingForce_ = forceParameters.getParameter< bool >("adaptDrivingForce", false);
            writeFrequency_ = forceParameters.getParameter< uint_t >("writeFrequency", 100);
            targetBulkVelocity_ = forceParameters.isDefined("targetBulkVelocity") ? forceParameters.getParameter< real_t >("targetBulkVelocity") : setup_.latticeVelocity;
            bulkVelocity_ = setup_.latticeVelocity;
            drivingForce_ = setup_.drivingForce;

            proportionalGain_ = forceParameters.getParameter< real_t >("proportionalGain");
            derivateGain_     = forceParameters.getParameter< real_t >("derivativeGain");
            integralGain_     = forceParameters.getParameter< real_t >("integralGain");
            maxRamp_          = forceParameters.getParameter< real_t >("maxRamp");

            minActuatingVariable_ = drivingForce_ - real_c(1000) * drivingForce_;
            maxActuatingVariable_ = drivingForce_ + real_c(1000) * drivingForce_;

            std::fill(errorHistory_, errorHistory_ + sizeof(errorHistory_) / sizeof(real_t), real_c(0.0));

            if (integralGain_ > real_c(0.0))
                errorIntegral_ = drivingForce_ / integralGain_;
            else
                errorIntegral_ = real_c(0.0);

            executionCounter_ = uint_c(0);
        }

        volume_ = 0.0;
        uint_t nBlocks = 0;
        for( auto it = blocks_->begin(); it != blocks_->end(); ++it ){
            auto* block = dynamic_cast< Block* >(it.get());
            const uint_t level = block->getLevel();

            const double scaleFactor = double_c(1.0) / double_c( 1 << level );
            const double cellVolume = scaleFactor * scaleFactor * scaleFactor;
            auto * flagField = block->template getData<FlagField_T>(ids.flagField);
            auto domainFlag = flagField->getFlag(setup.fluidUID);
            const auto ci = flagField->xyzSize();
            for(auto cellIt = ci.begin(); cellIt != ci.end(); ++cellIt) {
                if(isFlagSet(flagField->get(*cellIt), domainFlag)){
                    volume_ += cellVolume;
                }
            }
            nBlocks++;
        }

        const Vector3<uint_t> s(blocks_->getNumberOfXCellsPerBlock(), blocks_->getNumberOfYCellsPerBlock(), blocks_->getNumberOfZCellsPerBlock());
        getBlockAndGridSize(s, block_, grid_);
        mpi::allReduceInplace< double >(volume_, mpi::SUM, MPIManager::instance()->comm());
        
        resultCPU_.resize(nBlocks);

        WALBERLA_GPU_CHECK(gpuMalloc((void **)&resultGPU_, sizeof(double) * resultCPU_.size() ));
        WALBERLA_GPU_CHECK(gpuMemset(resultGPU_, 0.0, sizeof(double) * resultCPU_.size()));

        fluidBitNr_ = uint_c(std::log(setup_.fluidFlag) / std::log(2));

        fileName_ = setup_.outputFolder + "/" + setup_.fileStem + "DrivingForce.csv"; 
        filesystem::path drivingForceOuputFile( fileName_.c_str() );

        WALBERLA_ROOT_SECTION()
        {
            if(!setup_.loadSnapshot){
                if( filesystem::exists( drivingForceOuputFile ) )
                    std::remove( drivingForceOuputFile.string().c_str() );
            }

            if(!filesystem::exists( drivingForceOuputFile ) ){
                std::ofstream outfile( drivingForceOuputFile.c_str() );
                outfile << "SEP=," << "\n";

                outfile << "timestep," << "time," << "force," << "U_b" << "\n";
                outfile.close();
            }
        }
    }

    void writeStateToFile() const{
        const std::string filename = setup_.snapshotBaseFolder + "/" + setup_.fileStem + "DrivingForceSnapshot.file";
        std::ofstream file;
        file.open(filename);
        file << std::setprecision(16);
        file << bulkVelocity_ << std::endl;
        file << drivingForce_ << std::endl;
        file << proportionalGain_ << std::endl;
        file << derivateGain_ << std::endl;
        file << integralGain_ << std::endl;
        file << maxRamp_ << std::endl;
        file << minActuatingVariable_ << std::endl;
        file << maxActuatingVariable_ << std::endl;
        file << errorHistory_[0] << std::endl;
        file << errorHistory_[1] << std::endl;
        file << errorHistory_[2] << std::endl;
        file << errorIntegral_ << std::endl;
        file << executionCounter_ << std::endl;
        file.close();
    }

    void readStateFromFile(){
        const std::string filename = setup_.snapshotBaseFolder + "/" + setup_.fileStem + "DrivingForceSnapshot.file";
        std::ifstream file;
        file.open(filename);
        file >> bulkVelocity_;
        file >> drivingForce_;
        file >> proportionalGain_;
        file >> derivateGain_;
        file >> integralGain_;
        file >> maxRamp_;
        file >> minActuatingVariable_;
        file >> maxActuatingVariable_;
        file >> errorHistory_[0];
        file >> errorHistory_[1];
        file >> errorHistory_[2];
        file >> errorIntegral_;
        file >> executionCounter_;
        file.close();
    }


    uint_t forceUpdateFrequency() const {return forceUpdateFrequency_;}
    bool adaptDrivingForce() const {return adaptDrivingForce_;}
    double bulkVelocity() const { return bulkVelocity_; }
    void setBulkVelocity(const double bulkVelocity) { bulkVelocity_ = bulkVelocity; }
    void calculateBulkVelocity();

    real_t getDrivingForce() {
        calculateBulkVelocity();

        const real_t timestep = real_c(executionCounter_ * forceUpdateFrequency_) * setup_.dt;
        if(!adaptDrivingForce_){
            if(executionCounter_ % writeFrequency_ == 0){
                WALBERLA_ROOT_SECTION(){
                    std::ofstream outfile( fileName_.c_str(), std::ios_base::app );
                    outfile << executionCounter_ * forceUpdateFrequency_ << "," << timestep << "," << drivingForce_ << "," << bulkVelocity_ << "\n";
                    outfile.close();
                }
            }
            executionCounter_++;
            return drivingForce_;
        }

        const real_t error = targetBulkVelocity_ - bulkVelocity_;
        const real_t d = (error + real_c(3.0) * errorHistory_[0] - real_c(3.0) * errorHistory_[1] - errorHistory_[2]) * (real_c(1.0) / real_c(6.0));

        std::rotate(errorHistory_, errorHistory_ + 1, errorHistory_ + sizeof(errorHistory_) / sizeof(real_t));
        errorHistory_[sizeof(errorHistory_) / sizeof(real_t) - size_t(1)] = error;

        real_t newDrivingForce = proportionalGain_ * error + derivateGain_ * d + integralGain_ * errorIntegral_;

        if (std::fabs(drivingForce_ - newDrivingForce) < maxRamp_){
            errorIntegral_ += error;
            newDrivingForce = proportionalGain_ * error + derivateGain_ * d + integralGain_ * errorIntegral_;
        }

        const real_t maxValue = std::min(drivingForce_ + maxRamp_, maxActuatingVariable_);
        const real_t minValue = std::max(drivingForce_ - maxRamp_, minActuatingVariable_);

        drivingForce_ = std::min(std::max(minValue, newDrivingForce), maxValue);

        if(executionCounter_ % writeFrequency_ == 0){
            WALBERLA_ROOT_SECTION(){
                std::ofstream outfile( fileName_.c_str(), std::ios_base::app );
                outfile << executionCounter_ * forceUpdateFrequency_ << "," << timestep << "," << drivingForce_ << "," << bulkVelocity_ << "\n";
                outfile.close();
            }
        }

        executionCounter_++;
        return drivingForce_;
    }

private:
    const std::shared_ptr< StructuredBlockForest > blocks_;
    const Setup setup_;
    const IDs ids_;

    uint_t forceUpdateFrequency_;
    uint_t writeFrequency_;
    bool adaptDrivingForce_;

    real_t bulkVelocity_;
    real_t drivingForce_;
    real_t targetBulkVelocity_;

    real_t proportionalGain_;
    real_t derivateGain_;
    real_t integralGain_;

    real_t maxRamp_;
    real_t minActuatingVariable_;
    real_t maxActuatingVariable_;

    real_t errorHistory_[3];
    real_t errorIntegral_;

    uint_t executionCounter_;
    std::string fileName_;

    double volume_;
    uint_t fluidBitNr_;

    std::vector<double> resultCPU_;
    double* resultGPU_;

    Vector3<uint32_t> block_;
    Vector3<uint32_t> grid_;
};


} // namespace walberla